@@ -18,9 +18,13 @@
 #include <pwd.h>
 #include <sys/stat.h>
 
-#include "Tw/Tw.h"
-#include "Tw/Twkeys.h"
-#include "Tw/Twerrno.h"
+#include <Tw/Tw.h>
+#include <Tw/Twkeys.h>
+#include <Tw/Twerrno.h>
+
+#ifdef CONF__UNICODE
+# include <Tutf/Tutf.h>
+#endif
 
 #include "version.h"
 
@@ -73,7 +77,7 @@ static byte DM_Kill, quiet, logged_in;
 static TW_VOLATILE pid_t ServerPid = (pid_t)-1, AttachPid = (pid_t)-1;
 
 static byte * TW_CONST * Args;
-static TW_CONST byte *hw_name = "-hw=tty", *title;
+static TW_CONST byte *hw_name = "--hw=tty", *title, *TwEnvRC = NULL;
 static byte use_twdisplay = DM_ATTACH, Privileges = TW_PRIV_NONE;
 
 typedef struct s_data {
@@ -87,16 +91,17 @@ static struct s_data user, pass;
 static void Usage(void) {
     fprintf(stderr, "Usage: twdm [OPTIONS]\n"
 	    "Currently known options: \n"
-	    " -h, -help               display this help and exit\n"
-	    " -V, -version            output version information and exit\n"
-	    " -k, -kill               kill twin server upon display detach\n"
-	    " -q, -quiet              quiet; suppress diagnostic messages\n"
-	    " -attach                 use \"twattach\" to start display (default)\n"
-	    " -display                use \"twdisplay\" to start display\n"
-	    " -suidroot               tell twin to keep suid root privileges\n"
-	    " -sgidtty                tell twin to keep sgid tty privileges\n"
-	    " -title=<title>          set window title\n"
-	    " -hw=<arg>               set display hw to use (default: -hw=tty)\n");
+	    " -h, --help              display this help and exit\n"
+	    " -V, --version           output version information and exit\n"
+	    " -k, --kill              kill twin server upon display detach\n"
+	    " -q, --quiet             quiet; suppress diagnostic messages\n"
+	    " --attach                use \"twattach\" to start display (default)\n"
+	    " --display               use \"twdisplay\" to start display\n"
+	    " --envrc                 tell twin to run .twenvrc.sh to get environment\n"
+	    " --suidroot              tell twin to keep suid root privileges\n"
+	    " --sgidtty               tell twin to keep sgid tty privileges\n"
+	    " --title=<title>         set window title\n"
+	    " --hw=<arg>              set display hw to use (default: --hw=tty)\n");
 }
 
 static void ShowVersion(void) {
@@ -106,34 +111,36 @@ static void ShowVersion(void) {
 static void ParseArgs(void) {
     TW_CONST byte *s;
     while ((s = *Args)) {
-	if (!strcmp(s, "-h") || !strcmp(s, "-help")) {
+	if (!strcmp(s, "-h") || !Tw_option_strcmp(s, "-help")) {
 	    Usage();
 	    exit(0);
-	} else if (!strcmp(s, "-V") || !strcmp(s, "-version")) {
+	} else if (!strcmp(s, "-V") || !Tw_option_strcmp(s, "-version")) {
 	    ShowVersion();
 	    exit(0);
-	} else if (!strcmp(s, "-k") || !strcmp(s, "-kill")) {
+	} else if (!strcmp(s, "-k") || !Tw_option_strcmp(s, "-kill")) {
 	    DM_Kill = TRUE;
-	} else if (!strcmp(s, "-q") || !strcmp(s, "-quiet")) {
+	} else if (!strcmp(s, "-q") || !Tw_option_strcmp(s, "-quiet")) {
 	    quiet = TRUE;
-	} else if (!strcmp(s, "-attach")) {
+	} else if (!Tw_option_strcmp(s, "-attach")) {
 	    use_twdisplay = DM_ATTACH;
-	} else if (!strcmp(s, "-display")) {
+	} else if (!Tw_option_strcmp(s, "-display")) {
 	    use_twdisplay = DM_DISPLAY;
-	} else if (!strcmp(s, "-suidroot")) {
+	} else if (!Tw_option_strcmp(s, "-envrc")) {
+	    TwEnvRC = "--envrc";
+	} else if (!Tw_option_strcmp(s, "-suidroot")) {
 	    Privileges = TW_PRIV_SUIDROOT;
-	} else if (!strcmp(s, "-sgidtty")) {
+	} else if (!Tw_option_strcmp(s, "-sgidtty")) {
 	    Privileges = TW_PRIV_SGIDTTY;
-	} else if (!strncmp(s, "-title=", 7)) {
+	} else if (!Tw_option_strncmp(s, "-title=", 7)) {
 	    title = s+7;
-	} else if (!strncmp(s, "-hw=", 4)) {
+	} else if (!Tw_option_strncmp(s, "-hw=", 4)) {
 	    hw_name = s;
 	} else if (!strcmp(s, "--")) {
 	    /* '--' means end of options */
 	    break;
 	} else {
 	    fprintf(stderr, "twdm: unknown option: `%s'\n"
-		    "\ttry `twdm -help' for usage summary.\n", s);
+		    "\ttry `twdm --help' for usage summary.\n", s);
 	    exit(1);
 	}
 	Args++;
@@ -176,8 +183,8 @@ static byte InitServer(void) {
 	if (fd[0] != 1)
 	    close(fd[0]);
 
-	execl(BINDIR_PREFIX "twin", "twin", "-secure", "-nohw", NULL);
-	execlp("twin", "twin", "-secure", "-nohw", NULL);
+	execl(BINDIR_PREFIX "twin", "twin", "--secure", "--nohw", TwEnvRC, NULL);
+	execlp("twin", "twin", "--secure", "--nohw", TwEnvRC, NULL);
 	fprintf(stderr, "twdm: exec(twin) failed: %s\n", strerror(errno));
 	exit(1);
 	return FALSE;
@@ -193,6 +200,7 @@ static byte InitServer(void) {
 	do {
 	    i = read(fd[0], buff, 79);
 	} while (i < 0 && errno == EINTR);
+	close(fd[0]);
 	if (i > 33 && !memcmp(buff, "twin: starting in background as :", 33)) {
 	    while (buff[--i] == '\n')
 		;
@@ -207,8 +215,12 @@ static byte InitServer(void) {
 	    }
 	    return TRUE;
 	}
-
-	fprintf(stderr, "twdm: read() from twin failed: %s\n", strerror(errno));
+	
+	if (i <= 0)
+	    fprintf(stderr, "twdm: read() from twin failed: %s\n", strerror(errno));
+	else
+	    fprintf(stderr, "twdm: error starting twin:\n\t%.*s\n", i, buff);
+	    
 	return FALSE;
     }
     fprintf(stderr, "twdm: pipe() failed: %s\n", strerror(errno));
@@ -221,18 +233,21 @@ static void shortsleep(void) {
 }
 
 static byte InitAttach(void) {
-    char *attach, *prefix_attach;
-    char buff[] = "-twin@:\0\0\0";
+    char *attach, *path_attach;
+    char buff[] = "--twin@:\0\0\0";
 
-    if (use_twdisplay == DM_ATTACH || (use_twdisplay == 0 && !strcmp(hw_name, "-hw=tty"))) {
+    if (use_twdisplay == DM_ATTACH ||
+	(use_twdisplay == 0 &&
+	 !Tw_option_strcmp(hw_name, "-hw=tty"))) {
+	
         attach = "twattach";
-	prefix_attach = BINDIR_PREFIX "twattach";
+	path_attach = BINDIR_PREFIX "twattach";
     } else {
         attach = "twdisplay";
-	prefix_attach = BINDIR_PREFIX "twdisplay";
+	path_attach = BINDIR_PREFIX "twdisplay";
     }
 
-    strncpy(buff+6, DM_Display, 4);
+    strncpy(buff+7, DM_Display, 4);
 
     /* sleep a little to let server start up */
     shortsleep();
@@ -240,8 +255,8 @@ static byte InitAttach(void) {
     switch ((AttachPid = fork())) {
       case 0:
 	/* child */
-	execl(prefix_attach, attach, "-q", buff, hw_name, NULL);
-	execlp(attach, attach, "-q", buff, hw_name, NULL);
+	execl(path_attach, attach, "--quiet", buff, hw_name, NULL);
+	execlp(attach, attach, "--quiet", buff, hw_name, NULL);
 	fprintf(stderr, "twdm: exec(%s) failed: %s\n", attach, strerror(errno));
 	exit(1);
 	return FALSE;
@@ -284,7 +299,7 @@ static byte InitClient(void) {
     }
     
     if ((DM_MsgPort=TwCreateMsgPort
-	 (2, "DM", (time_t)0, (frac_t)0, (byte)0)) &&
+	 (2, "DM")) &&
 	(DM_Menu=TwCreateMenu
 	 (COL(BLACK,WHITE), COL(BLACK,GREEN), COL(HIGH|BLACK,WHITE), COL(HIGH|BLACK,BLACK),
 	  COL(RED,WHITE), COL(RED,GREEN), (byte)0)) &&
@@ -321,7 +336,7 @@ static byte InitClient(void) {
 
 	TwSetColTextWindow(DM_Window, COL(HIGH|RED,BLUE));
 	TwGotoXYWindow(DM_Window, 7, 8);
-	TwWriteRowWindow(DM_Window, 26, "L O G I N   F A I L E D  !");
+	TwWriteAsciiWindow(DM_Window, 26, "L O G I N   F A I L E D  !");
 			 
 	TwSetXYWindow(DM_Window, (X-40)/2, (Y-15)/2);
 	TwSetXYWindow(DM_user, 4, 2);
@@ -341,13 +356,13 @@ static byte InitClient(void) {
 static void ClearKey(void) {
     TwGotoXYWindow(DM_user, 0, 0);
     TwWriteMem(user.txt, ' ', user.len);
-    TwWriteRowWindow(DM_user, user.len, user.txt);
+    TwWriteAsciiWindow(DM_user, user.len, user.txt);
     TwGotoXYWindow(DM_user, user.len = user.x = 0, 0);
     user.txt[0] = '\0';
     
     TwGotoXYWindow(DM_pass, 0, 0);
     TwWriteMem(pass.txt, ' ', pass.len);
-    TwWriteRowWindow(DM_pass, pass.len, pass.txt);
+    TwWriteAsciiWindow(DM_pass, pass.len, pass.txt);
     TwGotoXYWindow(DM_pass, pass.len = pass.x = 0, 0);    
     pass.txt[0] = '\0';
 }
@@ -432,10 +447,10 @@ static void DelKey(twindow W, data u) {
 	
 	if (W == DM_user) {
 	    TwGotoXYWindow(W, 0, 0);
-	    TwWriteRowWindow(W, u->len, u->txt);
+	    TwWriteAsciiWindow(W, u->len, u->txt);
 	} else
 	    TwGotoXYWindow(W, u->len, 0);
-	TwWriteRowWindow(W, 1, " ");
+	TwWriteAsciiWindow(W, 1, " ");
 	TwGotoXYWindow(W, u->x, 0);	
     }
 }
@@ -486,10 +501,22 @@ static void WriteKey(twindow W, data u, udat len, byte *seq) {
 	u->len = _len;
 	if (W == DM_pass)
 	    TwWriteMem(seq, '*', len);
-	TwWriteRowWindow(W, len, seq);
+	TwWriteAsciiWindow(W, len, seq);
     }
 }
 
+#ifdef CONF__UNICODE
+static void WriteHWFontKey(twindow W, data u, udat len, hwfont *h_data) {
+    byte *d_data = (byte *)h_data, *s_data = d_data;
+    udat n = len;
+
+    /* hack warning: this assumes `h_data' is writable and correctly aligned */
+    while (n--)
+	*d_data++ = Tutf_UTF_16_to_CP437(*h_data++);
+    WriteKey(W, u, len, s_data);
+}
+#endif
+
 static void HandleKey(tevent_keyboard E) {
     data u;
     
@@ -552,16 +579,10 @@ static void HandleGadget(tevent_gadget E) {
     }
 }
 
-#if TW_RETSIGTYPE == void
-# define TW_RETFROMSIGNAL
-#else
-# define TW_RETFROMSIGNAL return 0;
-#endif
-
 static TW_RETSIGTYPE SignalPanic(int n) {
     signal(n, SIG_DFL);
     quit();
-    TW_RETFROMSIGNAL
+    TW_RETFROMSIGNAL(0);
 }
 
 static TW_RETSIGTYPE SignalChild(int n) {
@@ -576,7 +597,7 @@ static TW_RETSIGTYPE SignalChild(int n) {
 		AttachPid = (pid_t)-1;
 	}
     }
-    TW_RETFROMSIGNAL
+    TW_RETFROMSIGNAL(0);
 }
 
 static void InitSignals(void) {
@@ -608,7 +629,6 @@ int main(int argc, char *argv[]) {
     tmsg Msg;
     uldat err;
 
-    
     Args = (byte * TW_CONST *)argv + 1;
     
     ParseArgs();
@@ -654,10 +674,15 @@ int main(int argc, char *argv[]) {
 		{
 		    tevent_selectionnotify E = &Msg->Event.EventSelectionNotify;
 
+		    /* react as for keypresses */
 		    if (E->Magic == TW_SEL_TEXTMAGIC)
-			/* react as for keypresses */
 			WriteKey(E->ReqPrivate, E->ReqPrivate == DM_user ? &user : &pass,
 				 E->Len, E->Data);
+#ifdef CONF__UNICODE
+		    else if (E->Magic == TW_SEL_HWFONTMAGIC)
+			WriteHWFontKey(E->ReqPrivate, E->ReqPrivate == DM_user ? &user : &pass,
+				       E->Len / sizeof(hwfont), (hwfont *)E->Data);
+#endif
 		}
 		break;
 	      default:
