@@ -17,15 +17,22 @@
 # include "autoconf.h"
 #endif
 
+/* pull in "obj" field in <Tw/stattypes.h> */
+#define obj obj
+typedef struct s_obj *obj;
+
 #include "compiler.h"
 #include "version.h"
 #include "osincludes.h"
-#include "Tw/compiler.h"
-#include "Tw/datatypes.h"
-#include "Tw/datasizes.h"
-#include "Tw/uni_types.h"
-#include "Tw/version.h"
-#include "Tw/missing.h"
+#include <Tw/compiler.h>
+#include <Tw/datatypes.h>
+#include <Tw/datasizes.h>
+#include <Tw/stattypes.h>
+#include <Tw/endianity.h>
+#include <Tw/uni_types.h>
+#include <Tw/version.h>
+#include <Tw/missing.h>
+#include <Tw/mouse.h>
 
 
 #define Abs(x) ((x)>0 ? (x) : -(x))
@@ -69,6 +76,20 @@
 #define HUGEBUFF	131072
 
 
+
+/* return from signal macros */
+
+#define RETFROMSIGNAL(value)		return ARG_IFNOTVOID(value, RETSIGTYPE)
+#define ARG_IFNOTVOID(value, type)	CAT(ARG_IFNOTVOID_,type)(value)
+#define ARG_IFNOTVOID_void(value)
+#define ARG_IFNOTVOID_int(value)	(value)
+
+
+
+/* maximum number of arguments of a libTw function */
+#define TW_MAX_ARGS_N	20
+
+
 /* Macros for HW VGA (not ANSI!) colors */
 #define BLACK	((hwcol)0)
 #define BLUE	((hwcol)1)
@@ -81,6 +102,8 @@
 #define HIGH	((hwcol)8)
 #define MAXCOL	((hwcol)0xF)
 
+#define MAXHWCOL	COL(MAXCOL,MAXCOL)
+
 #define ANSI2VGA(col) (((col) & 0x1 ? RED   : 0) | \
 		       ((col) & 0x2 ? GREEN : 0) | \
 		       ((col) & 0x4 ? BLUE  : 0))
@@ -110,7 +133,7 @@
 # define HWFONT16(attr) ((byte)(attr))
 
 
-/* if sizeof(hwattr) == 4, bytes are { 'ascii_low', 'col', 'ascii_high', 'unused' } */
+/* if sizeof(hwattr) == 4, bytes are { 'ascii_low', 'col', 'ascii_high', 'extra' } */
 
 /* hwattr <-> hwcol+hwfont conversion */
 # define HWATTR32(col,ascii) (((byte32)(byte)(col) << 8) | (((byte32)(ascii) & 0xFF00) << 8) | (byte32)(byte)(ascii))
@@ -119,37 +142,40 @@
 # define HWCOL32(attr) ((hwcol)((attr) >> 8))
 # define HWFONT32(attr) ((byte16)(((attr) & 0xFF) | (((attr) >> 8) & 0xFF00)))
 
+# define HWATTR_EXTRA32(attr,extra) (((byte32)(byte)(extra) << 24) | ((byte32)(attr) & 0xFFFFFF))
+# define HWATTR_EXTRAMASK32(attr) ((attr) & 0xFF000000)
+# define HWEXTRA32(attr) ((byte)((attr) >> 24))
 
 /*
  * Notes about the timevalue struct:
  * 
  * it is used to represent both time intervals and absolute times;
- * the ->Seconds is a time_t numeric field.
+ * the ->Seconds is a tany numeric field.
  * DON'T assume time_t is 32 bit (or any other arbitrary size)
- * since in 19 Jan 2038 at 04:14:08 any signed, 32 bit time_t will overflow.
- * So use sizeof(time_t) if you really need.
+ * since in 19 Jan 2038 at 04:14:08 any signed, 32 bit tany will overflow.
+ * So use sizeof(tany) if you really need.
  * 
- * the ->Fraction is a frac_t numeric field (frac_t is signed).
- * As above, DON'T assume frac_t is 32 bit (or any other arbitrary size)
+ * the ->Fraction is a tany numeric field (tany is unsigned).
+ * As above, DON'T assume tany is 32 bit (or any other arbitrary size)
  * since in the future we may want a finer granularity than the nanosecond one
- * possible with a 32 bit frac_t.
+ * possible with a 32 bit tany.
  * So :
- * 1) use sizeof(frac_t) if you really need
- * 2) don't assume (frac_t)1 is a nanosecond (or any other arbitrary time),
+ * 1) use sizeof(tany) if you really need
+ * 2) don't assume (tany)1 is a nanosecond (or any other arbitrary time),
  *    but always use the form '1 NanoSECs', '250 MilliSECs + 7 MicroSECs', etc.
- * 3) if you _absolutely_ need to know to what time (frac_t)1 corresponds,
- *    use this: '1 FullSECs' is the number of (frac_t)1 intervals in a second.
+ * 3) if you _absolutely_ need to know to what time (tany)1 corresponds,
+ *    use this: '1 FullSECs' is the number of (tany)1 intervals in a second.
  * 4) for the moment, the only defined fractions of a second are:
  *    FullSECs, MilliSECs, MicroSECs, NanoSECs.
  *    Others may be added in the future (PicoSECs, FemtoSECs, AttoSECs, ...)
  */
 
 typedef struct s_timevalue  {
-    time_t Seconds;
-    frac_t Fraction;
+    tany Seconds;
+    tany Fraction;
 } timevalue;
 
-#define THOUSAND	((frac_t)1000)
+#define THOUSAND	((tany)1000)
 
 #define NanoSECs	* 1 /* i.e. (frac_t)1 is a nanosecond */
 #define MicroSECs	* (THOUSAND NanoSECs)
@@ -179,8 +205,7 @@ typedef struct s_ttydata ttydata;
 typedef struct s_remotedata remotedata;
 typedef struct s_draw_ctx draw_ctx;
 
-#define obj obj
-typedef struct s_obj *obj;
+/*typedef struct s_obj *obj;*/
 typedef struct s_fn_obj *fn_obj;
 typedef struct s_obj_parent *obj_parent;
 typedef struct s_widget *widget;
@@ -207,6 +232,8 @@ typedef struct s_mutex *mutex;
 typedef struct s_fn_mutex *fn_mutex;
 typedef struct s_module *module;
 typedef struct s_fn_module *fn_module;
+typedef struct s_extension *extension;
+typedef struct s_fn_extension *fn_extension;
 typedef struct s_display_hw *display_hw;
 typedef struct s_fn_display_hw *fn_display_hw;
 
@@ -276,6 +303,7 @@ struct s_ttydata {
     byte currG, G, G0, G1, saveG, saveG0, saveG1;
     byte utf, utf_count;
     hwfont utf_char;
+    void *InvCharset;	/* pointer to hwfont -> byte translation function */
     
     dat newLen, newMax;
     byte *newName;	/* buffer for xterm set window title escape seq */
@@ -310,8 +338,7 @@ struct s_draw_ctx {
 struct s_obj {
     uldat Id;
     fn_obj Fn;
-    obj Prev, Next;
-    obj_parent Parent;
+    obj Prev, Next, Parent;
 };
 
 struct s_obj_parent {
@@ -321,7 +348,7 @@ struct s_obj_parent {
 struct s_fn_obj {
     uldat Magic, Size, Used;
     obj (*Create)(fn_obj);
-    void (*Insert)(obj Obj, obj_parent, obj Prev, obj Next);
+    void (*Insert)(obj Obj, obj, obj Prev, obj Next);
     void (*Remove)(obj);
     void (*Delete)(obj);
     void (*ChangeField)(obj, udat field, uldat CLEARMask, uldat XORMask);
@@ -386,6 +413,8 @@ struct s_fn_widget {
     void (*Map)(widget, widget Parent);
     void (*UnMap)(widget);
     void (*MapTopReal)(widget, screen);
+    void (*Raise)(widget);
+    void (*Lower)(widget);
     void (*Own)(widget, msgport);
     void (*DisOwn)(widget);    
     void (*RecursiveDelete)(widget, msgport);
@@ -403,6 +432,7 @@ struct s_fn_widget {
 #define WIDGET_WANT_KEYS	0x0002
 #define WIDGET_WANT_MOUSE	0x0004
 #define WIDGET_WANT_CHANGES	0x0008
+#define WIDGET_AUTO_FOCUS	0x0010
 
 
 /* Widget->Flags */
@@ -410,6 +440,8 @@ struct s_fn_widget {
 #define WIDGETFL_USEFILL	0x03
 #define WIDGETFL_USEANY		0x07 /* mask of all above ones */
 
+#define WIDGETFL_NOTVISIBLE	0x8000
+
 #define w_USE(w, USExxx)	(((w)->Flags & WIDGETFL_USEANY) == CAT(WIDGETFL_,USExxx))
 
 
@@ -471,6 +503,8 @@ struct s_fn_gadget {
     void (*Map)(gadget, widget Parent);
     void (*UnMap)(gadget);
     void (*MapTopReal)(gadget, screen);
+    void (*Raise)(gadget);
+    void (*Lower)(gadget);
     void (*Own)(gadget, msgport);
     void (*DisOwn)(gadget);    
     void (*RecursiveDelete)(gadget, msgport);
@@ -494,6 +528,7 @@ struct s_fn_gadget {
 #define GADGET_WANT_KEYS	WIDGET_WANT_KEYS	/* 0x0002 */
 #define GADGET_WANT_MOUSE	WIDGET_WANT_MOUSE	/* 0x0004 */
 #define GADGET_WANT_CHANGES	WIDGET_WANT_CHANGES	/* 0x0008 */
+#define GADGET_AUTO_FOCUS	WIDGET_AUTO_FOCUS	/* 0x0010 */
 
 
 /* Gadget->Flags */
@@ -513,6 +548,8 @@ struct s_fn_gadget {
 #define GADGETFL_TOGGLE		0x0080
 #define GADGETFL_PRESSED	0x0100
 
+#define GADGETFL_NOTVISIBLE	0x8000
+
 
 /*              NOTE :
  * 
@@ -572,7 +609,8 @@ struct s_window {
 	struct s_wE E;
     } USE;
     /* window */
-    menu Menu;		/* from which the window depends */
+    menu Menu;
+    menuitem MenuItem; /* from which the window depends */
     dat NameLen;
     byte *Name; hwcol *ColName;
     hwfont *BorderPattern[2];
@@ -591,8 +629,8 @@ struct s_window {
 
 struct s_fn_window {
     uldat Magic, Size, Used;
-    window (*Create)(fn_window, dat NameLen, CONST byte *Name, CONST hwcol *ColName, menu Menu,
-		     hwcol ColText, uldat CursorType, uldat Attrib, uldat Flags,
+    window (*Create)(fn_window, msgport Owner, dat NameLen, CONST byte *Name, CONST hwcol *ColName,
+		     menu Menu, hwcol ColText, uldat CursorType, uldat Attrib, uldat Flags,
 		     dat XWidth, dat YWidth, dat ScrollBackLines);
     void (*Insert)(window, widget Parent, widget Prev, widget Next);
     void (*Remove)(window);
@@ -610,6 +648,8 @@ struct s_fn_window {
     void (*Map)(window, widget Parent);
     void (*UnMap)(window);
     void (*MapTopReal)(window, screen);
+    void (*Raise)(window);
+    void (*Lower)(window);
     void (*Own)(window, msgport);
     void (*DisOwn)(window);    
     void (*RecursiveDelete)(window, msgport);
@@ -618,13 +658,18 @@ struct s_fn_window {
     void (*RemoveHook)(window, fn_hook, fn_hook *Where);
     /* window */
     fn_widget Fn_Widget;
-    void (*WriteAscii)(window, ldat Len, CONST byte *Ascii);
-    void (*WriteString)(window, ldat Len, CONST byte *String);
-    void (*WriteHWFont)(window, ldat Len, CONST hwfont *HWFont);
-    void (*WriteHWAttr)(window, dat x, dat y, ldat Len, CONST hwattr *Attr);
-    byte (*WriteRow)(window, ldat Len, CONST byte *Text);
-    
+    void (*TtyWriteAscii)(window, ldat Len, CONST byte *Ascii);
+    void (*TtyWriteString)(window, ldat Len, CONST byte *String);
+    void (*TtyWriteHWFont)(window, ldat Len, CONST hwfont *HWFont);
+    void (*TtyWriteHWAttr)(window, dat x, dat y, ldat Len, CONST hwattr *Attr);
+
+    byte (*RowWriteAscii)(window, ldat Len, CONST byte *Ascii);
+    byte (*RowWriteString)(window, ldat Len, CONST byte *String);
+    byte (*RowWriteHWFont)(window, ldat Len, CONST hwfont *HWFont);
+    byte (*RowWriteHWAttr)(window, dat x, dat y, ldat Len, CONST hwattr *Attr);
+
     void (*GotoXY)(window, ldat X, ldat Y);
+    void (*SetTitle)(window, dat titlelen, byte *title);
     void (*SetColText)(window, hwcol ColText);
     void (*SetColors)(window, udat Bitmap,
 		      hwcol ColGadgets, hwcol ColArrows, hwcol ColBars, hwcol ColTabs, hwcol ColBorder,
@@ -632,7 +677,7 @@ struct s_fn_window {
     void (*Configure)(window, byte Bitmap, dat Left, dat Up, dat MinXWidth, dat MinYWidth,
 		      dat MaxXWidth, dat MaxYWidth);
     window (*Create4Menu)(fn_window, menu);
-    byte (*FindBorder)(window, dat u, dat v, byte Border, hwfont *PtrChar, hwcol *PtrColor);
+    byte (*FindBorder)(window, dat u, dat v, byte Border, hwattr *PtrAttr);
     row (*FindRow)(window, ldat RowN);
     row (*FindRowByCode)(window, udat Code, ldat *NumRow);
 };
@@ -643,6 +688,7 @@ struct s_fn_window {
 #define WINDOW_WANT_KEYS	WIDGET_WANT_KEYS	/* 0x0002 */
 #define WINDOW_WANT_MOUSE	WIDGET_WANT_MOUSE	/* 0x0004 */
 #define WINDOW_WANT_CHANGES	WIDGET_WANT_CHANGES	/* 0x0008 */
+#define WINDOW_AUTO_FOCUS	WIDGET_AUTO_FOCUS	/* 0x0010 */
 #define WINDOW_DRAG		0x0100
 #define WINDOW_RESIZE		0x0200
 #define WINDOW_CLOSE		0x0400
@@ -669,6 +715,7 @@ struct s_fn_window {
 #define WINDOWFL_ROWS_INSERT	0x0100
 #define WINDOWFL_ROWS_DEFCOL	0x0200
 #define WINDOWFL_ROWS_SELCURRENT	0x0400
+#define WINDOWFL_NOTVISIBLE	0x8000
 
 
 /* Window->State */
@@ -707,7 +754,7 @@ struct s_fn_window {
 #define SOLIDCURSOR	8
 
 /* window size limits */
-#define MIN_XWIN	4
+#define MIN_XWIN	5
 #define MIN_YWIN	2
 
 struct s_sB {	/* for SCREENFL_USEBG screens */
@@ -765,6 +812,9 @@ struct s_fn_screen {
     widget (*KbdFocus)(screen);
     void (*Map)(screen, widget Parent);
     void (*UnMap)(screen);
+    void (*MapTopReal)(screen, screen);
+    void (*Raise)(screen);
+    void (*Lower)(screen);
     void (*Own)(screen, msgport);
     void (*DisOwn)(screen);    
     void (*RecursiveDelete)(screen, msgport);
@@ -778,7 +828,7 @@ struct s_fn_screen {
     screen (*CreateSimple)(fn_screen, dat NameLen, CONST byte *Name, hwattr Bg);
     void (*BgImage)(screen, dat BgWidth, dat BgHeight, CONST hwattr *Bg);
     void (*DrawMenu)(screen, dat Xstart, dat Xend);
-    void (*ActivateMenu)(screen, menuitem, byte byMouse);
+    void (*ActivateMenu)(screen, menuitem, byte ByMouse);
     void (*DeActivateMenu)(screen);
 };
 
@@ -797,7 +847,7 @@ struct s_fn_screen {
 
 #define SCREENFL_BACK_SELECT	0x0010
 #define SCREENFL_BACK_PRESSED	0x0020
-
+#define SCREENFL_NOTVISIBLE	0x8000
 
 
 /* group -- gadget group */
@@ -857,7 +907,9 @@ struct s_fn_row {
     /* row */
     fn_obj Fn_Obj;
     byte (*SetText)(row, ldat Len, CONST byte *Text, byte DefaultCol);
-/*    byte (*SetHWFont)(row, ldat Len, CONST hwfont *HWFont, byte DefaultCol); */
+    byte (*SetHWFont)(row, ldat Len, CONST hwfont *HWFont, byte DefaultCol);
+    void (*Raise)(row);
+    void (*Lower)(row);
 };
 
 /*Flags : */
@@ -884,6 +936,7 @@ struct s_menuitem {
     /* menuitem */
     window Window;
     dat Left, ShortCut;
+    ldat WCurY; 
 };
 struct s_fn_menuitem {
     uldat Magic, Size, Used;
@@ -896,8 +949,10 @@ struct s_fn_menuitem {
     /* row */
     fn_obj Fn_Obj;
     byte (*SetText)(row, ldat Len, CONST byte *Text, byte DefaultCol);
-/*    byte (*SetHWFont)(row, ldat Len, CONST hwfont *HWFont, byte DefaultCol); */
+    byte (*SetHWFont)(row, ldat Len, CONST hwfont *HWFont, byte DefaultCol);
     /* menuitem */	
+    void (*Raise)(menuitem);
+    void (*Lower)(menuitem);
     fn_row Fn_Row;
     menuitem (*Create4Menu)(fn_menuitem, obj Parent, window Window, udat Code,
 			    byte Flags, ldat Len, CONST byte *Name);
@@ -930,8 +985,9 @@ struct s_fn_menu {
     fn_obj Fn_Obj;
     row (*SetInfo)(menu, byte Flags, ldat Len, CONST byte *Text, CONST hwcol *ColText);
     menuitem (*FindItem)(menu, dat i);
-    menuitem (*GetSelectItem)(menu);
-    void (*SetSelectItem)(menu, menuitem);
+    menuitem (*GetSelectedItem)(menu);
+    menuitem (*RecursiveGetSelectedItem)(menu, dat *depth);
+    void (*SetSelectedItem)(menu, menuitem);
 };
 
 
@@ -956,6 +1012,7 @@ struct s_fn_menu {
 
 #define MSG_USER_FIRST		((udat)0x2000)
 #define MSG_USER_CONTROL	((udat)0x2000)
+#define MSG_USER_CONTROL_REPLY	((udat)0x2001)
 #define MSG_USER_CLIENTMSG	((udat)0x2100)
 
 /*
@@ -1023,7 +1080,7 @@ struct s_event_control {
     widget W;
     udat Code, Len;
     dat X, Y;
-    byte Data[1]; /* Data[Len] == '\0' */
+    byte Data[sizeof(uldat)]; /* [Len] bytes actually */
 };
 
 /* some MSG_CONTROL codes */
@@ -1036,8 +1093,13 @@ struct s_event_control {
 typedef struct s_event_clientmsg event_clientmsg;
 struct s_event_clientmsg {
     widget W;
-    udat Code, Len;
-    byte Data[1]; /* [Len] bytes actually */
+    udat Code, Format;
+    uldat Len;
+    union {
+	byte b[sizeof(uldat)];
+	udat d[sizeof(uldat)/sizeof(udat)];
+	uldat l[1];
+    } Data; /* [Len] bytes actually */
 };
 
 typedef struct s_event_display event_display;
@@ -1092,6 +1154,7 @@ struct s_event_menu {
     window W;
     udat Code, pad;
     menu Menu;
+    row Row;
 };
 
 typedef struct s_event_selection event_selection;
@@ -1111,12 +1174,12 @@ struct s_event_selectionnotify {
     uldat Magic;
     byte MIME[MAX_MIMELEN];
     uldat Len;
-    byte Data[1]; /* Data[] is Len bytes actually */
+    byte Data[sizeof(uldat)]; /* Data[] is Len bytes actually */
 };
 /*SelectionNotify Magic*/
 #define SEL_APPEND	0x00000000
 #define SEL_TEXTMAGIC	0x54657874
-#define SEL_HWFONTMAGIC 0x4877666E
+#define SEL_HWFONTMAGIC 0x4877666E /* it's unicode */
 #define SEL_FILEMAGIC	0x46696c65
 #define SEL_URLMAGIC	0xAB1691BA
 #define SEL_DATAMAGIC	0xDA1AA1AD /* check MIME if you get this */
@@ -1187,12 +1250,14 @@ struct s_msgport {
     widget FirstW, LastW;	/* widgets owned by this MsgPort */
     group FirstGroup, LastGroup; /* groups done by this MsgPort */
     mutex FirstMutex, LastMutex;/* mutexes owned by this MsgPort */
+    uldat CountE, SizeE;        /* number of extensions used by this MsgPort */
+    extension *Es;              /* extensions used by this MsgPort */
     display_hw AttachHW;	/* that was attached as told by MsgPort */
 };
 struct s_fn_msgport {
     uldat Magic, Size, Used;
     msgport (*Create)(fn_msgport, byte NameLen, CONST byte *Name,
-		      time_t PauseSec, frac_t PauseFraction,
+		      tany PauseSec, tany PauseFraction,
 		      byte WakeUp, void (*Handler)(msgport));
     void (*Insert)(msgport, all, msgport Prev, msgport Next);
     void (*Remove)(msgport);
@@ -1200,6 +1265,8 @@ struct s_fn_msgport {
     void (*ChangeField)(msgport, udat field, uldat CLEARMask, uldat XORMask);
     /* msgport */
     fn_obj Fn_Obj;
+    void (*UseExtension)(msgport, extension);
+    void (*UnuseExtension)(msgport, extension);
 };
 /* MsgPort->WakeUp: */
 #define TIMER_ALWAYS	((byte)1)
@@ -1261,6 +1328,37 @@ struct s_fn_module {
 
 
 
+struct s_extension {
+    uldat Id;
+    fn_extension Fn;
+    extension Prev, Next; /* in the same All */
+    all All;
+    /* module */
+    uldat NameLen, Used;
+    byte *Name;
+    void *Handle, *Private;
+    /* extension */
+    tany (*CallB)(extension, topaque len, CONST byte *data, void *return_type); /* call extension-specific functions */
+    void (*Quit)(extension); /* how to quit this extension if it is not dlopen()ed */
+};
+struct s_fn_extension {
+    uldat Magic, Size, Used;
+    extension (*Create)(fn_extension, uldat NameLen, CONST byte *Name);
+    void (*Insert)(extension, all, extension Prev, extension Next);
+    void (*Remove)(extension);
+    void (*Delete)(extension);
+    void (*ChangeField)(extension, udat field, uldat CLEARMask, uldat XORMask);
+    /* module */    
+    fn_obj Fn_Obj;
+    byte (*DlOpen)(extension);
+    void (*DlClose)(extension);
+    /* extension */
+    fn_module Fn_Module;
+    extension (*Query)(byte namelen, CONST byte *name);
+};
+
+
+
 struct s_display_hw {
     uldat Id;
     fn_display_hw Fn;
@@ -1298,7 +1396,7 @@ struct s_display_hw {
     void (*HWSelectionRequest)(obj Requestor, uldat ReqPrivate);
     void (*HWSelectionNotify)(uldat ReqPrivate, uldat Magic,
 			      CONST byte MIME[MAX_MIMELEN], uldat Len, CONST byte *Data);
-    void *HWSelectionPrivate;
+    tany HWSelectionPrivate;
 	
     byte (*CanDragArea)(dat Xstart, dat Ystart, dat Xend, dat Yend, dat DstXstart, dat DstYstart);
     void (*DragArea)(dat Xstart, dat Ystart, dat Xend, dat Yend, dat DstXstart, dat DstYstart);
@@ -1312,6 +1410,8 @@ struct s_display_hw {
 
     void (*Beep)(void);
     void (*Configure)(udat resource, byte todefault, udat value);
+    void (*ConfigureKeyboard)(udat resource, byte todefault, udat value);
+    void (*ConfigureMouse)(udat resource, byte todefault, udat value);
     void (*SetPalette)(udat N, udat R, udat G, udat B);
     void (*ResetPalette)(void);
 
@@ -1448,7 +1548,7 @@ struct s_fn_display_hw {
 
 /* IDs */
 #define NOID		((uldat)0)
-#define ERRID		((uldat)-1)
+#define BADID		((uldat)-1)
 
 #define MAXID		((uldat)0x0FFFFFFFul)
 #define magic_mask	((uldat)0xF0000000ul)
@@ -1467,7 +1567,9 @@ struct s_fn_display_hw {
 #define msg_magic_id		0xA
 #define mutex_magic_id		0xB
 #define module_magic_id		0xC
-#define display_hw_magic_id	0xD
+#define extension_magic_id	0xD
+#define display_hw_magic_id	0xE
+#define all_magic_id		0xF
 
 /*
  * These must have consecutive values, but obj_magic_STR can be changed
@@ -1475,21 +1577,23 @@ struct s_fn_display_hw {
  * To avoid troubles with strlen(), you should not use '\0' or "\0"
  * for any of the values below.
  */
-#define base_magic_CHR		'0'
-#define obj_magic_STR		"0"
-#define widget_magic_STR	"1"
-#define gadget_magic_STR	"2"
-#define window_magic_STR	"3"
-#define screen_magic_STR	"4"
-#define group_magic_STR		"5"
-#define row_magic_STR		"6"
-#define menuitem_magic_STR	"7"
-#define menu_magic_STR		"8"
-#define msgport_magic_STR	"9"
-#define msg_magic_STR		"A"
-#define mutex_magic_STR		"B"
-#define module_magic_STR	"C"
-#define display_hw_magic_STR	"D"
+#define base_magic_CHR		'\x30'
+#define obj_magic_STR		"\x30"
+#define widget_magic_STR	"\x31"
+#define gadget_magic_STR	"\x32"
+#define window_magic_STR	"\x33"
+#define screen_magic_STR	"\x34"
+#define group_magic_STR		"\x35"
+#define row_magic_STR		"\x36"
+#define menuitem_magic_STR	"\x37"
+#define menu_magic_STR		"\x38"
+#define msgport_magic_STR	"\x39"
+#define msg_magic_STR		"\x3A"
+#define mutex_magic_STR		"\x3B"
+#define module_magic_STR	"\x3C"
+#define extension_magic_STR	"\x3D"
+#define display_hw_magic_STR	"\x3E"
+#define all_magic_STR		"\x3F"
 
 #define obj_magic	((uldat)0x0dead0b1ul)
 #define widget_magic	((uldat)0x161d9743ul)
@@ -1504,11 +1608,12 @@ struct s_fn_display_hw {
 #define msg_magic	((uldat)0xA3a61ce4ul) /* this gets compiled in libTw ! */
 #define mutex_magic	((uldat)0xB0faded0ul)
 #define module_magic	((uldat)0xCb0f1278ul)
-#define display_hw_magic ((uldat)0xDdbcc609ul)
+#define extension_magic ((uldat)0xDe81ec51ul)
+#define display_hw_magic ((uldat)0xEdbcc609ul)
+#define all_magic	((uldat)0xFa11Fa11ul)
 
 
-#define magic_n		14 /* max top hex digit of the above ones + 1 */
-
+#define magic_n		16 /* max top hex digit of the above ones + 1 */
 
 /*
  *   B I G   F A T   WARNING:
@@ -1530,8 +1635,10 @@ struct s_fn_display_hw {
 #define IS_MSGPORT(O)	IS_OBJ(msgport,O)
 #define IS_MUTEX(O)	IS_OBJ(mutex,O)
 #define IS_MSG(O)	IS_OBJ(msg,O)
-#define IS_MODULE(O)	IS_OBJ(module,O)
+#define IS_MODULE(O)	(IS_OBJ(module,O) || IS_OBJ(extension,O))
+#define IS_EXTENSION(O)	IS_OBJ(extension,O)
 #define IS_DISPLAY_HW(O) IS_OBJ(display_hw,O)
+#define IS_ALL(O)	IS_OBJ(all,O)
 
 /* in the same order as the #defines above ! */
 struct s_fn {
@@ -1548,23 +1655,26 @@ struct s_fn {
     fn_mutex f_mutex;
     fn_msg f_msg;
     fn_module f_module;
+    fn_extension f_extension;
     fn_display_hw f_display_hw;
+    fn_obj f_all;
 };
 
 struct s_setup {
     dat MaxMouseSnap;
     udat MinAllocSize;
     byte Flags;
-    byte SelectionButton, PasteButton;
+    byte ButtonSelection, ButtonPaste;
     byte DeltaXShade, DeltaYShade;
 };
 /* All->Setup->Flags */
 #define SETUP_SHADOWS		0x01
 #define SETUP_BLINK		0x02
-#define SETUP_ALWAYSCURSOR	0x04
-#define SETUP_HIDEMENU		0x08
-#define SETUP_MENUINFO		0x10
-#define SETUP_EDGESCROLL	0x20
+#define SETUP_CURSOR_ALWAYS	0x04
+#define SETUP_MENU_HIDE		0x08
+#define SETUP_MENU_INFO		0x10
+#define SETUP_MENU_RELAX	0x20
+#define SETUP_SCREEN_SCROLL	0x40
 
 
 #define MAX_XSHADE	9
@@ -1578,18 +1688,18 @@ struct s_setup {
  * do not change it!
  */
 /*#define STATE_BUTTON(n)	((byte)(n)) */
-#define STATE_RESIZE		((byte)10)
-#define STATE_DRAG		((byte)11)
-#define STATE_SCROLL		((byte)12)
-#define STATE_GADGET		((byte)13)
-#define STATE_MENU		((byte)14)
-#define STATE_SCREEN		((byte)15)
-#define STATE_SCREENBUTTON	((byte)16)
-#define STATE_ROOT		((byte)17)
-#define STATE_DEFAULT		((byte)18)
+#define STATE_DRAG		((byte)10)
+#define STATE_RESIZE		((byte)15)
+#define STATE_SCROLL		((byte)16)
+#define STATE_GADGET		((byte)26)
+#define STATE_MENU		((byte)27)
+#define STATE_SCREEN		((byte)28)
+#define STATE_BUTTON_SCREEN	((byte)29)
+#define STATE_ROOT		((byte)30)
+#define STATE_DEFAULT		((byte)31)
 
 /* mask for all the above */
-#define STATE_ANY		((byte)0x1f)
+#define STATE_ANY		((byte)0x1F)
 
 /* furher All->State flags */
 #define STATE_FL_BYMOUSE	((byte)0x40)
@@ -1598,19 +1708,28 @@ struct s_setup {
 /* values returned by FnWindow->FindBorder (modeled after STATE_*) */
 /*#define POS_BUTTON(n)		((byte)(n))*/
 #define POS_TITLE		((byte)10)
-#define POS_CORNER		((byte)11)
-#define POS_SIDE		((byte)12)
+#define POS_SIDE_LEFT		((byte)11)
+#define POS_SIDE_UP		((byte)12)
+#define POS_SIDE_RIGHT		((byte)13)
+#define POS_SIDE_DOWN		((byte)14)
+#define POS_BUTTON_RESIZE	((byte)15)
 
-#define POS_BAR_BACK		((byte)13)
-#define POS_BAR_FWD		((byte)14)
-#define POS_TAB			((byte)15)
-#define POS_ARROW_BACK		((byte)16)
-#define POS_ARROW_FWD		((byte)17)
+#define POS_X_BAR_BACK		((byte)16)
+#define POS_X_BAR_FWD		((byte)17)
+#define POS_X_TAB		((byte)18)
+#define POS_X_ARROW_BACK	((byte)19)
+#define POS_X_ARROW_FWD		((byte)20)
 
-#define POS_INSIDE		((byte)18)
-#define POS_MENU		((byte)19)
-#define POS_SCREENBUTTON	((byte)21)
-#define POS_ROOT		((byte)22)
+#define POS_Y_BAR_BACK		((byte)21)
+#define POS_Y_BAR_FWD		((byte)22)
+#define POS_Y_TAB		((byte)23)
+#define POS_Y_ARROW_BACK	((byte)24)
+#define POS_Y_ARROW_FWD		((byte)25)
+
+#define POS_INSIDE		((byte)26)
+#define POS_MENU		((byte)27)
+#define POS_BUTTON_SCREEN	((byte)29)
+#define POS_ROOT		((byte)30)
 
 typedef struct s_selection {
     timevalue Time;
@@ -1624,14 +1743,19 @@ typedef struct s_selection {
 
 typedef struct s_button_vec {
     hwfont shape[2];
-    num pos;
+    sbyte pos;
     byte exists;
     byte changed;
 } button_vec;
 
 
 struct s_all {
-    uldat Id; /* for compatibility with s_obj */
+    /* for compatibility with s_obj */
+    uldat Id;
+    fn_obj Fn;
+    obj Prev, Next;
+    obj Parent;
+    
     screen FirstScreen, LastScreen;
     msgport FirstMsgPort, LastMsgPort, RunMsgPort;
     mutex FirstMutex, LastMutex;
@@ -1672,64 +1796,7 @@ struct s_all {
 #define ENTER      	((udat)'\r')
 #define ESCAPE     	((udat)'\033')
 
-#define HOLD		((udat)1)
-#define HOLD_LEFT	((udat)1)
-#define HOLD_MIDDLE	((udat)2)
-#define HOLD_RIGHT	((udat)4)
-#define HOLD_ANY	(HOLD_LEFT|HOLD_MIDDLE|HOLD_RIGHT)
-#define HOLD_CODE(n)	(HOLD << (n)) /* n is 0,1,2 */
 
-#define PRESS_LEFT	((udat)0x08)
-#define PRESS_MIDDLE	((udat)0x18)
-#define PRESS_RIGHT	((udat)0x28)
-#define PRESS_ANY	((udat)0x38)
-#define PRESS_CODE(n)	((udat)0x08 | ((udat)(n) << 4)) /* n is 0,1,2 */
-
-#define DOWN_LEFT	(HOLD_LEFT|PRESS_LEFT)
-#define DOWN_MIDDLE	(HOLD_MIDDLE|PRESS_MIDDLE)
-#define DOWN_RIGHT	(HOLD_RIGHT|PRESS_RIGHT)
-#define DOWN_ANY	(HOLD_ANY|PRESS_ANY)
-#define DOWN_CODE(n)	(HOLD_CODE(n)|PRESS_CODE(n)) /* n is 0,1,2 */
-
-#define RELEASE_LEFT	((udat)0x10)
-#define RELEASE_MIDDLE	((udat)0x20)
-#define RELEASE_RIGHT	((udat)0x30)
-#define RELEASE_ANY	((udat)0x30)
-#define RELEASE_CODE(n)	((udat)0x10 + ((udat)(n) << 4)) /* n is 0,1,2 */
-
-#define DRAG_MOUSE	((udat)0x40)
-
-#define MOTION_MOUSE	((udat)0x00)
-
-#define ANY_ACTION_MOUSE	(PRESS_ANY | RELEASE_ANY | DRAG_MOUSE | MOTION_MOUSE)
-
-#define MAX_MOUSE_CODE	(udat)0x48
-
-#define isPRESS(code)	((code) & 0x08)
-#define isRELEASE(code)	((code) & ANY_ACTION_MOUSE && !((code) & (DRAG_MOUSE|(udat)0x08)))
-#define isDRAG(code)	((code) & DRAG_MOUSE)
-#define isMOTION(code)	(!(code))
-
-#define isSINGLE_PRESS(code) (isPRESS(code) && ((code) == DOWN_LEFT || (code) == DOWN_MIDDLE || (code) == DOWN_RIGHT))
-#define isSINGLE_DRAG(code) (isDRAG(code) && ((code) == (DRAG_MOUSE|HOLD_LEFT) || (code) == (DRAG_MOUSE|HOLD_MIDDLE) || (code) == (DRAG_MOUSE|HOLD_RIGHT)))
-#define isSINGLE_RELEASE(code) (isRELEASE(code) && !((code) & HOLD_ANY))
-
-/*
- * These macros can be used only for proper mouse codes.
- * The button numbers are: 0 (Left), 1 (Middle), 2 (Right)
- */
-
-/* if (Code & HOLD_ANY) then HOLD_N(n) returns the lowest button # pressed in Code */
-#define HOLD_N(Code)	(ffs(Code) - 1)
-
-/* if (isSINGLE_PRESS(Code)) then PRESS_N(n) return the button # pressed in Code */
-#define PRESS_N(Code)	((Code) >> 4)
-
-/* if (isSINGLE_DRAG(Code)) then DRAG_N(n) return the button # pressed in Code */
-#define DRAG_N(Code)	HOLD_N(Code)
-
-/* if (isSINGLE_RELEASE(Code)) then RELEASE_N(n) return the button # released in Code */
-#define RELEASE_N(Code) (((Code) >> 4) - 1)
 
 
 /**********************************/
@@ -1746,12 +1813,19 @@ struct s_all {
 /* don't use codes above or equal to this one! */
 #define COD_RESERVED		0xF800
 
+
+
+
 /* INLINE/define stuff: */
 
+
+
+
+
 #ifdef DEBUG_MALLOC
   /*
    * with the current MkDep, DEBUG_MALLOC gets defined only if doing
-   * `make DEBUG_MALLOC=1 ...' and the C file that includes twin.h actually
+   * `make DEBUG_MALLOC=y ...' and the C file that includes twin.h actually
    * checks for DEBUG_MALLOC. Anyway, this is acceptable :-)
    */
   extern byte *S;
