@@ -2,7 +2,7 @@
 
 #define T_TEMPLATE T_MAP(TEMPLATE)
 
-hwfont T_CAT(T_CAT(Tutf_,T_TEMPLATE),_to_UTF_16)[0x100] = {
+hwfont T_CAT3(Tutf_,T_TEMPLATE,_to_UTF_16)[0x100] = {
 #define EL(x) T_UTF(UTF_16,x),
     T_LIST(T_TEMPLATE,EL)
 #undef EL
@@ -18,7 +18,7 @@ static utf_to_ch T_CAT(array_,T_TEMPLATE) [] = {
 
 hwfont T_CAT(Tutf_UTF_16_to_,T_TEMPLATE)(hwfont c) {
     static utf_to_ch key;
-    utf_to_ch *res;
+    TW_CONST utf_to_ch *res;
     
     /* T_TEMPLATE obviously cannot contain all unicode chars. this is just a best effort. */
     if (!T_CAT(flag_,T_TEMPLATE)) {
@@ -31,7 +31,7 @@ hwfont T_CAT(Tutf_UTF_16_to_,T_TEMPLATE)(hwfont c) {
 	/* direct-to-font area */
 	(c >= ' ' && c <= '~') ||
 	/* ASCII area */
-	(c > '~' && c < 0x100 && T_CAT(T_CAT(Tutf_,T_TEMPLATE),_to_UTF_16)[c] == c))
+	(c > '~' && c < 0x100 && T_CAT3(Tutf_,T_TEMPLATE,_to_UTF_16)[c] == c))
 	/* c has the same meaning in Unicode and this charset... sheer luck! */
 
 	return c & 0x00ff;
@@ -43,7 +43,7 @@ hwfont T_CAT(Tutf_UTF_16_to_,T_TEMPLATE)(hwfont c) {
 	c = res->ch;
     else if (c > '~')
 	/* try to approximate */
-	c = T_CAT(Tutf_IBM437_to_,T_MAP(ASCII)) [ Tutf_UTF_16_to_IBM437(c) ];
+	c = T_CAT(Tutf_CP437_to_,T_MAP(ASCII)) [ Tutf_UTF_16_to_CP437(c) ];
     /* else c = c; */
 
     return key.ch = c;
