@@ -10,6 +10,13 @@
  *
  */
 
+#include <Tw/Twkeys.h>
+
+#ifdef CONF__UNICODE
+# include <Tutf/Tutf.h>
+# include <Tutf/Tutf_defs.h>
+#endif
+
 #include "twin.h"
 #include "main.h"
 #include "data.h"
@@ -20,12 +27,6 @@
 #include "hw_dirty.h"
 #include "common.h"
 
-#include "Tw/Twkeys.h"
-
-#ifdef CONF__UNICODE
-# include "Tutf/Tutf.h"
-# include "Tutf/Tutf_defs.h"
-#endif
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
@@ -33,6 +34,9 @@
 #include <X11/Xatom.h>
 #include <X11/Xmd.h>                /* CARD32 */
 
+
+#define THIS "hw_X11"
+
 /*
  * a user reported his system lacks DefaultRootWindow() ...
  */
@@ -102,7 +106,7 @@
 
 /* Display variables */
 
-#define NEST 2
+#define NEST 4
 
 
 
@@ -112,7 +116,7 @@ struct x11_data {
 #ifdef CONF__UNICODE
     Tutf_function xUTF_16_to_charset;
 #endif    
-    Display     *xdisplay;
+    Display     *dpy;
     Window       xwindow;
     GC           xgc;
     XGCValues    xsgc;
@@ -124,72 +128,42 @@ struct x11_data {
     uldat        XReqCount;
     XSelectionRequestEvent XReq[NEST];
     unsigned long xcol[MAXCOL+1];
-    Atom xWM_PROTOCOLS, xWM_DELETE_WINDOW;
+    Atom xWM_PROTOCOLS, xWM_DELETE_WINDOW, xTARGETS;
 };
 
-#define xdata	((struct x11_data *)HW->Private)
-#define xwidth	(xdata->xwidth)
-#define xheight	(xdata->xheight)
-#define xwfont	(xdata->xwfont)
-#define xhfont	(xdata->xhfont)
-#define xupfont	(xdata->xupfont)
+#define xdata		((struct x11_data *)HW->Private)
+#define xwidth		(xdata->xwidth)
+#define xheight		(xdata->xheight)
+#define xwfont		(xdata->xwfont)
+#define xhfont		(xdata->xhfont)
+#define xupfont		(xdata->xupfont)
 #define xUTF_16_to_charset	(xdata->xUTF_16_to_charset)
-#define xdisplay	(xdata->xdisplay)
-#define xwindow	(xdata->xwindow)
-#define xgc	(xdata->xgc)
-#define xsgc	(xdata->xsgc)
-#define xsfont	(xdata->xsfont)
+#define dpy	(xdata->dpy)
+#define xwindow		(xdata->xwindow)
+#define xgc		(xdata->xgc)
+#define xsgc		(xdata->xsgc)
+#define xsfont		(xdata->xsfont)
 #define xwindow_AllVisible	(xdata->xwindow_AllVisible)
 #define xfont_map	(xdata->xfont_map)
 #define xRequestor(j)	(xdata->xRequestor[j])
 #define xReqPrivate(j)	(xdata->xReqPrivate[j])
 #define xReqCount	(xdata->xReqCount)
 #define XReqCount	(xdata->XReqCount)
-#define XReq(j)	(xdata->XReq[j])
-#define xcol	(xdata->xcol)
-#define xWM_PROTOCOLS (xdata->xWM_PROTOCOLS)
-#define xWM_DELETE_WINDOW (xdata->xWM_DELETE_WINDOW)
+#define XReq(j)		(xdata->XReq[j])
+#define xcol		(xdata->xcol)
+#define xWM_PROTOCOLS	(xdata->xWM_PROTOCOLS)
+#define xWM_DELETE_WINDOW	(xdata->xWM_DELETE_WINDOW)
+#define xTARGETS	(xdata->xTARGETS)
 
 static void X11_SelectionNotify_up(Window win, Atom prop);
 static void X11_SelectionRequest_up(XSelectionRequestEvent *req);
 
 
 static void X11_Beep(void) {
-    XBell(xdisplay, 0);
+    XBell(dpy, 0);
     setFlush();
 }
 
-static struct {
-    KeySym xkey;
-    udat tkey;
-    byte len;
-    byte *seq;
-} X11_keys[] = {
-
-#define IS(sym,l,s) { XK_##sym, TW_##sym, l, s },
-#  include "hw_keys.h"
-#undef IS
-
-};
-
-static uldat X11_keyn = sizeof(X11_keys) / sizeof(X11_keys[0]);
-
-static byte X11_RemapKeys(void) {
-    uldat i;
-    
-    for (i = 0; i < X11_keyn; i++) {
-	XRebindKeysym(xdisplay, X11_keys[i].xkey, (KeySym *)0, 0, X11_keys[i].seq, X11_keys[i].len);
-	if (i && X11_keys[i-1].xkey >= X11_keys[i].xkey) {
-	    printk("\n"
-		   "      ERROR: twin compiled from a bad server/hw_keys.h file"
-		   "             (data in file is not sorted). X11 support is unusable.\n"
-		   "      X11_InitHW() failed: internal error.\n");
-	    return FALSE;
-	}
-    }
-    return TRUE;
-}
-
 static void X11_Configure(udat resource, byte todefault, udat value) {
     XKeyboardControl xctrl;
     
@@ -202,12 +176,12 @@ static void X11_Configure(udat resource, byte todefault, udat value) {
 	break;
       case HW_BELLPITCH:
 	xctrl.bell_pitch = todefault ? -1 : value;
-	XChangeKeyboardControl(xdisplay, KBBellPitch, &xctrl);
+	XChangeKeyboardControl(dpy, KBBellPitch, &xctrl);
 	setFlush();
 	break;
       case HW_BELLDURATION:
 	xctrl.bell_duration = todefault ? -1 : value;
-	XChangeKeyboardControl(xdisplay, KBBellDuration, &xctrl);
+	XChangeKeyboardControl(dpy, KBBellDuration, &xctrl);
 	setFlush();
 	break;
       case HW_MOUSEMOTIONEVENTS:
@@ -218,13 +192,49 @@ static void X11_Configure(udat resource, byte todefault, udat value) {
     }
 }
 
-    
+
+static struct {
+    KeySym xkey;
+    udat tkey;
+    byte len;
+    CONST byte *seq;
+} X11_keys[] = {
+
+#define IS(sym,l,s) { XK_##sym, TW_##sym, l, s },
+#  include "hw_keys.h"
+#undef IS
+
+};
+
+static uldat X11_keyn = sizeof(X11_keys) / sizeof(X11_keys[0]);
+
+/*
+ * calling XRebindKeysym() on non-english keyboards disrupts keys that use Mode_switch (AltGr),
+ * so we implement a manual translation using binary search on the table above.
+ * But first, check the table is sorted, or binary search would become bugged.
+ */
+static byte X11_CheckRemapKeys(void) {
+    uldat i;
+
+    for (i = 0; i < X11_keyn; i++) {
+	if (i && X11_keys[i-1].xkey >= X11_keys[i].xkey) {
+	    printk("\n"
+		   "      ERROR: twin compiled from a bad server/hw_keys.h file"
+		   "             (data in file is not sorted). -hw=X11 driver is unusable.\n"
+		   "      X11_InitHW() failed: internal error.\n");
+	    return FALSE;
+	}
+    }
+    return TRUE;
+}
+
 /* convert an X11 KeySym into a libTw key code and ASCII sequence */
 
 static udat X11_LookupKey(XKeyEvent *ev, udat *ShiftFlags, udat *len, char *seq) {
     static udat lastTW = TW_Null;
     static uldat lastI = MAXULDAT;
-    static KeySym sym, lastXK = NoSymbol;
+    static KeySym lastXK = NoSymbol;
+    KeySym sym;
     
     uldat i, low, up, _len = *len;
 
@@ -241,7 +251,9 @@ static udat X11_LookupKey(XKeyEvent *ev, udat *ShiftFlags, udat *len, char *seq)
 	*ShiftFlags |= KBD_NUM_LOCK;
     
     *len = XLookupString(ev, seq, _len, &sym, NULL);
-    
+#ifdef DEBUG_HW_X11
+    printf(THIS ": keysym=%d (%s) string='%.*s'\n", (int)sym, XKeysymToString(sym), (int)len, seq);
+#endif
     if (sym == XK_BackSpace && ev->state & (ControlMask|Mod1Mask)) {
 	if (ev->state & ControlMask)
 	    *len = 1, *seq = '\x1F';
@@ -250,7 +262,7 @@ static udat X11_LookupKey(XKeyEvent *ev, udat *ShiftFlags, udat *len, char *seq)
 	return TW_BackSpace;
     }
 
-    if (sym >= ' ' && sym <= '~') {
+    if ((sym >= ' ' && sym <= '~') || (sym >= 0xA0 && sym <= 0xFF)) {
 	/* turn ALT+A into ESC+A etc. */
 	if (ev->state & Mod1Mask && *len == 1 && (byte)sym == *seq) {
 	    *len = 2;
@@ -280,16 +292,17 @@ static udat X11_LookupKey(XKeyEvent *ev, udat *ShiftFlags, udat *len, char *seq)
 	}
 	lastXK = sym;
     }
-    if (*len == 0 && lastI < X11_keyn && X11_keys[lastI].len) {
-	/* XLookupString() returned empty string, steal sequence from hw_keys.h */
+    /* XLookupString() returned empty string, or no ShiftFlags: steal sequence from hw_keys.h */
+    if (lastI < X11_keyn && X11_keys[lastI].len && (*len == 0 || *ShiftFlags == 0)) {
 	if (_len > X11_keys[lastI].len)
 	    CopyMem(X11_keys[lastI].seq, seq, *len = X11_keys[lastI].len);
     }
     return lastTW;
-};
+}
 
 
 static void X11_HandleEvent(XEvent *event) {
+    /* this can stay static, X11_HandleEvent() is not reentrant */
     static byte seq[SMALLBUFF];
     dat x, y, dx, dy;
     udat len = SMALLBUFF, TW_key, ShiftFlags;
@@ -321,21 +334,44 @@ static void X11_HandleEvent(XEvent *event) {
 	    break;
 	}
 	
-	dx= event->xbutton.state;
-	dy= (dx & Button1Mask ? HOLD_LEFT : 0) |
+	dx = event->xbutton.state;
+	dy =(dx & Button1Mask ? HOLD_LEFT : 0) |
 	    (dx & Button2Mask ? HOLD_MIDDLE : 0) |
-	    (dx & Button3Mask ? HOLD_RIGHT : 0);
+	    (dx & Button3Mask ? HOLD_RIGHT : 0) |
+#ifdef HOLD_WHEEL_REV
+	    (dx & Button4Mask ? HOLD_WHEEL_REV : 0) |
+#endif
+#ifdef HOLD_WHEEL_FWD
+	    (dx & Button5Mask ? HOLD_WHEEL_FWD : 0) |
+#endif
+	    0;
+	
+	
+	/*
+	 * when sending ButtonRelease event, X11 reports that button
+	 * as still held down in event->xbutton.state ButtonMask...
+	 * 
+	 * instead when sending ButtonPress event, X11 reports that
+	 * button as still NOT held down in event->xbutton.state ButtonMask...
+	 * 
+	 * fix both cases here.
+	 */
 	dx = event->xbutton.button;
-	
+	dx = (dx == Button1 ? HOLD_LEFT :
+	      dx == Button2 ? HOLD_MIDDLE :
+	      dx == Button3 ? HOLD_RIGHT :
+#ifdef HOLD_WHEEL_REV
+	      dx == Button4 ? HOLD_WHEEL_REV :
+#endif
+#ifdef HOLD_WHEEL_FWD
+	      dx == Button5 ? HOLD_WHEEL_FWD :
+#endif
+	      0);
 	if (event->type == ButtonPress)
-	    dy|= (dx == Button1 ? HOLD_LEFT :
-		  dx == Button2 ? HOLD_MIDDLE :
-		  dx == Button3 ? HOLD_RIGHT : 0);
+	    dy |= dx;
 	else
-	    dy &= ~(dx == Button1 ? HOLD_LEFT :
-		    dx == Button2 ? HOLD_MIDDLE :
-		    dx == Button3 ? HOLD_RIGHT : 0);
-	
+	    dy &= ~dx;
+	    
 	MouseEventCommon(x, y, 0, 0, dy);
 	
 	break;
@@ -363,7 +399,7 @@ static void X11_HandleEvent(XEvent *event) {
 	xwindow_AllVisible = event->xvisibility.state == VisibilityUnobscured;
 	break;
       case SelectionClear:
-	HW->HWSelectionPrivate = NULL; /* selection now owned by some other X11 client */
+	HW->HWSelectionPrivate = 0; /* selection now owned by some other X11 client */
 	TwinSelectionSetOwner((obj)HW, SEL_CURRENTTIME, SEL_CURRENTTIME);
 	break;
       case SelectionRequest:
@@ -392,8 +428,8 @@ static void X11_KeyboardEvent(int fd, display_hw D_HW) {
     SaveHW;
     SetHW(D_HW);
     
-    while (XPending(xdisplay) > 0) {
-	XNextEvent(xdisplay, &event);
+    while (XPending(dpy) > 0) {
+	XNextEvent(dpy, &event);
 	X11_HandleEvent(&event);
     }
     
@@ -408,17 +444,17 @@ static hwcol _col;
 #ifdef CONF__UNICODE
 # define XDRAW(col, buf, buflen) \
     if (xsgc.foreground != xcol[COLFG(col)]) \
-	XSetForeground(xdisplay, xgc, xsgc.foreground = xcol[COLFG(col)]); \
+	XSetForeground(dpy, xgc, xsgc.foreground = xcol[COLFG(col)]); \
     if (xsgc.background != xcol[COLBG(col)]) \
-	XSetBackground(xdisplay, xgc, xsgc.background = xcol[COLBG(col)]); \
-	XDrawImageString16(xdisplay, xwindow, xgc, xbegin, ybegin + xupfont, buf, buflen)
+	XSetBackground(dpy, xgc, xsgc.background = xcol[COLBG(col)]); \
+    XDrawImageString16(dpy, xwindow, xgc, xbegin, ybegin + xupfont, buf, buflen)
 #else
 # define XDRAW(col, buf, buflen) \
     if (xsgc.foreground != xcol[COLFG(col)]) \
-	XSetForeground(xdisplay, xgc, xsgc.foreground = xcol[COLFG(col)]); \
+	XSetForeground(dpy, xgc, xsgc.foreground = xcol[COLFG(col)]); \
     if (xsgc.background != xcol[COLBG(col)]) \
-	XSetBackground(xdisplay, xgc, xsgc.background = xcol[COLBG(col)]); \
-	XDrawImageString(xdisplay, xwindow, xgc, xbegin, ybegin + xupfont, buf, buflen)
+	XSetBackground(dpy, xgc, xsgc.background = xcol[COLBG(col)]); \
+    XDrawImageString(dpy, xwindow, xgc, xbegin, ybegin + xupfont, buf, buflen)
 #endif
 
 INLINE void X11_Mogrify(dat x, dat y, uldat len) {
@@ -499,17 +535,17 @@ static void X11_ShowCursor(uldat type, dat x, dat y) {
 	if ((type & 0x40) && ((COLFG(v) & WHITE) == (COLBG(v) & WHITE)))
 	    v ^= COL(WHITE,0);
 	if (xsgc.foreground != xcol[COLFG(v)])
-	    XSetForeground(xdisplay, xgc, xsgc.foreground = xcol[COLFG(v)]);
+	    XSetForeground(dpy, xgc, xsgc.foreground = xcol[COLFG(v)]);
 	if (xsgc.background != xcol[COLBG(v)])
-	    XSetBackground(xdisplay, xgc, xsgc.background = xcol[COLBG(v)]);
+	    XSetBackground(dpy, xgc, xsgc.background = xcol[COLBG(v)]);
 #ifdef CONF__UNICODE
 	f = xUTF_16_to_charset(HWFONT(V));
 	c.byte1 = f >> 8;
 	c.byte2 = f & 0xFF;
-	XDrawImageString16(xdisplay, xwindow, xgc, x * xwfont, y * xhfont + xupfont, &c, 1);
+	XDrawImageString16(dpy, xwindow, xgc, x * xwfont, y * xhfont + xupfont, &c, 1);
 #else
 	c = HWFONT(V);
-	XDrawImageString(xdisplay, xwindow, xgc, x * xwfont, y * xhfont + xupfont, &c, 1);
+	XDrawImageString(dpy, xwindow, xgc, x * xwfont, y * xhfont + xupfont, &c, 1);
 #endif
     }
     if (type & 0xF) {
@@ -518,12 +554,12 @@ static void X11_ShowCursor(uldat type, dat x, dat y) {
 	
 	/* doesn't work as expected on paletted visuals... */
 	if (xsgc.foreground != xcol[COLFG(HWCOL(V)) ^ COLBG(HWCOL(V))])
-	    XSetForeground(xdisplay, xgc, xsgc.foreground = xcol[COLFG(HWCOL(V)) ^ COLBG(HWCOL(V))]);
+	    XSetForeground(dpy, xgc, xsgc.foreground = xcol[COLFG(HWCOL(V)) ^ COLBG(HWCOL(V))]);
 	
-	XSetFunction(xdisplay, xgc, xsgc.function = GXxor);
-	XFillRectangle(xdisplay, xwindow, xgc,
+	XSetFunction(dpy, xgc, xsgc.function = GXxor);
+	XFillRectangle(dpy, xwindow, xgc,
 		       x * xwfont, y * xhfont + xhfont - i, xwfont, i);
-	XSetFunction(xdisplay, xgc, xsgc.function = GXcopy);
+	XSetFunction(dpy, xgc, xsgc.function = GXcopy);
     }
 }
 
@@ -572,7 +608,7 @@ static void X11_FlushVideo(void) {
 }
 
 static void X11_FlushHW(void) {
-    XFlush(xdisplay);
+    XFlush(dpy);
     clrFlush();
 }
 
@@ -587,7 +623,7 @@ static void X11_CheckResize(dat *x, dat *y) {
 
 static void X11_Resize(dat x, dat y) {
     if (x != HW->X || y != HW->Y) {
-	XResizeWindow(xdisplay, xwindow,
+	XResizeWindow(dpy, xwindow,
 		      xwidth = xwfont * (HW->X = x),
 		      xheight = xhfont * (HW->Y = y));
 	setFlush();
@@ -606,8 +642,8 @@ static byte X11_SelectionImport_X11(void) {
  */
 static void X11_SelectionExport_X11(void) {
     if (!HW->HWSelectionPrivate) {
-	XSetSelectionOwner(xdisplay, XA_PRIMARY, xwindow, CurrentTime);
-	HW->HWSelectionPrivate = (void *)xwindow;
+	XSetSelectionOwner(dpy, XA_PRIMARY, xwindow, CurrentTime);
+	HW->HWSelectionPrivate = (tany)xwindow;
 	setFlush();
     }
 }
@@ -616,19 +652,16 @@ static void X11_SelectionExport_X11(void) {
  * notify our Selection to X11
  */
 static void X11_SelectionNotify_X11(uldat ReqPrivate, uldat Magic, CONST byte MIME[MAX_MIMELEN],
-				    uldat Len, byte CONST * Data) {
+				    uldat Len, CONST byte * Data) {
     XEvent ev;
-    static Atom xa_targets = None;
-    if (xa_targets == None)
-	xa_targets = XInternAtom (xdisplay, "TARGETS", False);
 	
     if (XReqCount == 0) {
-	printk("hw_X11.c: X11_SelectionNotify_X11(): unexpected Twin Selection Notify event!\n");
+	printk(THIS ".c: X11_SelectionNotify_X11(): unexpected Twin Selection Notify event!\n");
 	return;
     }
 #if 0
     else {
-	printk("hw_X11.c: X11_SelectionNotify_X11(): %d nested Twin Selection Notify events\n", XReqCount);
+	printk(THIS ".c: X11_SelectionNotify_X11(): %d nested Twin Selection Notify events\n", XReqCount);
     }
 #endif
     
@@ -641,7 +674,7 @@ static void X11_SelectionNotify_X11(uldat ReqPrivate, uldat Magic, CONST byte MI
     ev.xselection.target    = XReq(XReqCount).target;
     ev.xselection.time      = XReq(XReqCount).time;
 
-    if (XReq(XReqCount).target == xa_targets) {
+    if (XReq(XReqCount).target == xTARGETS) {
 	/*
 	 * On some systems, the Atom typedef is 64 bits wide.
 	 * We need to have a typedef that is exactly 32 bits wide,
@@ -650,20 +683,41 @@ static void X11_SelectionNotify_X11(uldat ReqPrivate, uldat Magic, CONST byte MI
 	typedef CARD32 Atom32;
 	Atom32 target_list[2];
 	
-	target_list[0] = (Atom32) xa_targets;
+	target_list[0] = (Atom32) xTARGETS;
 	target_list[1] = (Atom32) XA_STRING;
-	XChangeProperty (xdisplay, XReq(XReqCount).requestor, XReq(XReqCount).property,
-			 xa_targets, 8*sizeof(target_list[0]), PropModeReplace,
+	XChangeProperty (dpy, XReq(XReqCount).requestor, XReq(XReqCount).property,
+			 xTARGETS, 8*sizeof(target_list[0]), PropModeReplace,
 			 (char *)target_list,
 			 sizeof(target_list)/sizeof(target_list[0]));
 	ev.xselection.property = XReq(XReqCount).property;
     } else if (XReq(XReqCount).target == XA_STRING) {
-	XChangeProperty (xdisplay, XReq(XReqCount).requestor, XReq(XReqCount).property,
-			 XA_STRING, 8, PropModeReplace,
-			 Data, Len);
+#ifdef CONF__UNICODE
+	uldat l;
+	byte *_Data = NULL, *d;
+	TW_CONST hwfont *s;
+	
+	/* X11 selection contains text, not unicode */
+	if (Magic == SEL_HWFONTMAGIC) {
+	    if ((_Data = d = (byte *)AllocMem(Len))) {
+		s = (TW_CONST hwfont *)Data;
+		for (l = Len; l; l--)
+		    *d++ = Tutf_UTF_16_to_CP437(*s++);
+		Data = _Data;
+		Len /= sizeof(hwfont);
+	    } else
+		Len = 0;
+	}
+#endif
+	XChangeProperty (dpy, XReq(XReqCount).requestor, XReq(XReqCount).property,
+			 XA_STRING, 8, PropModeReplace, Data, Len);
 	ev.xselection.property = XReq(XReqCount).property;
+
+#ifdef CONF__UNICODE
+	if (Magic == SEL_HWFONTMAGIC && _Data)
+	    FreeMem(_Data);
+#endif
     }
-    XSendEvent (xdisplay, XReq(XReqCount).requestor, False, 0, &ev);
+    XSendEvent (dpy, XReq(XReqCount).requestor, False, 0, &ev);
     setFlush();
 }
 
@@ -678,12 +732,12 @@ static void X11_SelectionNotify_up(Window win, Atom prop) {
     byte *data, *buff = NULL;
 
     if (xReqCount == 0) {
-	printk("hw_X11.c: X11_SelectionNotify_up(): unexpected X Selection Notify event!\n");
+	printk(THIS ".c: X11_SelectionNotify_up(): unexpected X Selection Notify event!\n");
 	return;
     }
 #if 0
     else {
-	printk("hw_X11.c: X11_SelectionNotify_up(): %d nested X Selection Notify event\n", xReqCount);
+	printk(THIS ".c: X11_SelectionNotify_up(): %d nested X Selection Notify event\n", xReqCount);
     }
 #endif
     if (prop == None)
@@ -692,7 +746,7 @@ static void X11_SelectionNotify_up(Window win, Atom prop) {
     xReqCount--;
 
     do {
-	if ((XGetWindowProperty(xdisplay, win, prop,
+	if ((XGetWindowProperty(dpy, win, prop,
 				nread/4, bytes_after/4, False,
 				AnyPropertyType, &actual_type, &actual_fmt,
 				&nitems, &bytes_after, &data)
@@ -724,23 +778,23 @@ static void X11_SelectionRequest_X11(obj Requestor, uldat ReqPrivate) {
     if (!HW->HWSelectionPrivate) {
 
 	if (xReqCount == NEST) {
-	    printk("hw_X11.c: X11_SelectionRequest_X11(): too many nested Twin Selection Request events!\n");
+	    printk(THIS ".c: X11_SelectionRequest_X11(): too many nested Twin Selection Request events!\n");
 	    return;
 	}
 #if 0
 	else {
-	    printk("hw_X11.c: X11_SelectionRequest_X11(): %d nested Twin Selection Request events\n", xReqCount+1);
+	    printk(THIS ".c: X11_SelectionRequest_X11(): %d nested Twin Selection Request events\n", xReqCount+1);
 	}
 #endif
 	xRequestor(xReqCount) = Requestor;
 	xReqPrivate(xReqCount) = ReqPrivate;
 	xReqCount++;
 	
-	if (XGetSelectionOwner(xdisplay, XA_PRIMARY) == None)
-	    X11_SelectionNotify_up(DefaultRootWindow(xdisplay), XA_CUT_BUFFER0);
+	if (XGetSelectionOwner(dpy, XA_PRIMARY) == None)
+	    X11_SelectionNotify_up(DefaultRootWindow(dpy), XA_CUT_BUFFER0);
 	else {
-	    Atom prop = XInternAtom (xdisplay, "VT_SELECTION", False);
-	    XConvertSelection(xdisplay, XA_PRIMARY, XA_STRING, prop, xwindow, CurrentTime);
+	    Atom prop = XInternAtom (dpy, "VT_SELECTION", False);
+	    XConvertSelection(dpy, XA_PRIMARY, XA_STRING, prop, xwindow, CurrentTime);
 	    setFlush();
 	    /* we will get an X11 SelectionNotify event */
 	}
@@ -754,17 +808,18 @@ static void X11_SelectionRequest_X11(obj Requestor, uldat ReqPrivate) {
  */
 static void X11_SelectionRequest_up(XSelectionRequestEvent *req) {
     if (XReqCount == NEST) {
-	printk("hw_X11.c: X11_SelectionRequest_up(): too many nested X Selection Request events!\n");
+	printk(THIS ".c: X11_SelectionRequest_up(): too many nested X Selection Request events!\n");
 	return;
     }
 #if 0
     else {
-	printk("hw_X11.c: X11_SelectionRequest_up(): %d nested X Selection Request events\n", XReqCount+1);
+	printk(THIS ".c: X11_SelectionRequest_up(): %d nested X Selection Request events\n", XReqCount+1);
     }
 #endif
     CopyMem(req, &XReq(XReqCount), sizeof(XSelectionRequestEvent));
     TwinSelectionRequest((obj)HW, XReqCount++, TwinSelectionGetOwner());
     /* we will get a HW->HWSelectionNotify (i.e. X11_SelectionNotify_X11) call */
+    /* the call **CAN** arrive while we are still inside TwinSelectionRequest() !!! */
 }
 
 static byte X11_CanDragArea(dat Left, dat Up, dat Rgt, dat Dwn, dat DstLeft, dat DstUp) {
@@ -789,7 +844,7 @@ static void X11_DragArea(dat Left, dat Up, dat Rgt, dat Dwn, dat DstLeft, dat Ds
 	    HW->TT = (uldat)-1;
 	}
     }
-    XCopyArea(xdisplay, xwindow, xwindow, xgc,
+    XCopyArea(dpy, xwindow, xwindow, xgc,
 	      Left*xwfont, Up*xhfont, (Rgt-Left+1)*xwfont, (Dwn-Up+1)*xhfont,
 	      DstLeft*xwfont, DstUp*xhfont);
     setFlush();
@@ -798,14 +853,14 @@ static void X11_DragArea(dat Left, dat Up, dat Rgt, dat Dwn, dat DstLeft, dat Ds
 
 static void X11_QuitHW(void) {
 
-    if (xsfont) XFreeFont(xdisplay, xsfont);
-    if (xgc != None) XFreeGC(xdisplay, xgc);
+    if (xsfont) XFreeFont(dpy, xsfont);
+    if (xgc != None) XFreeGC(dpy, xgc);
     if (xwindow != None) {
-	XUnmapWindow(xdisplay, xwindow);
-	XDestroyWindow(xdisplay, xwindow);
+	XUnmapWindow(dpy, xwindow);
+	XDestroyWindow(dpy, xwindow);
     }
-    XCloseDisplay(xdisplay);
-    xdisplay = NULL;
+    XCloseDisplay(dpy);
+    dpy = NULL;
     
     if (HW->keyboard_slot != NOSLOT)
 	UnRegisterRemote(HW->keyboard_slot);
@@ -830,7 +885,7 @@ static int X11_Die(Display *d) {
      */
     forallHW {
 	if (HW->QuitHW == X11_QuitHW && HW->Private
-	    && d == xdisplay) { /* expands to HW->Private->xdisplay */
+	    && d == dpy) { /* expands to HW->Private->dpy */
 	    
 	    HW->NeedHW |= NEEDPanicHW, NeedHW |= NEEDPanicHW;
 	    
@@ -857,19 +912,19 @@ static Tutf_function X11_UTF_16_to_charset_function(CONST byte *charset) {
     if (!charset) {
 	/* attempt to autodetect encoding. */
 	fp = xsfont->properties;
-	fontatom = XInternAtom (xdisplay, "FONT", False);
+	fontatom = XInternAtom (dpy, "FONT", False);
 	i = xsfont->n_properties;
 	
 	while (i--) {
 	    if (fp->name == fontatom) {
-		fontname = XGetAtomName(xdisplay, fp->card32);
+		fontname = XGetAtomName(dpy, fp->card32);
 		/*fprintf(stderr, "    X11_UTF_16_to...: font name: `%s\'\n", fontname);*/
 		break;
 	    }
 	    fp++;
 	}
 	if (fontname && !strcmp(fontname, "vga")) {
-	    charset = T_NAME_IBM437;
+	    charset = T_NAME_CP437;
 	} else if (fontname) {
 	    i = 2;
 	    for (s = fontname + strlen(fontname) - 1; i && s >= fontname; s--) {
@@ -883,14 +938,14 @@ static Tutf_function X11_UTF_16_to_charset_function(CONST byte *charset) {
 	if (!charset) {
 	    if (xsfont->min_byte1 < xsfont->max_byte1) {
 		/* font is more than just 8-bit. For now, assume it's unicode */
-		printk("    X11_InitHW: font `%s\' has no known charset encoding,\n"
+		printk("    X11_InitHW: font `%."STR(SMALLBUFF)"s\' has no known charset encoding,\n"
 		       "                assuming Unicode.\n", fontname);
 		return NULL;
 	    }
 	    /* else assume codepage437. gross. */
-	    printk("    X11_InitHW: font `%s\' has no known charset encoding,\n"
-		   "                assuming IBM437 codepage (\"VGA\").\n", fontname);
-	    return Tutf_UTF_16_to_IBM437;
+	    printk("    X11_InitHW: font `%."STR(SMALLBUFF)"s\' has no known charset encoding,\n"
+		   "                assuming CP437 codepage (\"VGA\").\n", fontname);
+	    return Tutf_UTF_16_to_CP437;
 	}
     }
     
@@ -902,8 +957,8 @@ static Tutf_function X11_UTF_16_to_charset_function(CONST byte *charset) {
     }
     
     if (i == (uldat)-1) {
-	printk("      X11_InitHW(): libTutf warning: unknown charset `%." STR(SMALLBUFF) "s', assuming `IBM437'\n", charset);
-	return Tutf_UTF_16_to_IBM437;
+	printk("      X11_InitHW(): libTutf warning: unknown charset `%." STR(SMALLBUFF) "s', assuming `CP437'\n", charset);
+	return Tutf_UTF_16_to_CP437;
     }
     
     return Tutf_UTF_16_to_charset_function(i);
@@ -920,7 +975,7 @@ static hwfont X11_UTF_16_to_UTF_16(hwfont c) {
 #endif
 
     
-#ifdef CONF_THIS_MODULE
+#ifdef THIS_MODULE
 static
 #endif
 byte X11_InitHW(void) {
@@ -932,7 +987,7 @@ byte X11_InitHW(void) {
     XSizeHints *xhints;
     XEvent event;
     int i;
-    byte *s, *dpy = NULL, *dpy0 = NULL,
+    byte *s, *dpy_ = NULL, *dpy0 = NULL,
     *fontname = NULL, *fontname0 = NULL,
     *charset = NULL, *charset0 = NULL,
     name[] = "twin :??? on X";
@@ -953,9 +1008,11 @@ byte X11_InitHW(void) {
 	    arg += 2; /* `X11' is same as `X' */
 
 	if (*arg == '@') {
-	    s = strchr(dpy = ++arg, ',');
-	    if (s) *(dpy0 = s) = '\0';
-	    arg = s;
+	    if ((s = strchr(dpy_ = ++arg, ','))) {
+		*(dpy0 = s) = '\0';
+		arg = s + 1;
+	    } else
+		arg = NULL;
 	}
 
 	while (arg && *arg) {
@@ -980,7 +1037,8 @@ byte X11_InitHW(void) {
 	    } else if (!strncmp(arg, "noinput", 7)) {
 		arg += 7;
 		noinput = TRUE;
-	    }
+	    } else
+		arg = strchr(arg, ',');
 	}
     }
 
@@ -989,21 +1047,21 @@ byte X11_InitHW(void) {
     xReqCount = XReqCount = 0;
     HW->keyboard_slot = NOSLOT;
     
-    if ((xdisplay = XOpenDisplay(dpy))) do {
+    if ((dpy = XOpenDisplay(dpy_))) do {
 	
 	(void)XSetIOErrorHandler(X11_Die);
-	
-	if (!X11_RemapKeys())
+
+	if (!X11_CheckRemapKeys())
 	    break;
 
-	xscreen = DefaultScreen(xdisplay);
-	xdepth  = DefaultDepth(xdisplay, xscreen);
+	xscreen = DefaultScreen(dpy);
+	xdepth  = DefaultDepth(dpy, xscreen);
 	
 	for (i = 0; i <= MAXCOL; i++) {
 	    xcolor.red   = 257 * (udat)Palette[i].Red;
 	    xcolor.green = 257 * (udat)Palette[i].Green;
 	    xcolor.blue  = 257 * (udat)Palette[i].Blue;
-	    if (!XAllocColor(xdisplay, DefaultColormap(xdisplay, xscreen), &xcolor)) {
+	    if (!XAllocColor(dpy, DefaultColormap(dpy, xscreen), &xcolor)) {
 		printk("      X11_InitHW() failed to allocate colors\n");
 		break;
 	    }
@@ -1016,30 +1074,30 @@ byte X11_InitHW(void) {
 	xattr.event_mask = ExposureMask | VisibilityChangeMask |
 	    StructureNotifyMask | SubstructureNotifyMask |
 	    KeyPressMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask;
-	
-	if (((fontname && (xsfont = XLoadQueryFont(xdisplay, fontname))) ||
-	     (xsfont = XLoadQueryFont(xdisplay, "vga")) ||
-	     (xsfont = XLoadQueryFont(xdisplay, "fixed"))) &&
+
+	if (((fontname && (xsfont = XLoadQueryFont(dpy, fontname))) ||
+	     (xsfont = XLoadQueryFont(dpy, "vga")) ||
+	     (xsfont = XLoadQueryFont(dpy, "fixed"))) &&
 	    (xwfont = xsfont->min_bounds.width,
 	     xwidth = xwfont * (HW->X = GetDisplayWidth()),
 	     xhfont = (xupfont = xsfont->ascent) + xsfont->descent,
 	     xheight = xhfont * (HW->Y = GetDisplayHeight()),
-	     xwindow = XCreateWindow(xdisplay, RootWindow(xdisplay, xscreen), 0, 0,
+	     xwindow = XCreateWindow(dpy, DefaultRootWindow(dpy), 0, 0,
 				     xwidth, xheight, 0, xdepth, InputOutput,
-				     DefaultVisual(xdisplay, xscreen),
+				     DefaultVisual(dpy, xscreen),
 				     CWBackPixel | CWEventMask, &xattr)) &&
-	    (xgc = XCreateGC(xdisplay, xwindow, 0, NULL)) &&
+
+	    (xsgc.foreground = xsgc.background = xcol[0],
+	     xsgc.graphics_exposures = False,
+	     xsgc.font = xsfont->fid,
+	     xgc = XCreateGC(dpy, xwindow, GCFont|GCForeground|GCBackground|GCGraphicsExposures, &xsgc)) &&
+
 	    (xhints = XAllocSizeHints())) {
 	    
-	    xsgc.foreground = xsgc.background = xcol[0];
-	    xsgc.graphics_exposures = False;
-	    XChangeGC(xdisplay, xgc, GCForeground|GCBackground|GCGraphicsExposures, &xsgc);
-	    XSetFont(xdisplay, xgc, xsfont->fid);
-	    
-	    strcpy(name+5, TWDisplay);
-	    strcat(name+5, " on X");
-	    XStoreName(xdisplay, xwindow, name);
-	    
+	    sprintf(name, "twin %s on X", TWDisplay);
+	    XStoreName(dpy, xwindow, name);
+
+
 #ifdef CONF__UNICODE
 	    if (!(xUTF_16_to_charset = X11_UTF_16_to_charset_function(charset)))
 		xUTF_16_to_charset = X11_UTF_16_to_UTF_16;
@@ -1048,27 +1106,28 @@ byte X11_InitHW(void) {
 	     * ask ICCCM-compliant window manager to tell us when close window
 	     * has been chosen, rather than just killing us
 	     */
-	    xWM_PROTOCOLS = XInternAtom(xdisplay, "WM_PROTOCOLS", False);
-	    xWM_DELETE_WINDOW = XInternAtom(xdisplay, "WM_DELETE_WINDOW", False);
-	    
-	    XChangeProperty(xdisplay, xwindow, xWM_PROTOCOLS, XA_ATOM, 32, PropModeReplace,
+	    xWM_PROTOCOLS = XInternAtom(dpy, "WM_PROTOCOLS", False);
+	    xWM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
+	    xTARGETS = XInternAtom(dpy, "TARGETS", False);
+
+	    XChangeProperty(dpy, xwindow, xWM_PROTOCOLS, XA_ATOM, 32, PropModeReplace,
 			    (unsigned char *) &xWM_DELETE_WINDOW, 1);
 	    
 	    xhints->flags = PResizeInc;
 	    xhints->width_inc  = xwfont;
 	    xhints->height_inc = xhfont;
-	    XSetWMNormalHints(xdisplay, xwindow, xhints);
+	    XSetWMNormalHints(dpy, xwindow, xhints);
 	    
-	    XMapWindow(xdisplay, xwindow);
+	    XMapWindow(dpy, xwindow);
 	    
 	    do {
-		XNextEvent(xdisplay, &event);
+		XNextEvent(dpy, &event);
 	    } while (event.type != MapNotify);
 	    
 	    XFree(xhints); xhints = NULL;
 	    
 	    HW->mouse_slot = NOSLOT;
-	    HW->keyboard_slot = RegisterRemote(i = XConnectionNumber(xdisplay), (obj)HW,
+	    HW->keyboard_slot = RegisterRemote(i = XConnectionNumber(dpy), (obj)HW,
 					       X11_KeyboardEvent);
 	    if (HW->keyboard_slot == NOSLOT)
 		break;
@@ -1096,7 +1155,7 @@ byte X11_InitHW(void) {
 	    HW->HWSelectionExport  = X11_SelectionExport_X11;
 	    HW->HWSelectionRequest = X11_SelectionRequest_X11;
 	    HW->HWSelectionNotify  = X11_SelectionNotify_X11;
-	    HW->HWSelectionPrivate = NULL;
+	    HW->HWSelectionPrivate = 0;
 	    
 	    if (drag) {
 		HW->CanDragArea = X11_CanDragArea;
@@ -1141,8 +1200,8 @@ byte X11_InitHW(void) {
 	    return TRUE;
 	}
     } while (0); else {
-	if (dpy || (dpy = getenv("DISPLAY")))
-	    printk("      X11_InitHW() failed to open display %s\n", arg);
+	if (dpy_ || (dpy_ = getenv("DISPLAY")))
+	    printk("      X11_InitHW() failed to open display %."STR(SMALLBUFF)"s\n", HW->Name);
 	else
 	    printk("      X11_InitHW() failed: DISPLAY is not set\n");
     }
@@ -1151,7 +1210,7 @@ byte X11_InitHW(void) {
     if (fontname0) *fontname0 = ',';
     if (charset0) *charset0 = ',';
 	
-    if (xdisplay)
+    if (dpy)
 	X11_QuitHW();
 
     FreeMem(HW->Private);
@@ -1159,11 +1218,8 @@ byte X11_InitHW(void) {
     return FALSE;
 }
 
-#ifdef CONF_THIS_MODULE
+#ifdef THIS_MODULE
 
-#include "version.h"
-MODULEVERSION;
-		       
 byte InitModule(module Module) {
     Module->Private = X11_InitHW;
     return TRUE;
@@ -1173,4 +1229,4 @@ byte InitModule(module Module) {
 void QuitModule(module Module) {
 }
 
-#endif /* CONF_THIS_MODULE */
+#endif /* THIS_MODULE */
