@@ -29,34 +29,40 @@
 #include "fdlist.h"
 #include "version.h"
 
-#include "Tw/Tw.h"
-#include "Tw/Twerrno.h"
-#include "Tw/Twstat.h"
+#include <Tw/Tw.h>
+#include <Tw/Twerrno.h>
+#include <Tw/Twstat.h>
+#include <Tw/Twstat_defs.h>
 
-/* HW specific headers */
-
-#ifdef CONF_HW_X11
-# include "HW/hw_X11.h"
-#endif
-
-#ifdef CONF_HW_TWIN
-# include "HW/hw_twin.h"
-#endif
-
-#ifdef CONF_HW_TTY
-# include "HW/hw_tty.h"
-#endif
-
-#ifdef CONF_HW_GGI
-# include "HW/hw_ggi.h"
-#endif
 
 
 
 
 #ifdef CONF__MODULES
 
-# ifdef LIBDIR
+
+# if defined(HAVE_DLFCN_H) && defined(HAVE_DLOPEN)
+
+#  include <dlfcn.h>
+# define my(fn) fn
+# define my_handle void *
+# define my_dlopen_extra_args , RTLD_NOW|RTLD_GLOBAL
+# define my_VERSION ".so." TWIN_VERSION_STR
+
+#elif defined(HAVE_LTDL_H) && defined(HAVE_LT_DLOPEN)
+
+# include <ltdl.h>
+# define my(fn) lt_##fn
+# define my_handle lt_dlhandle
+# define my_dlopen_extra_args
+# define my_VERSION ".la"
+
+#else
+# error nor dlopen() nor lt_dlopen() module loading API available!  
+#endif
+
+
+#ifdef LIBDIR
 #  define DIR_LIB_TWIN_MODULES_ LIBDIR "/twin/modules/"
 # else
 #  define DIR_LIB_TWIN_MODULES_ "./"
@@ -66,8 +72,6 @@ static CONST byte *conf_destdir_lib_twin_modules_ = DIR_LIB_TWIN_MODULES_;
 
 #endif /* CONF__MODULES */
 
-
-
 static CONST byte *MYname;
 
 static dat TryDisplayWidth, TryDisplayHeight;
@@ -80,6 +84,8 @@ byte nullMIME[TW_MAX_MIMELEN];
 udat ErrNo;
 byte CONST * ErrStr;
 
+byte CONST * HOME;
+
 #define L 0x55
 #define M 0xAA
 #define H 0xFF
@@ -277,8 +283,6 @@ static void RemoteEvent(int FdCount, fd_set *FdSet) {
 
 #ifdef CONF__MODULES
 
-#include <dlfcn.h>
-
 static struct s_fn_module _FnModule = {
     module_magic, (uldat)sizeof(struct s_module), (uldat)1,
 	(void *)NoOp, /* CreateModule */
@@ -301,32 +305,32 @@ static module DlLoadAny(uldat len, byte *name) {
     byte *_name;
     
     if ((Module->Name = CloneStrL(name, len)) &&
-	(_name = AllocMem(len + strlen(conf_destdir_lib_twin_modules_) + 1))) {
+	(_name = AllocMem(len + strlen(conf_destdir_lib_twin_modules_) + strlen(my_VERSION) + 1))) {
 	
-	sprintf(_name, "%s%s", conf_destdir_lib_twin_modules_, Module->Name);
+	sprintf(_name, "%s%.*s%s", conf_destdir_lib_twin_modules_, (int)len, name, my_VERSION);
 
-	if ((Module->Handle = dlopen(_name, RTLD_NOW|RTLD_GLOBAL)) &&
+	if ((Module->Handle = my(dlopen)(_name my_dlopen_extra_args)) &&
 	    /*
 	     * Module MUST have a InitModule function, as it needs to set
 	     * Module->Private to its xxx_InitHW() startup code.
 	     */
-	    (init_dl = dlsym(Module->Handle, "InitModule")) &&
-	    init_dl(Module))
+	    (init_dl = (byte (*)(module)) my(dlsym)((my_handle)Module->Handle, "InitModule")) &&
+	    init_dl(Module)) {
 
+	    FreeMem(_name);
 	    return Module;
-	else
-	    ErrStr = dlerror();
+	} else
+	    ErrStr = my(dlerror)();
+	FreeMem(_name);
     } else
 	ErrStr = "Out of memory!";
     return (module)0;
 }
 
-static byte module_InitHW(void) {
+static byte module_InitHW(byte *arg, uldat len) {
     byte *name, *tmp;
     byte *(*InitD)(void);
-    byte *arg = HW->Name;
-    uldat len = HW->NameLen;
-    module Module;
+    module Module = NULL;
 
     if (!arg || len <= 4)
 	return FALSE;
@@ -340,32 +344,35 @@ static byte module_InitHW(void) {
     if (name)
 	len = name - arg;
     
-    if ((name = AllocMem(len + 10))) {
-	sprintf(name, "HW/hw_%.*s.so", (int)len, arg);
+    if ((name = AllocMem(len + 7))) {
+	sprintf(name, "HW/hw_%.*s", (int)len, arg);
 			
-	Module = DlLoadAny(len+9, name);
+	Module = DlLoadAny(len + 6, name);
 	
-	FreeMem(name);
-	    
 	if (Module) {
-	    printk("twdisplay: starting display driver module `HW/hw_%.*s.so'...\n", (int)len, arg);
+	    printk("twdisplay: starting display driver module `%."STR(SMALLBUFF)"s'...\n", name);
 	    
 	    if ((InitD = Module->Private) && InitD()) {
-		printk("twdisplay: ...module `HW/hw_%.*s.so' successfully started.\n", (int)len, arg);
+		printk("twdisplay: ...module `%."STR(SMALLBUFF)"s' successfully started.\n", name);
 		HW->Module = Module; Module->Used++;
+		
+		FreeMem(name);
 		return TRUE;
 	    }
 	    /*Delete(Module);*/
 	}
     } else
 	ErrStr = "Out of memory!";
-
-    if (Module) {
-	printk("twdisplay: ...module `HW/hw_%.*s.so' failed to start.\n", (int)len, arg);
-    } else
-	printk("twdisplay: unable to load display driver module `HW/hw_%.*s.so' :\n"
-		"      %s\n", (int)len, arg, ErrStr);
     
+    if (Module) {
+	printk("twdisplay: ...module `%."STR(SMALLBUFF)"s' failed to start.\n", name ? name : (byte *)"(NULL)");
+    } else
+	printk("twdisplay: unable to load display driver module `%."STR(SMALLBUFF)"s' :\n"
+	       "      %."STR(SMALLBUFF)"s\n", name ? name : (byte *)"(NULL)", ErrStr);
+
+    if (name)
+    	FreeMem(name);
+
     return FALSE;
 }
 
@@ -392,19 +399,32 @@ static struct s_display_hw _HW = {
 	&_FnDisplayHW,
 };
 
-#if defined(CONF_HW_X11) || defined(CONF_HW_TWIN) || defined(CONF_HW_TTY) || defined(CONF_HW_GGI)
+#if defined(CONF__MODULES) || defined(CONF_HW_TWIN)
 static byte check4(byte *s, byte *arg) {
+    if (arg && !strncmp(s, arg, strlen(s))) {
+	printk("twin: trying given `--hw=%."STR(SMALLBUFF)"s' display driver.\n", s);
+	return TRUE;
+    }
+    return FALSE;
+}
+#endif /* defined(CONF__MODULES) || defined(CONF_HW_TWIN) */
+
+#if defined(CONF__MODULES) || defined(CONF_HW_GFX) || defined(CONF_HW_X11) || defined(CONF_HW_TTY) || defined(CONF_HW_GGI)
+static byte autocheck4(byte *s, byte *arg) {
     if (arg && strncmp(s, arg, strlen(s))) {
-	printk("twdisplay: `-hw=%s' given, skipping `-hw=%s' display driver.\n",
+	printk("twdisplay: `--hw=%."STR(SMALLBUFF)"s' given, skipping `--hw=%."STR(SMALLBUFF)"s' display driver.\n",
 		arg, s);
 	return FALSE;
     } else if (arg)
-	printk("twdisplay: trying given `-hw=%s' display driver.\n", s);
+	printk("twdisplay: trying given `--hw=%."STR(SMALLBUFF)"s' display driver.\n", s);
     else
-	printk("twdisplay: autoprobing `-hw=%s' display driver.\n", s);
+	printk("twdisplay: autoprobing `--hw=%."STR(SMALLBUFF)"s' display driver.\n", s);
     return TRUE;
 }
+#endif /* defined(CONF__MODULES) || defined(CONF_HW_GFX) || defined(CONF_HW_X11) || defined(CONF_HW_TTY) || defined(CONF_HW_GGI) */
 
+
+#if defined(CONF__MODULES) || defined(CONF_HW_GFX) || defined(CONF_HW_X11) || defined(CONF_HW_TWIN) || defined(CONF_HW_TTY) || defined(CONF_HW_GGI)
 static void fix4(byte *s, display_hw D_HW) {
     uldat len;
     if (!D_HW->NameLen) {
@@ -417,19 +437,19 @@ static void fix4(byte *s, display_hw D_HW) {
 	}
     }
 }
-#endif /* defined(CONF_HW_X11) || defined(CONF_HW_TWIN) || defined(CONF_HW_TTY) || defined(CONF_HW_GGI) */
+#endif /* defined(CONF__MODULES) || defined(CONF_HW_GFX) || defined(CONF_HW_X11) || defined(CONF_HW_TWIN) || defined(CONF_HW_TTY) || defined(CONF_HW_GGI) */
 
 static void warn_NoHW(uldat len, char *arg, uldat tried) {
 #ifdef CONF__MODULES
     if (!tried && !arg)
 	printk("twdisplay: no display driver compiled into twdisplay.\n"
-	       "      please run as `twdisplay [-twin@<TWDISPLAY>] -hw=<display>'\n");
+	       "      please run as `twdisplay [--twin@<TWDISPLAY>] --hw=<display>'\n");
     else
 #endif
     {
 	printk("twdisplay: All display drivers failed");
 	if (arg)
-	    printk(" for `-hw=%.*s\'", (int)len, arg);
+	    printk(" for `--hw=%.*s\'", (int)len, arg);
 	else
 	    printk(".");
 	printk("\n");
@@ -444,6 +464,64 @@ static void UpdateFlagsHW(void) {
     }
 }
 
+#ifdef CONF__MODULES
+# define DEF_INITHW(hw) \
+static byte CAT(hw,_InitHW)(void) { \
+    byte *arg; \
+    uldat len; \
+    if (HW->Name && HW->NameLen) { \
+	arg = HW->Name; \
+	len = HW->NameLen; \
+    } else { \
+	arg = "-hw=" STR(hw); \
+	len = strlen(arg); \
+    } \
+    return module_InitHW(arg, len); \
+}
+#else
+# define DEF_INITHW(hw)
+#endif
+
+
+/* HW specific functions */
+
+#ifdef CONF_HW_GFX
+# include "HW/hw_gfx.h"
+#else
+  DEF_INITHW(gfx)
+#endif
+
+#ifdef CONF_HW_X11
+# include "HW/hw_X11.h"
+# define X_InitHW X11_InitHW
+#else
+  DEF_INITHW(X)
+#endif
+	
+#ifdef CONF_HW_TWIN
+# include "HW/hw_twin.h"
+# define twin_InitHW TW_InitHW
+#else
+  DEF_INITHW(twin)
+#endif
+
+#ifdef CONF_HW_TTY
+# include "HW/hw_tty.h"
+#else
+  DEF_INITHW(tty)
+#endif
+
+#ifdef CONF_HW_GGI
+# include "HW/hw_ggi.h"
+# define ggi_InitHW GGI_InitHW
+#else
+  DEF_INITHW(ggi)
+#endif
+
+#undef DEF_INITHW
+
+
+
 /*
  * InitDisplayHW runs HW specific InitXXX() functions, starting from best setup
  * and falling back in case some of them fails.
@@ -463,26 +541,33 @@ static byte InitDisplayHW(display_hw D_HW) {
     else
 	arg = NULL;
 
+#define     TRY4(hw) (    check4(STR(hw), arg) && (tried++, CAT(hw,_InitHW)()) && (fix4(STR(hw), D_HW), TRUE))
+#define AUTOTRY4(hw) (autocheck4(STR(hw), arg) && (tried++, CAT(hw,_InitHW)()) && (fix4(STR(hw), D_HW), TRUE))
+
     success =
-#ifdef CONF_HW_X11
-	(check4("X", arg) && (tried++, X11_InitHW()) && (fix4("X", D_HW), TRUE)) ||
+#if defined(CONF__MODULES) || defined(CONF_HW_GFX)
+	AUTOTRY4(gfx) ||
 #endif
-#ifdef CONF_HW_TWIN
-	(check4("twin", arg) && (tried++, TW_InitHW()) && (fix4("twin", D_HW), TRUE)) ||
+#if defined(CONF__MODULES) || defined(CONF_HW_X11)
+	AUTOTRY4(X) ||
+#endif
+#if defined(CONF__MODULES) || defined(CONF_HW_TWIN)
+	AUTOTRY4(twin) ||
+#endif
+#if defined(CONF__MODULES) || defined(CONF_HW_DISPLAY)
+#if 0 /* cannot use `--hw=display' inside twdisplay! */
+	TRY4(display) ||
+# endif
+#endif
+#if defined(CONF__MODULES) || defined(CONF_HW_TTY)
+	AUTOTRY4(tty) ||
+#endif
+#if defined(CONF__MODULES) || defined(CONF_HW_GGI)
+	AUTOTRY4(ggi) ||
 #endif
 
-#if 0 /* cannot use `-hw=display' inside twdisplay! */
-	(check4("display", arg) && (tried++, display_InitHW()) && (fix4("display", D_HW), TRUE)) ||
-#endif
-	
-#ifdef CONF_HW_TTY
-	(check4("tty", arg) && (tried++, tty_InitHW()) && (fix4("tty", D_HW), TRUE)) ||
-#endif
-#ifdef CONF_HW_GGI
-	(check4("ggi", arg) && (tried++, GGI_InitHW()) && (fix4("ggi", D_HW), TRUE)) ||
-#endif
 #ifdef CONF__MODULES
-	module_InitHW() ||
+	module_InitHW(D_HW->Name, D_HW->NameLen) ||
 #endif
 	(warn_NoHW(arg ? D_HW->NameLen - 4 : 0, arg, tried), FALSE);
 
@@ -537,7 +622,7 @@ static byte IsValidHW(uldat len, CONST byte *arg) {
 
 static display_hw AttachDisplayHW(uldat len, CONST byte *arg, uldat slot, byte flags) {
     if ((len && len <= 4) || CmpMem("-hw=", arg, Min2(len,4))) {
-	printk("twdisplay: specified `%.*s\' is not `-hw=<display>\'\n",
+	printk("twdisplay: specified `%.*s\' is not `--hw=<display>\'\n",
 		(int)len, arg);
 	return (display_hw)0;
     }
@@ -710,7 +795,7 @@ static byte ResizeDisplay(void) {
 
     TryDisplayWidth = TryDisplayHeight = 0;
     
-    if (change && (Tmsg = TwCreateMsg(TW_MSG_DISPLAY, sizeof(struct s_tevent_display)))) {
+    if (change && (Tmsg = TwCreateMsg(TW_MSG_DISPLAY, TW_SIZEOF_TEVENT_DISPLAY))) {
 	Tmsg->Event.EventDisplay.Code = TW_DPY_Resize;
 	Tmsg->Event.EventDisplay.Len  = 0;
 	Tmsg->Event.EventDisplay.X    = DisplayWidth;
@@ -732,7 +817,10 @@ void DragAreaHW(dat Left, dat Up, dat Rgt, dat Dwn, dat DstLeft, dat DstUp) {
     
 void SetPaletteHW(udat N, udat R, udat G, udat B) {
     if (N <= MAXCOL) {
-	palette c = {R, G, B};
+	palette c;
+	c.Red = R;
+	c.Green = G;
+	c.Blue = B;
 	if (CmpMem(&Palette[N], &c, sizeof(palette))) {
 	    Palette[N] = c;
 	    HW->SetPalette(N, R, G, B);
@@ -764,7 +852,7 @@ static void HandleMsg(tmsg Msg) {
 	 * Cast it to (obj) as expected by HW displays...
 	 * we will cast it back when needed
 	 */
-	HW->HWSelectionRequest((obj)(Msg->Event.EventSelectionRequest.Requestor),
+	HW->HWSelectionRequest((obj)(topaque)Msg->Event.EventSelectionRequest.Requestor,
 			       Msg->Event.EventSelectionRequest.ReqPrivate);
 	break;
       case TW_MSG_SELECTIONNOTIFY:
@@ -879,11 +967,11 @@ void SelectionExport(void) {
  * treats it as opaque. We will cast it back to (uldat) when needed.
  */
 obj TwinSelectionGetOwner(void) {
-    return (void *)TwGetOwnerSelection();
+    return (obj)(topaque)TwGetOwnerSelection();
 }
 
 /* HW back-end function: set selection owner */
-void TwinSelectionSetOwner(obj Owner, time_t Time, frac_t Frac) {
+void TwinSelectionSetOwner(obj Owner, tany Time, tany Frac) {
     tmsg Msg;
 
     if ((Msg=TwCreateMsg(TW_MSG_SELECTIONCLEAR, sizeof(tevent_common)))) {
@@ -900,7 +988,7 @@ void TwinSelectionNotify(obj Requestor, uldat ReqPrivate, uldat Magic, CONST byt
     printk("twdisplay: Selection Notify to 0x%08x\n", (uldat)Requestor);
 #endif
     /* cast back Requestor from fake (obj) to its original (uldat) */
-    TwNotifySelection((uldat)Requestor, ReqPrivate, Magic, MIME, Len, Data);
+    TwNotifySelection((topaque)Requestor, ReqPrivate, Magic, MIME, Len, Data);
 }
 
 /* HW back-end function: request selection */
@@ -909,17 +997,17 @@ void TwinSelectionRequest(obj Requestor, uldat ReqPrivate, obj Owner) {
     printk("twdisplay: Selection Request from 0x%08x, Owner is 0x%08x\n", (uldat)Requestor, (uldat)Owner);
 #endif
     /* cast back Owner from the fake (obj) to (uldat) */
-    TwRequestSelection((uldat)Owner, ReqPrivate);
+    TwRequestSelection((topaque)Owner, ReqPrivate);
 }
 
-static byte StdAddEventMouse(udat CodeMsg, udat Code, dat MouseX, dat MouseY) {
+static byte StdAddMouseEvent(udat Code, dat MouseX, dat MouseY) {
     tevent_mouse Event;
     tmsg Msg;
 
     if (HW->FlagsHW & FlHWNoInput)
 	return TRUE;
 
-    if ((Msg=TwCreateMsg(CodeMsg, sizeof(event_mouse)))) {
+    if ((Msg=TwCreateMsg(TW_MSG_WIDGET_MOUSE, sizeof(event_mouse)))) {
 	Event = &Msg->Event.EventMouse;
 
 	Event->Code = Code;
@@ -935,7 +1023,7 @@ static byte StdAddEventMouse(udat CodeMsg, udat Code, dat MouseX, dat MouseY) {
 
 byte MouseEventCommon(dat x, dat y, dat dx, dat dy, udat Buttons) {
     dat prev_x, prev_y;
-    udat OldButtons;
+    udat OldButtons, i;
     mouse_state *OldState;
     udat result;
     byte ret = TRUE;
@@ -961,30 +1049,15 @@ byte MouseEventCommon(dat x, dat y, dat dx, dat dy, udat Buttons) {
     
     if (Buttons != OldButtons || ((MouseMotionN || OldButtons) && (x != prev_x || y != prev_y))) {
 	
-	if (MouseMotionN && !OldButtons && (x != prev_x || y != prev_y)) {
-	    if (!StdAddEventMouse(TW_MSG_WIDGET_MOUSE, MOTION_MOUSE, x, y))
-		ret = FALSE;
-	} else if (OldButtons && (x != prev_x || y != prev_y)) {
-	    if (!StdAddEventMouse(TW_MSG_WIDGET_MOUSE, DRAG_MOUSE | OldButtons, x, y))
-		ret = FALSE;
+	if ((MouseMotionN || OldButtons) && (x != prev_x || y != prev_y)) {
+	    ret = StdAddMouseEvent(MOVE_MOUSE | OldButtons, x, y);
 	}
-	if ((Buttons & HOLD_LEFT) != (OldButtons & HOLD_LEFT)) {
-	    result = (Buttons & HOLD_LEFT ? DOWN_LEFT : RELEASE_LEFT) | (OldButtons &= ~HOLD_LEFT);
-	    OldButtons |= Buttons & HOLD_LEFT;
-	    if (!StdAddEventMouse(TW_MSG_WIDGET_MOUSE, result, x, y))
-		ret = FALSE;
-	}
-	if ((Buttons & HOLD_MIDDLE) != (OldButtons & HOLD_MIDDLE)) {
-	    result = (Buttons & HOLD_MIDDLE ? DOWN_MIDDLE : RELEASE_MIDDLE) | (OldButtons &= ~HOLD_MIDDLE);
-	    OldButtons |= Buttons & HOLD_MIDDLE;
-	    if (!StdAddEventMouse(TW_MSG_WIDGET_MOUSE, result, x, y))
-		ret = FALSE;
-	}
-	if ((Buttons & HOLD_RIGHT) != (OldButtons & HOLD_RIGHT)) {
-	    result = (Buttons & HOLD_RIGHT ? DOWN_RIGHT : RELEASE_RIGHT) | (OldButtons &= ~HOLD_RIGHT);
-	    OldButtons |= Buttons & HOLD_RIGHT;
-	    if (!StdAddEventMouse(TW_MSG_WIDGET_MOUSE, result, x, y))
-		ret = FALSE;
+	for (i = 0; i < BUTTON_N_MAX && ret; i++) {
+	    if ((Buttons & HOLD_CODE(i)) != (OldButtons & HOLD_CODE(i))) {
+		result = (Buttons & HOLD_CODE(i) ? PRESS_CODE(i) : RELEASE_CODE(i)) | (OldButtons &= ~HOLD_CODE(i));
+		OldButtons |= Buttons & HOLD_CODE(i);
+		ret = StdAddMouseEvent(result, x, y);
+	    }
 	}
     }
     return ret;
@@ -1094,14 +1167,14 @@ static void MainLoop(int Fd) {
     
     if (num_fds < 0 && errno != EINTR) {
 	QuitDisplayHW(HW);
-	printk("twdisplay: select(): %s\n", strerror(errno));
+	printk("twdisplay: select(): %."STR(SMALLBUFF)"s\n", strerror(errno));
 	exit(1);
     }
     if (TwInPanic()) {
 	err = TwErrno;
 	detail = TwErrnoDetail;
     	QuitDisplayHW(HW);
-	printk("%s: libTw error: %s%s\n", MYname,
+	printk("%."STR(SMALLBUFF)"s: libTw error: %."STR(SMALLBUFF)"s%."STR(SMALLBUFF)"s\n", MYname,
 		TwStrError(err), TwStrErrorDetail(err, detail));
 	exit(1);
     }
@@ -1109,9 +1182,9 @@ static void MainLoop(int Fd) {
     detail = TwErrnoDetail;
     sys_errno = errno;
     QuitDisplayHW(HW);
-    printk("%s: shouldn't happen! Please report:\n"
-	    "\tlibTw TwErrno: %d(%d),\t%s%s\n"
-	    "\tsystem  errno: %d,\t%s\n", MYname,
+    printk("%."STR(SMALLBUFF)"s: shouldn't happen! Please report:\n"
+	    "\tlibTw TwErrno: %d(%d),\t%."STR(SMALLBUFF)"s%."STR(SMALLBUFF)"s\n"
+	    "\tsystem  errno: %d,\t%."STR(SMALLBUFF)"s\n", MYname,
 	    err, detail, TwStrError(err), TwStrErrorDetail(err, detail),
 	    sys_errno, strerror(sys_errno));
     exit(1);
@@ -1124,20 +1197,21 @@ dat GetDisplayHeight(void) {
     return DisplayHeight;
 }
 
-
 static void Usage(void) {
-    fputs("Usage: twdisplay [OPTIONS] -hw=<display> [...]\n"
+    fputs("Usage: twdisplay [OPTIONS]\n"
 	  "Currently known options: \n"
-	  " -h, -help               display this help and exit\n"
-	  " -V, -version            output version information and exit\n"
-	  " -s, -share              allow multiple simultaneous displays (default)\n"
-	  " -x, -excl               request exclusive display - detach all others\n"
-	  " -v                      verbose output (default)\n"
-	  " -q                      quiet - don't report messages from twin server\n"
-	  " -f                      force running even with wrong protocol version\n"
-	  " -twin@<TWDISPLAY>       specify server to contact instead of $TWDISPLAY\n"
-	  " -hw=<display>[,options] start the given display\n"
-	  "Currently known display methods: \n"
+	  " -h, --help               display this help and exit\n"
+	  " -V, --version            output version information and exit\n"
+	  " -s, --share              allow multiple simultaneous displays (default)\n"
+	  " -x, --excl               request exclusive display - detach all others\n"
+	  " -v, --verbose            verbose output (default)\n"
+	  " -q, --quiet              quiet - don't report messages from twin server\n"
+	  " -f, --force              force running even with wrong protocol version\n"
+	  " --twin@<TWDISPLAY>       specify server to contact instead of $TWDISPLAY\n"
+	  " --hw=<display>[,options] start the given display (only one --hw=... allowed)\n"
+	  "                          (default: autoprobe all displays until one succeeds)\n"
+	  "Currently known display drivers: \n"
+	  "\tgfx[@<XDISPLAY>]\n"
 	  "\tX[@<XDISPLAY>]\n"
 	  "\ttwin[@<TWDISPLAY>]\n"
 	  "\ttty[@<tty device>]\n"
@@ -1146,8 +1220,8 @@ static void Usage(void) {
 
 static void TryUsage(CONST char *opt) {
     if (opt)
-	fprintf(stdout, "twdisplay: unknown option `%s'\n", opt);
-    fputs("           try `twdisplay -help' for usage summary.\n", stdout);
+	fprintf(stdout, "twdisplay: unknown option `%."STR(SMALLBUFF)"s'\n", opt);
+    fputs("           try `twdisplay --help' for usage summary.\n", stdout);
 }
 
 static void ShowVersion(void) {
@@ -1159,17 +1233,26 @@ static byte VersionsMatch(byte force) {
     uldat cv = TW_PROTOCOL_VERSION, lv = TwLibraryVersion(), sv = TwServerVersion();
 	
     if (lv != sv || lv != cv) {
-	printk("twdisplay: %s: socket protocol version mismatch!%s\n"
+	printk("twdisplay: %."STR(SMALLBUFF)"s: socket protocol version mismatch!%."STR(SMALLBUFF)"s\n"
 		"           client is %d.%d.%d, library is %d.%d.%d, server is %d.%d.%d\n",
 		(force ? "warning" : "fatal"), (force ? " (ignored)" : ""),
-		TWVER_MAJOR(cv), TWVER_MINOR(cv), TWVER_PATCH(cv),
-		TWVER_MAJOR(lv), TWVER_MINOR(lv), TWVER_PATCH(lv),
-		TWVER_MAJOR(sv), TWVER_MINOR(sv), TWVER_PATCH(sv));
+		TW_VER_MAJOR(cv), TW_VER_MINOR(cv), TW_VER_PATCH(cv),
+		TW_VER_MAJOR(lv), TW_VER_MINOR(lv), TW_VER_PATCH(lv),
+		TW_VER_MAJOR(sv), TW_VER_MINOR(sv), TW_VER_PATCH(sv));
 	return FALSE;
     }
     return TRUE;
 }
 
+static void MergeHyphensArgv(int argc, char **argv) {
+    char *S;
+    while (argc) {
+	if ((S = *argv) && S[0] == '-' && S[1] == '-' && S[2] && S[3])
+	    (*argv)++;
+	argv++, argc--;
+    }
+}
+
 TW_DECL_MAGIC(display_magic);
 
 int main(int argc, char *argv[]) {
@@ -1181,6 +1264,8 @@ int main(int argc, char *argv[]) {
     uldat chunk;
     int Fd;
     
+    MergeHyphensArgv(argc, argv);
+    
     MYname = argv[0];
     
     while (*++argv) {
@@ -1194,17 +1279,17 @@ int main(int argc, char *argv[]) {
 	    flags |= TW_ATTACH_HW_EXCLUSIVE;
 	else if (!strcmp(*argv, "-s") || !strcmp(*argv, "-share"))
 	    flags &= ~TW_ATTACH_HW_EXCLUSIVE;
-	else if (!strcmp(*argv, "-v"))
+	else if (!strcmp(*argv, "-v") || !strcmp(*argv, "-verbose"))
 	    flags |= TW_ATTACH_HW_REDIRECT;
-	else if (!strcmp(*argv, "-q"))
+	else if (!strcmp(*argv, "-q") || !strcmp(*argv, "-quiet"))
 	    flags &= ~TW_ATTACH_HW_REDIRECT;
-	else if (!strcmp(*argv, "-f"))
+	else if (!strcmp(*argv, "-f") || !strcmp(*argv, "-force"))
 	    force = 1;
 	else if (!strncmp(*argv, "-twin@", 6))
 	    dpy = *argv + 6;
 	else if (!strncmp(*argv, "-hw=", 4)) {
 	    if (!strncmp(*argv+4, "display", 7)) {
-		printk("%s: argument `-hw=display' is for internal use only.\n", MYname);
+		printk("%."STR(SMALLBUFF)"s: argument `--hw=display' is for internal use only.\n", MYname);
 		TryUsage(NULL);
 		return 1;
 	    }
@@ -1219,20 +1304,20 @@ int main(int argc, char *argv[]) {
 		else if (*buff == '@' && buff[1]) {
 		    if (buff[1] == '-') {
 			/*
-			 * using server tty makes no sense for twdisplay
+			 * using server controlling tty makes no sense for twdisplay
 			 */
-			printk("%s: `%s' makes sense only with twattach.\n", MYname, *argv);
+			printk("%."STR(SMALLBUFF)"s: `%."STR(SMALLBUFF)"s' makes sense only with twattach.\n", MYname, *argv);
 			return 1;
 		    } else if (tty) {
 			if (!strcmp(buff+1, tty))
 			    /* attach twin to our tty */
 			    ourtty = 1;
 		    } else {
-			printk("%s: ttyname() failed: cannot find controlling tty!\n", MYname);
+			printk("%."STR(SMALLBUFF)"s: ttyname() failed: cannot find controlling tty!\n", MYname);
 			return 1;
 		    }
 		} else {
-		    printk("%s: malformed display hw `%s'\n", MYname, *argv);
+		    printk("%."STR(SMALLBUFF)"s: malformed display hw `%."STR(SMALLBUFF)"s'\n", MYname, *argv);
 		    return 1;
 		}
 
@@ -1261,8 +1346,8 @@ int main(int argc, char *argv[]) {
     }
     
     if (!arg) {
-	Usage();
-	return 1;
+	/* if user did not specify any `--hw=<dpy>', autoprobe */
+	arg = "";
     }
 
 #ifdef CONF__ALLOC
@@ -1273,9 +1358,11 @@ int main(int argc, char *argv[]) {
     }
 #endif
     
-    TWDisplay = origTWDisplay = CloneStr(dpy ? dpy : (byte *)getenv("TWDISPLAY"));
+    origTWDisplay = CloneStr(getenv("TWDISPLAY"));
+    TWDisplay = dpy ? CloneStr(dpy) : origTWDisplay;
     origTERM = CloneStr(getenv("TERM"));
-
+    HOME = CloneStr(getenv("HOME"));
+    
     InitSignals();
     InitTtysave();
     
@@ -1300,7 +1387,7 @@ int main(int argc, char *argv[]) {
 	    return 1;
 	}
 	    
-	if (!(TMsgPort = TwCreateMsgPort(9, "twdisplay", 0, 0, 0)))
+	if (!(TMsgPort = TwCreateMsgPort(9, "twdisplay")))
 	    break;
 
 	DisplayWidth = TryDisplayWidth = TwGetDisplayWidth();
@@ -1327,7 +1414,7 @@ int main(int argc, char *argv[]) {
 	flags &= TW_ATTACH_HW_REDIRECT;
 	
 	if (flags)
-	    printk("reported messages ...\n");
+	    printk("messages reported by twin server...\n");
 	
 	for (;;) {
 	    buff = TwAttachGetReply(&chunk);
@@ -1374,12 +1461,12 @@ int main(int argc, char *argv[]) {
 	     */
 	    MainLoop(Fd);
 	else if (ret)
-	    printk("%s: twin said we can quit... strange!\n", MYname);
+	    printk("%."STR(SMALLBUFF)"s: twin said we can quit... strange!\n", MYname);
 	
 	Quit(!ret);
     } while (0);
     
-    printk("%s: libTw error: %s%s\n", MYname,
+    printk("%."STR(SMALLBUFF)"s: libTw error: %."STR(SMALLBUFF)"s%."STR(SMALLBUFF)"s\n", MYname,
 	    TwStrError(TwErrno), TwStrErrorDetail(TwErrno, TwErrnoDetail));
     return 1;
 }
