@@ -26,7 +26,7 @@
 #include "methods.h"
 #include "draw.h"
 #include "remote.h"
-#include "extensions.h"
+#include "extreg.h"
 
 #include "resize.h"
 #include "hw.h"
@@ -39,40 +39,26 @@
 # include "dl.h"
 #endif
 
-#include "Tw/Tw.h"
+#include <Tw/Tw.h>
 
 
 /* HW specific headers */
 
+#ifdef CONF_HW_GFX
+# include "HW/hw_gfx.h"
+#endif
+
 #ifdef CONF_HW_X11
 # include "HW/hw_X11.h"
 #endif
 
-#if defined(CONF__MODULES) || defined(CONF_SOCKET)
-
-# ifdef CONF_HW_DISPLAY
-#  include "HW/hw_display.h"
-# endif
-
-# ifdef CONF_HW_TWIN
-#  include "HW/hw_twin.h"
-# endif
-
-#else
+#if !defined(CONF__MODULES) && !defined(CONF_SOCKET)
 
 # undef CONF_HW_DISPLAY
 # undef CONF_HW_TWIN
 
 #endif
 
-#ifdef CONF_HW_TTY
-# include "HW/hw_tty.h"
-#endif
-
-#ifdef CONF_HW_GGI
-# include "HW/hw_ggi.h"
-#endif
-
 
 
 
@@ -94,7 +80,7 @@ static dat TryDisplayWidth, TryDisplayHeight;
 
 static dat AccelVideo[4] = { MAXDAT, MAXDAT, MINDAT, MINDAT };
 byte   StrategyFlag;
-frac_t StrategyDelay = (frac_t)0;
+tany   StrategyDelay = (tany)0;
 
 static udat ConfigureHWValue[HW_CONFIGURE_MAX];
 static byte ConfigureHWDefault[HW_CONFIGURE_MAX];
@@ -114,6 +100,8 @@ udat GetDisplayHeight(void) {
 }
 
 void UpdateFlagsHW(void) {
+    StrategyReset(); /* reset StrategyFlag */
+    
     NeedOldVideo = ExpensiveFlushVideo = FALSE;
     CanDragArea = TRUE;
 
@@ -169,36 +157,35 @@ void RunNoHW(byte print_info) {
 }
 
 
-void warn_NoHW(uldat len, char *arg, uldat tried) {
+static void warn_NoHW(uldat len, char *arg, uldat tried) {
 #ifdef CONF__MODULES
     if (!tried && !arg)
 	printk("twin: no display driver compiled into twin.\n"
-	       "      please run as `twin -hw=<display>'\n");
+	       "      please run as `twin --hw=<display>'\n");
     else
 #endif
     {
-	printk("twin: All display drivers failed");
+	printk("twin: all display drivers failed");
 	if (arg)
-	    printk(" for `-hw=%.*s\'", (int)len, arg);
+	    printk(" for `--hw=%.*s\'\n", Min2((int)len,SMALLBUFF), arg);
 	else
-	    printk(".");
-	printk("\n");
+	    printk(".\n");
     }
 }
 
 #ifdef CONF__MODULES
 
-static byte module_InitHW(void) {
+static byte module_InitHW(byte *arg, uldat len) {
     byte *name, *tmp;
     byte *(*InitD)(void);
-    byte *arg = HW->Name;
-    uldat len = HW->NameLen;
     module Module;
 
-    if (!arg || len <= 4)
+    if (!arg || !len)
 	return FALSE;
     
-    arg += 4; len -= 4; /* skip "-hw=" */
+    if (len >= 4 && !memcmp(arg, "-hw=", 4)) {
+	arg += 4; len -= 4; /* skip "-hw=" */
+    }
     
     name = memchr(arg, '@', len);
     tmp = memchr(arg, ',', len);
@@ -207,48 +194,124 @@ static byte module_InitHW(void) {
     if (name)
 	len = name - arg;
     
-    if ((name = AllocMem(len + 10))) {
-	sprintf(name, "HW/hw_%.*s.so", (int)len, arg);
-			
-	Module = DlLoadAny(len+9, name);
+    if ((name = AllocMem(len + 7))) {
+	sprintf(name, "HW/hw_%.*s", (int)len, arg);
+
+	Module = DlLoadAny(len + 6, name);
 	
 	if (Module) {
-	    printk("twin: starting display driver module `HW/hw_%.*s.so'...\n", (int)len, arg);
-	    
+	    printk("twin: starting display driver module `%."STR(SMALLBUFF)"s'...\n", name);
 	    if ((InitD = Module->Private) && InitD()) {
-		printk("twin: ...module `%s' successfully started.\n", name);
+		printk("twin: ...module `%."STR(SMALLBUFF)"s' successfully started.\n", name);
 		FreeMem(name);
 		HW->Module = Module; Module->Used++;
 		return TRUE;
 	    }
 	    Delete(Module);
 	}
-    }
-
-    if (Module) {
-	printk("twin: ...module `%s' failed to start.\n", name);
     } else
-	printk("twin: unable to load display driver module `%s' :\n"
-			"      %s\n", name, ErrStr);
+	Error(NOMEMORY);
+    
+    if (Module) {
+	printk("twin: ...module `%."STR(SMALLBUFF)"s' failed to start.\n", name ? name : (byte *)"(NULL)");
+    } else
+	printk("twin: unable to load display driver module `%."STR(SMALLBUFF)"s' :\n"
+	       "      %."STR(SMALLBUFF)"s\n", name ? name : (byte *)"(NULL)", ErrStr);
+    if (name)
+	FreeMem(name);
     
-    FreeMem(name);
     return FALSE;
 }
 
 #endif /* CONF__MODULES */
 
-#if defined(CONF_HW_X11) || defined(CONF_HW_TWIN) || defined(CONF_HW_DISPLAY) || defined(CONF_HW_TTY) || defined(CONF_HW_GGI)
+
+#ifdef CONF__MODULES
+# define DEF_INITHW(hw) \
+static byte CAT(hw,_InitHW)(void) { \
+    byte *arg; \
+    uldat len; \
+    if (HW->Name && HW->NameLen) { \
+	arg = HW->Name; \
+	len = HW->NameLen; \
+    } else { \
+	arg = "-hw=" STR(hw); \
+	len = strlen(arg); \
+    } \
+    return module_InitHW(arg, len); \
+}
+#else
+# define DEF_INITHW(hw)
+#endif
+
+
+/* HW specific functions */
+
+#ifdef CONF_HW_GFX
+# include "HW/hw_gfx.h"
+#else
+  DEF_INITHW(gfx)
+#endif
+
+#ifdef CONF_HW_X11
+# include "HW/hw_X11.h"
+# define X_InitHW X11_InitHW
+#else
+  DEF_INITHW(X)
+#endif
+	
+#ifdef CONF_HW_TWIN
+# include "HW/hw_twin.h"
+# define twin_InitHW TW_InitHW
+#else
+  DEF_INITHW(twin)
+#endif
+
+#ifdef CONF_HW_DISPLAY
+# include "HW/hw_display.h"
+#else
+  DEF_INITHW(display)
+#endif
+
+#ifdef CONF_HW_TTY
+# include "HW/hw_tty.h"
+#else
+  DEF_INITHW(tty)
+#endif
+
+#ifdef CONF_HW_GGI
+# include "HW/hw_ggi.h"
+# define ggi_InitHW GGI_InitHW
+#else
+  DEF_INITHW(ggi)
+#endif
+
+#undef DEF_INITHW
+
+
+#if defined(CONF__MODULES) || defined(CONF_HW_DISPLAY)
 static byte check4(byte *s, byte *arg) {
-    if (arg && strncmp(s, arg, strlen(s))) {
+    if (arg && !strncmp(s, arg, strlen(s))) {
+	printk("twin: trying given `--hw=%."STR(SMALLBUFF)"s' display driver.\n", s);
+	return TRUE;
+    }
+    return FALSE;
+}
+#endif /* defined(CONF__MODULES) || defined(CONF_HW_DISPLAY) */
+
+static byte autocheck4(byte *s, byte *arg) {
+    if (arg && !strncmp(s, arg, strlen(s))) {
+	printk("twin: trying given `--hw=%."STR(SMALLBUFF)"s' display driver.\n", s);
+	return TRUE;
+    }
+    if (arg) {
 	/*
-	printk("twin: `-hw=%s' given, skipping `-hw=%s' display driver.\n",
+	printk("twin: `-hw=%."STR(SMALLBUFF)"s' given, skipping `-hw=%."STR(SMALLBUFF)"s' display driver.\n",
 		arg, s);
 	 */
 	return FALSE;
-    } else if (arg)
-	printk("twin: trying given `-hw=%s' display driver.\n", s);
-    else
-	printk("twin: autoprobing `-hw=%s' display driver.\n", s);
+    }
+    printk("twin: autoprobing `--hw=%."STR(SMALLBUFF)"s' display driver.\n", s);
     return TRUE;
 }
 
@@ -264,7 +327,6 @@ static void fix4(byte *s, display_hw D_HW) {
 	}
     }
 }
-#endif /* defined(CONF_HW_X11) || defined(CONF_HW_TWIN) || defined(CONF_HW_DISPLAY) || defined(CONF_HW_TTY) || defined(CONF_HW_GGI) */
 
 /*
  * InitDisplayHW runs HW specific InitXXX() functions, starting from best setup
@@ -285,27 +347,36 @@ byte InitDisplayHW(display_hw D_HW) {
     else
 	arg = NULL;
 
+#define AUTOTRY4(hw) (autocheck4(STR(hw), arg) && (tried++, CAT(hw,_InitHW)()) && (fix4(STR(hw), D_HW), TRUE))
+#define     TRY4(hw) (    check4(STR(hw), arg) && (tried++, CAT(hw,_InitHW)()) && (fix4(STR(hw), D_HW), TRUE))
+    
     success =
-#ifdef CONF_HW_X11
-	(check4("X", arg) && (tried++, X11_InitHW()) && (fix4("X", D_HW), TRUE)) ||
+#if defined(CONF__MODULES) || defined(CONF_HW_GFX)
+	AUTOTRY4(gfx) ||
 #endif
-#ifdef CONF_HW_TWIN
-	(check4("twin", arg) && (tried++, TW_InitHW()) && (fix4("twin", D_HW), TRUE)) ||
+#if defined(CONF__MODULES) || defined(CONF_HW_X11)
+	AUTOTRY4(X) ||
 #endif
-#ifdef CONF_HW_DISPLAY
-	(check4("display", arg) && (tried++, display_InitHW()) && (fix4("display", D_HW), TRUE)) ||
+#if defined(CONF__MODULES) || defined(CONF_HW_TWIN)
+	AUTOTRY4(twin) ||
 #endif
-#ifdef CONF_HW_TTY
-	(check4("tty", arg) && (tried++, tty_InitHW()) && (fix4("tty", D_HW), TRUE)) ||
+#if defined(CONF__MODULES) || defined(CONF_HW_DISPLAY)
+	TRY4(display) ||
 #endif
-#ifdef CONF_HW_GGI
-	(check4("ggi", arg) && (tried++, GGI_InitHW()) && (fix4("ggi", D_HW), TRUE)) ||
+#if defined(CONF__MODULES) || defined(CONF_HW_TTY)
+	AUTOTRY4(tty) ||
+#endif
+#if defined(CONF__MODULES) || defined(CONF_HW_GGI)
+	AUTOTRY4(ggi) ||
 #endif
 #ifdef CONF__MODULES
-	module_InitHW() ||
+	module_InitHW(D_HW->Name, D_HW->NameLen) ||
 #endif
 	(warn_NoHW(arg ? D_HW->NameLen - 4 : 0, arg, tried), FALSE);
 
+#undef     TRY4
+#undef AUTOTRY4
+
     if (success) {
 	D_HW->Quitted = FALSE;
 	
@@ -360,7 +431,7 @@ void QuitDisplayHW(display_hw D_HW) {
 static byte IsValidHW(uldat len, CONST byte *arg) {
     CONST byte *slash = memchr(arg, '/', len), *at = memchr(arg, '@', len), *comma = memchr(arg, ',', len);
     if (slash && (!at || slash < at) && (!comma || slash < comma)) {
-	printk("twin: slash ('/') not allowed in display HW name: %.*s\n", (int)len, arg);
+	printk("twin: slash ('/') not allowed in display HW name: %.*s\n", Min2((int)len,SMALLBUFF), arg);
 	return FALSE;
     }
     return TRUE;
@@ -371,8 +442,8 @@ display_hw AttachDisplayHW(uldat len, CONST byte *arg, uldat slot, byte flags) {
 
     if ((len && len <= 4) || CmpMem("-hw=", arg, Min2(len,4))) {
 	printk("twin: specified `%.*s\' is not a known option.\n"
-		"      try `twin -help' for usage summary.\n",
-		(int)len, arg);
+		"      try `twin --help' for usage summary.\n",
+	       Min2((int)len,SMALLBUFF), arg);
 	return NULL;
     }
     
@@ -436,35 +507,49 @@ byte DetachDisplayHW(uldat len, CONST byte *arg, byte flags) {
     return done;
 }
 
+/* initialize all required HW displays. Since we are at it, also parse command line */
 byte InitHW(void) {
     byte **arglist = orig_argv;
-    byte ret = FALSE;
-    byte flags = 0;
-    int hwcount = 0;
+    byte ret = FALSE, flags = 0, nohw = FALSE;
+    udat hwcount = 0;
     
     WriteMem(ConfigureHWDefault, '\1', HW_CONFIGURE_MAX); /* set everything to default (-1) */
     
     for (arglist = orig_argv; *arglist; arglist++) {
-	if (!strcmp(*arglist, "-nohw")) {
-	    if (hwcount > 0) {
-		printk("twin: `-hw=' and `-nohw' options used together. make up your mind.\n");
-		return FALSE;
-	    }
-	    hwcount = -1;
-	} else if (!strcmp(*arglist, "-x")) {
-	    flags = TW_ATTACH_HW_EXCLUSIVE;
-	} else if (!strcmp(*arglist, "-secure")) {
+	if (!strcmp(*arglist, "-nohw"))
+	    nohw = TRUE;
+	else if (!strcmp(*arglist, "-x") || !strcmp(*arglist, "-excl"))
+	    flags |= TW_ATTACH_HW_EXCLUSIVE;
+	else if (!strcmp(*arglist, "-s") || !strcmp(*arglist, "-share"))
+	    flags &= ~TW_ATTACH_HW_EXCLUSIVE;
+	else if (!strcmp(*arglist, "-secure"))
 	    flag_secure = TRUE;
-	} else if (!strncmp(*arglist, "-hw=", 4)) {
-	    if (hwcount == -1) {
-		printk("twin: `-hw=' and `-nohw' options used together. make up your mind.\n");
-		return FALSE;
-	    } else if (hwcount > 0 && (flags & TW_ATTACH_HW_EXCLUSIVE)) {
-		printk("twin: `-x' (exclusive) used with multiple `-hw='. make up your mind.\n");
-		return FALSE;
-	    }
+	else if (!strcmp(*arglist, "-envrc"))
+	    flag_envrc = TRUE;
+	else if (!strncmp(*arglist, "-hw=", 4))
+	    hwcount++;
+	else
+	    printk("twin: ignoring unknown option `%."STR(SMALLBUFF)"s'\n", *arglist);
+    }
+
+    if (nohw && hwcount > 0) {
+	printk("twin: `--hw=' and `--nohw' options used together. make up your mind.\n");
+	return ret;
+    }
+    if (flags & TW_ATTACH_HW_EXCLUSIVE) {
+	if (hwcount == 0 || hwcount > 1) {
+	    printk("twin: `--excl' used with%s `--hw='. make up your mind.\n",
+		   hwcount == 0 ? "out" : " multiple");
+	    return ret;
 	}
     }
+
+    /*
+     * execute .twenvrc.sh if needed and read its output to set
+     * environment variables (mostly useful for twdm)
+     */
+    RunTwEnvRC();
+
     for (arglist = orig_argv; *arglist; arglist++) {
 	if (!strcmp(*arglist, "-nohw")) {
 	    RunNoHW(ret = TRUE);
@@ -646,7 +731,11 @@ void ConfigureHW(udat resource, byte todefault, udat value) {
 
 void SetPaletteHW(udat N, udat R, udat G, udat B) {
     if (N <= MAXCOL) {
-	palette c = {R, G, B};
+	palette c;
+	c.Red = R;
+	c.Green = G;
+	c.Blue = B;
+	
 	if (CmpMem(&Palette[N], &c, sizeof(palette))) {
 	    Palette[N] = c;
 	    forHW {
@@ -683,12 +772,12 @@ obj TwinSelectionGetOwner(void) {
 static void SelectionClear(msgport Owner) {
     msg Msg;
     
-    if ((Msg = Do(Create,Msg)(FnMsg, MSG_SELECTIONCLEAR, sizeof(event_common))))
+    if ((Msg = Do(Create,Msg)(FnMsg, MSG_SELECTIONCLEAR, 0)))
 	SendMsg(Owner, Msg);
 }
 
 /* HW back-end function: set selection owner */
-void TwinSelectionSetOwner(obj Owner, time_t Time, frac_t Frac) {
+void TwinSelectionSetOwner(obj Owner, tany Time, tany Frac) {
     timevalue T;
     if (Time == SEL_CURRENTTIME)
 	CopyMem(&All->Now, &T, sizeof(timevalue));
@@ -727,7 +816,7 @@ void TwinSelectionNotify(obj Requestor, uldat ReqPrivate, uldat Magic, CONST byt
 	if (!Data)
 	    Len = 0;
 
-	if ((NewMsg = Do(Create,Msg)(FnMsg, MSG_SELECTIONNOTIFY, sizeof(event_selectionnotify) + Len))) {
+	if ((NewMsg = Do(Create,Msg)(FnMsg, MSG_SELECTIONNOTIFY, Len))) {
 	    Event = &NewMsg->Event;
 	    Event->EventSelectionNotify.W = NULL;
 	    Event->EventSelectionNotify.Code = 0;
@@ -759,7 +848,7 @@ void TwinSelectionRequest(obj Requestor, uldat ReqPrivate, obj Owner) {
 	if (Owner->Id >> magic_shift == msgport_magic >> magic_shift) {
 	    msg NewMsg;
 	    event_any *Event;
-	    if ((NewMsg = Do(Create,Msg)(FnMsg, MSG_SELECTIONREQUEST, sizeof(event_selectionrequest)))) {
+	    if ((NewMsg = Do(Create,Msg)(FnMsg, MSG_SELECTIONREQUEST, 0))) {
 
 		Event = &NewMsg->Event;
 		Event->EventSelectionRequest.W = NULL;
@@ -883,8 +972,8 @@ INLINE void SyncOldVideo(void) {
 #define MaxRecentBeepHW ((byte)30)
 
 void FlushHW(void) {
-    static timevalue LastBeep = {(time_t)0, (frac_t)0};
-    timevalue tmp = {(time_t)0, 100 MilliSECs};
+    static timevalue LastBeep = {(tany)0, (tany)0};
+    timevalue tmp = {(tany)0, 100 MilliSECs};
     byte doBeep = FALSE, saved = FALSE, mangled = FALSE;
     byte saveChangedVideoFlag, saveValidOldVideo;
     /*
@@ -977,7 +1066,7 @@ void SyntheticKey(widget W, udat Code, udat ShiftFlags, byte Len, byte *Seq) {
     msg Msg;
 
     if (W && Len && Seq &&
-	(Msg=Do(Create,Msg)(FnMsg, MSG_WIDGET_KEY, Len + sizeof(event_keyboard)))) {
+	(Msg=Do(Create,Msg)(FnMsg, MSG_WIDGET_KEY, Len))) {
 	
 	Event = &Msg->Event.EventKeyboard;
 	Event->W = W;
@@ -1137,7 +1226,7 @@ void EnableMouseMotionEvents(byte enable) {
 
 byte MouseEventCommon(dat x, dat y, dat dx, dat dy, udat Buttons) {
     dat prev_x, prev_y;
-    udat OldButtons;
+    udat OldButtons, i;
     mouse_state *OldState;
     udat result;
     byte ret = TRUE;
@@ -1160,60 +1249,45 @@ byte MouseEventCommon(dat x, dat y, dat dx, dat dy, udat Buttons) {
     OldState->x = x;
     OldState->y = y;
     
-    OldState->keys = Buttons;
+    OldState->keys = Buttons &= HOLD_ANY;
 
     /* keep it available */
     All->MouseHW = HW;
 
     if (Buttons != OldButtons || ((alsoMotionEvents || OldButtons) && (x != prev_x || y != prev_y))) {
 	
-	if (alsoMotionEvents && !OldButtons && (x != prev_x || y != prev_y)) {
-	    if (!StdAddEventMouse(MSG_MOUSE, MOTION_MOUSE, x, y))
-		ret = FALSE;
-	} else if (OldButtons && (x != prev_x || y != prev_y)) {
-	    if (!StdAddEventMouse(MSG_MOUSE, DRAG_MOUSE | OldButtons, x, y))
-		ret = FALSE;
+	if ((alsoMotionEvents || OldButtons) && (x != prev_x || y != prev_y)) {
+	    ret = StdAddMouseEvent(MOVE_MOUSE | OldButtons, x, y);
 	}
-	if ((Buttons & HOLD_LEFT) != (OldButtons & HOLD_LEFT)) {
-	    result = (Buttons & HOLD_LEFT ? DOWN_LEFT : RELEASE_LEFT) | (OldButtons &= ~HOLD_LEFT);
-	    OldButtons |= Buttons & HOLD_LEFT;
-	    if (!StdAddEventMouse(MSG_MOUSE, result, x, y))
-		ret = FALSE;
+	for (i = 0; i < BUTTON_N_MAX && ret; i++) {
+	    if ((Buttons & HOLD_CODE(i)) != (OldButtons & HOLD_CODE(i))) {
+		result = (Buttons & HOLD_CODE(i) ? PRESS_CODE(i) : RELEASE_CODE(i)) | (OldButtons &= ~HOLD_CODE(i));
+		OldButtons |= Buttons & HOLD_CODE(i);
+		ret = StdAddMouseEvent(result, x, y);
+	    }
 	}
-	if ((Buttons & HOLD_MIDDLE) != (OldButtons & HOLD_MIDDLE)) {
-	    result = (Buttons & HOLD_MIDDLE ? DOWN_MIDDLE : RELEASE_MIDDLE) | (OldButtons &= ~HOLD_MIDDLE);
-	    OldButtons |= Buttons & HOLD_MIDDLE;
-	    if (!StdAddEventMouse(MSG_MOUSE, result, x, y))
-		ret = FALSE;
-	}
-	if ((Buttons & HOLD_RIGHT) != (OldButtons & HOLD_RIGHT)) {
-	    result = (Buttons & HOLD_RIGHT ? DOWN_RIGHT : RELEASE_RIGHT) | (OldButtons &= ~HOLD_RIGHT);
-	    OldButtons |= Buttons & HOLD_RIGHT;
-	    if (!StdAddEventMouse(MSG_MOUSE, result, x, y))
-		ret = FALSE;
-	}
-
     }
     return ret;
 }
 
-byte StdAddEventMouse(udat CodeMsg, udat Code, dat MouseX, dat MouseY) {
+byte StdAddMouseEvent(udat Code, dat MouseX, dat MouseY) {
     msg Msg;
     event_mouse *Event;
 
     if (HW && HW == All->MouseHW && HW->FlagsHW & FlHWNoInput)
 	return TRUE;
     
-    if ((Code & ANY_ACTION_MOUSE)==DRAG_MOUSE
+    if ((Code & MOUSE_ACTION_ANY) == MOVE_MOUSE
 	&& (Msg = Ext(WM,MsgPort)->LastMsg)
 	&& Msg->Type==MSG_MOUSE
 	&& (Event=&Msg->Event.EventMouse)
 	&& Event->Code==Code) {
+	/* merge the two events */
 	Event->X=MouseX;
 	Event->Y=MouseY;
 	return TRUE;
     }
-    if ((Msg=Do(Create,Msg)(FnMsg, CodeMsg, sizeof(event_mouse)))) {
+    if ((Msg=Do(Create,Msg)(FnMsg, MSG_MOUSE, 0))) {
 	Event=&Msg->Event.EventMouse;
 	Event->Code=Code;
 	Event->ShiftFlags=(udat)0;
@@ -1232,7 +1306,7 @@ byte KeyboardEventCommon(udat Code, udat ShiftFlags, udat Len, CONST byte *Seq)
     if (HW->FlagsHW & FlHWNoInput)
 	return TRUE;
 
-    if ((Msg=Do(Create,Msg)(FnMsg, MSG_KEY, Len + sizeof(event_keyboard)))) {
+    if ((Msg=Do(Create,Msg)(FnMsg, MSG_KEY, Len))) {
 	Event = &Msg->Event.EventKeyboard;
 	    
 	Event->Code = Code;
