@@ -7,7 +7,7 @@
  */
 
 extern byte StrategyFlag;
-extern frac_t StrategyDelay;
+extern tany StrategyDelay;
 /* strategy */
 #define HW_UNSET  0
 #define HW_ACCEL  1
@@ -38,7 +38,7 @@ void UpdateFlagsHW(void);
 
 void EnableMouseMotionEvents(byte enable);
 
-byte StdAddEventMouse(udat CodeMsg, udat Code, dat MouseX, dat MouseY);
+byte StdAddMouseEvent(udat Code, dat MouseX, dat MouseY);
 void SyntheticKey(widget W, udat Code, udat ShiftFlags, byte Len, byte *Seq);
 
 #endif /* _TWIN_HW_MULTI_H */
