@@ -21,17 +21,20 @@ byte InitRCOptions(void) {
 #ifdef CONF_OPT_BLINK
 	| SETUP_BLINK
 #endif
-#ifdef CONF_OPT_ALWAYSCURSOR
-	| SETUP_ALWAYSCURSOR
+#ifdef CONF_OPT_CURSOR_ALWAYS
+	| SETUP_CURSOR_ALWAYS
 #endif
-#ifdef CONF_OPT_HIDEMENU
-	| SETUP_HIDEMENU
+#ifdef CONF_OPT_MENU_HIDE
+	| SETUP_MENU_HIDE
 #endif
-#ifdef CONF_OPT_MENUINFO
-	| SETUP_MENUINFO
+#ifdef CONF_OPT_MENU_RELAX
+	| SETUP_MENU_RELAX
 #endif
-#ifdef CONF_OPT_EDGESCROLL
-	| SETUP_EDGESCROLL
+#ifdef CONF_OPT_MENU_INFO
+	| SETUP_MENU_INFO
+#endif
+#ifdef CONF_OPT_SCREEN_SCROLL
+	| SETUP_SCREEN_SCROLL
 #endif
 	;
 
