@@ -19,7 +19,7 @@
 #include "util.h"
 #include "wm.h"
 
-#include "extensions.h"
+#include "extreg.h"
 #include "methods.h"
 #include "hw.h"
 #include "common.h"
@@ -36,9 +36,17 @@
 # include "dl.h"
 #endif
 
-#include "Tw/Twkeys.h"
+#include <Tw/Twkeys.h>
 #include "hotkey.h"
 
+#ifdef CONF__UNICODE
+# include <Tutf/Tutf.h>
+# include <Tutf/Tutf_defs.h>
+#endif
+
+
+
+
 #define MAX_RUNCYCLE      1024 /* kill a queue after this number of steps */
 
 #define MAX_RUNSTACK 28
@@ -178,23 +186,29 @@ static run *ReverseRunList(run *l) {
 
 static ldat Pos2Ctx(udat Pos) {
     switch (Pos) {
-      case POS_TITLE: return CTX_TITLE;
-      case POS_CORNER:return CTX_CORNER;
-      case POS_SIDE:  return CTX_SIDE;
+      case POS_TITLE:
+      case POS_SIDE_UP:	return CTX_TITLE;
+      case POS_SIDE_LEFT:
+      case POS_SIDE_RIGHT:
+      case POS_SIDE_DOWN:return CTX_SIDE;
+      case POS_BUTTON_RESIZE:return CTX_CORNER;
 
-      case POS_BAR_BACK:
-      case POS_BAR_FWD:
-      case POS_TAB:
-      case POS_ARROW_BACK:
-      case POS_ARROW_FWD:
-	return CTX_BARS;
+      case POS_X_BAR_BACK:
+      case POS_X_BAR_FWD:
+      case POS_X_TAB:
+      case POS_X_ARROW_BACK:
+      case POS_X_ARROW_FWD:
+      case POS_Y_BAR_BACK:
+      case POS_Y_BAR_FWD:
+      case POS_Y_TAB:
+      case POS_Y_ARROW_BACK:
+      case POS_Y_ARROW_FWD:return CTX_BARS;
 	
       case POS_INSIDE:return CTX_INSIDE;
       case POS_MENU:  return CTX_MENU;
 
-      case POS_SCREENBUTTON:
-      case POS_ROOT:
-	return CTX_ROOT;
+      case POS_BUTTON_SCREEN:
+      case POS_ROOT:  return CTX_ROOT;
       default:
 	if (Pos < BUTTON_MAX)
 	    return CTX_BUTTON(Pos);
@@ -213,7 +227,7 @@ static node RCFindKeyBind(ldat label, ldat shiftflags) {
 
 static node RCFindMouseBind(ldat code, ldat ctx) {
     node l = MouseList;
-    ldat hc = code & (HOLD_CODE(3) | HOLD_CODE(4));
+    ldat hc = code & (PRESS_|RELEASE_);
     code &= HOLD_ANY;
     
     for (; l; l = l->next) {
@@ -396,7 +410,8 @@ static byte RCSteps(run *r) {
     ldat flag;
     str *argv;
     byte state, ret;
-
+    int nfd;
+    
     W = RCCheck4WidgetId(r);
     S = ScreenOf(W);
     C = &r->C;
@@ -414,6 +429,11 @@ static byte RCSteps(run *r) {
 	      case -1: /* error */
 		break;
 	      case 0:  /* child */
+		nfd = open("/dev/null", O_RDWR);
+		close(0); close(1); close(2);
+		if (nfd >= 0) {
+		    dup2(nfd, 0); dup2(nfd, 1); dup2(nfd, 2);
+		}
 		execvp((char *)n->x.v.argv[0], (char **)n->x.v.argv);
 		exit(1);
 		break;
@@ -507,15 +527,15 @@ static byte RCSteps(run *r) {
 	    }
 	    break;
 	  case SLEEP:
-	    r->SW.WakeUp.Seconds = n->x.f.a;
-	    r->SW.WakeUp.Fraction = (frac_t)0;
+	    r->SW.WakeUp.Seconds = n->x.f.a / 1000;
+	    r->SW.WakeUp.Fraction = n->x.f.a % 1000;
 	    IncrTime(&r->SW.WakeUp, &All->Now);
 	    state = Ssleep;
 	    break;
 	  case STDERR:
 	    argv = n->x.v.argv;
 	    while (*argv)
-		printk("%s ", *argv++);
+		printk("%."STR(SMALLBUFF)"s ", *argv++);
 	    printk("\n");
 	    break;
 	  case SYNTHETICKEY:
@@ -642,18 +662,18 @@ static byte RCSteps(run *r) {
 	    break;
 	  case LOWER:
 	    if (W && S)
-		MakeLastWidget(W, FALSE);
+		LowerWidget(W, FALSE);
 	    break;
 	  case RAISE:
 	    if (W && S)
-		MakeFirstWidget(W, FALSE);
+		RaiseWidget(W, FALSE);
 	    break;
 	  case RAISELOWER:
 	    if (W && S) {
 		if ((widget)W == S->FirstW)
-		    MakeLastWidget(W, TRUE);
+		    LowerWidget(W, TRUE);
 		else
-		    MakeFirstWidget(W, TRUE);
+		    RaiseWidget(W, TRUE);
 	    }
 	    break;
 	  case ROLL:
@@ -745,8 +765,8 @@ static byte RCSleep(timevalue *_t) {
     timevalue *t = _t;
     run *r = Sleep;
 
-    t->Seconds = MAXTIME_T;
-    t->Fraction = (frac_t)0; /* not MAXFRAC_T as Normalize() would overflow */
+    t->Seconds = MAXTANY;
+    t->Fraction = (tany)0; /* not MAXTANY as Normalize() would overflow */
 
     while (r) {
 	if (CmpTime(&r->SW.WakeUp, t) < 0)
@@ -787,7 +807,7 @@ static void RCReload(void) {
     /* this would garble -hw=tty display */
     else
 	printk("twin: failed to load the RC parser:\n"
-		"      %s\n", ErrStr);
+		"      %."STR(SMALLBUFF)"s\n", ErrStr);
 # endif
 #endif
     
@@ -817,7 +837,7 @@ static void RCReload(void) {
 
 	FillButtonWin();
 	UpdateOptionWin();
-	HideMenu(!!(All->SetUp->Flags & SETUP_HIDEMENU));
+	HideMenu(!!(All->SetUp->Flags & SETUP_MENU_HIDE));
     }
 }
 
@@ -898,14 +918,14 @@ byte RC_VMQueue(CONST wm_ctx *C) {
 	    ctx = Pos2Ctx(C->Pos);
 	    
 	    if (isSINGLE_PRESS(C->Code)) {
-		Code = (C->Code & HOLD_ANY) | HOLD_CODE(3);
+		Code = HOLD_CODE(PRESS_N(C->Code)) | PRESS_;
 		n = RCFindMouseBind((ldat)Code, ctx);
 		if (n) {
 		    ClickWinId = C->W ? C->W->Id : NOID;
 		    n = n->body;
 		}
 	    } else if (isSINGLE_RELEASE(C->Code)) {
-		Code = HOLD_CODE(RELEASE_N(C->Code)) | HOLD_CODE(4);
+		Code = HOLD_CODE(RELEASE_N(C->Code)) | RELEASE_;
 		n = RCFindMouseBind((ldat)Code, ctx);
 		if (n && MouseClickReleaseSameCtx(ClickWinId, C->W ? C->W->Id : NOID,
 						  Pos2Ctx(ClickWindowPos), ctx, n->x.ctx))
@@ -971,6 +991,7 @@ byte RC_VM(timevalue *t) {
 void QuitRC(void) {
     ResetBorderPattern();
     RCKillAll();
+    shm_quit();
 }
 
 
@@ -1088,35 +1109,40 @@ byte InitRC(void) {
     /* now mouse binds */
     static struct node M[] = {
 	/* Mouse   1   0  Close */
-	{ HOLD_LEFT|HOLD_CODE(4),  "0", M+1, N+13,NULL, { { CTX_0, } } },
+	{ HOLD_LEFT|RELEASE_,  "0", M+1, N+13,NULL, { { CTX_0, } } },
 	/* Mouse  123  1  RaiseLower */
-	{ HOLD_ANY |HOLD_CODE(4),  "1", M+2, N+7, NULL, { { CTX_1, } } },
+	{ HOLD_ANY |RELEASE_,  "1", M+2, N+7, NULL, { { CTX_1, } } },
 	/* Mouse  123  3  Roll Toggle */
-	{ HOLD_ANY |HOLD_CODE(4),  "2", M+3, N+6, NULL, { { CTX_2, } } },
+	{ HOLD_ANY |RELEASE_,  "2", M+3, N+6, NULL, { { CTX_2, } } },
 	/* Mouse  H1   TS Interactive Move */
-	{ HOLD_LEFT|HOLD_CODE(3), "TS", M+4, N+0, NULL, { { CTX_TITLE|CTX_SIDE, } } },
+	{ HOLD_LEFT|PRESS_,   "TS", M+4, N+0, NULL, { { CTX_TITLE|CTX_SIDE, } } },
 	/* Mouse  H1   C  Interactive Resize */
-	{ HOLD_LEFT|HOLD_CODE(3),  "C", M+5, N+1, NULL, { { CTX_CORNER, } } },
+	{ HOLD_LEFT|PRESS_,    "C", M+5, N+1, NULL, { { CTX_CORNER, } } },
 	/* Mouse  H1   C  Interactive Scroll */
-	{ HOLD_LEFT|HOLD_CODE(3),  "B", M+6, N+2, NULL, { { CTX_BARS, } } },
+	{ HOLD_LEFT|PRESS_,    "B", M+6, N+2, NULL, { { CTX_BARS, } } },
 	/* Mouse  H1   R _UnFocus */
-	{ HOLD_LEFT|HOLD_CODE(3),  "R", M+8, M+7, NULL, { { CTX_ROOT, } } },
+	{ HOLD_LEFT|PRESS_,    "R", M+8, M+7, NULL, { { CTX_ROOT, } } },
 	{ USERFUNC, "_UnFocus", },
 	/* Mouse  H2  R  WindowList */
-	{ HOLD_MIDDLE|HOLD_CODE(3),"R", M+9, N+10,NULL, { { CTX_ROOT, } } },
+	{ HOLD_MIDDLE|PRESS_,  "R", M+9, N+10,NULL, { { CTX_ROOT, } } },
 	/* Mouse  H3  A  Interactive Menu */
-	{ HOLD_RIGHT|HOLD_CODE(3), "A", M+10,K+1, NULL, { { CTX_ANY, } } },
+	{ HOLD_RIGHT|PRESS_,   "A", M+10,K+1, NULL, { { CTX_ANY, } } },
 	/* Mouse  H1  M  Interactive Screen */
-	{ HOLD_LEFT|HOLD_CODE(3),  "M", NULL,M+11,NULL, { { CTX_MENU, } } },
+	{ HOLD_LEFT|PRESS_,    "M", NULL,M+11,NULL, { { CTX_MENU, } } },
 	{ INTERACTIVE, NULL, NULL, NULL, NULL, { { 0, SCREEN, }, } }
     };
+#ifdef CONF__UNICODE
+# define UD_ARROW T_UTF_16_UP_DOWN_ARROW
+#else
+# define UD_ARROW '\x12'
+#endif
     static button_vec V[] = {
-	{ {'[',    ']'},      0, TRUE, FALSE },
-	{ {'\x12', '\x12'} , -2, TRUE, FALSE },
-	{ {'>',    '<'} ,    -4, TRUE, FALSE }
+	{ {'[',      ']'     },  0, TRUE, FALSE },
+	{ {UD_ARROW, UD_ARROW}, -2, TRUE, FALSE },
+	{ {'>',      '<'     }, -4, TRUE, FALSE }
     };
 
-    str Seq = "";
+    byte *Seq = "";
     /*
      * this is really heavy on the compiler...
      * but it should be able to optimize it
@@ -1140,8 +1166,8 @@ byte InitRC(void) {
     WriteMem(All->ButtonVec, 0, sizeof(All->ButtonVec));
     CopyMem(V, All->ButtonVec, sizeof(V));
     
-    All->SetUp->SelectionButton = HOLD_LEFT;
-    All->SetUp->PasteButton = HOLD_MIDDLE;
+    All->SetUp->ButtonSelection = HOLD_LEFT;
+    All->SetUp->ButtonPaste = HOLD_MIDDLE;
     All->SetUp->DeltaXShade = 3;
     All->SetUp->DeltaXShade = 2;
 
@@ -1150,7 +1176,7 @@ byte InitRC(void) {
 	InitRCOptions();
 	UpdateOptionWin();
 	FillButtonWin();
-	HideMenu(!!(All->SetUp->Flags & SETUP_HIDEMENU));
+	HideMenu(!!(All->SetUp->Flags & SETUP_MENU_HIDE));
 	Act(DrawMenu,All->FirstScreen)(All->FirstScreen, 0, MAXDAT);
 	
 	return TRUE;
