@@ -31,20 +31,28 @@
 #include "data.h"
 #include "methods.h"
 #include "draw.h"
+#include "printk.h"
 #include "resize.h"
 #include "util.h"
 
 #include "hw.h"
 
-#include "Tw/Tw.h"
-#include "Tw/Twstat.h"
+#include <Tw/Tw.h>
+#include <Tw/Twstat.h>
 
 #ifdef CONF__UNICODE
-# include "Tutf/Tutf.h"
+# include <Tutf/Tutf.h>
 #endif
 
 /***************/
 
+#ifdef CONF__UNICODE
+extern hwattr extra_POS_INSIDE;
+#else
+# define extra_POS_INSIDE 0
+#endif
+
+
 byte NeedUpdateCursor;
 
 void FlushCursor(void) {
@@ -63,7 +71,7 @@ void FlushCursor(void) {
 	Window = FindCursorWindow();
 	
 	if (Window && ((Window->Flags & WINDOWFL_CURSOR_ON) ||
-		       (All->SetUp->Flags & SETUP_ALWAYSCURSOR))) {
+		       (All->SetUp->Flags & SETUP_CURSOR_ALWAYS))) {
 
 	    W = (widget)Window;
 	    
@@ -76,11 +84,17 @@ void FlushCursor(void) {
 		CurY >= 0 && CurY < (Window->YWidth - HasBorder) &&
 		InitDrawCtx(W, (dat)CurX, (dat)CurY, (dat)CurX, (dat)CurY, FALSE, &D) &&
 		((Window == (window)Screen->FirstW && !Window->FirstW) ||
+#if 1
+		 /* widgets and gadgets cannot contain cursor, but they can obscure it */
+		 W == RecursiveFindWidgetAt((widget)Screen, (dat)D.X1, (dat)D.Y1 - Screen->YLimit)
+#else
 		 Window == WindowParent(RecursiveFindWidgetAt
-					((widget)Screen, (dat)D.X1, (dat)D.Y1 - Screen->YLimit)))) {
+					((widget)Screen, (dat)D.X1, (dat)D.Y1 - Screen->YLimit))
+#endif
+		 )) {
 		
 		MoveToXY((dat)D.X1, (dat)D.Y1);
-		if ((type = Window->CursorType) == NOCURSOR && All->SetUp->Flags & SETUP_ALWAYSCURSOR)
+		if ((type = Window->CursorType) == NOCURSOR && All->SetUp->Flags & SETUP_CURSOR_ALWAYS)
 		    type = LINECURSOR;
 	    }
 	}
@@ -115,14 +129,16 @@ byte CheckResizeWindowContents(window Window) {
 }
 
 byte ResizeWindowContents(window Window) {
-    hwattr *NewCont, *saveNewCont, *OldCont, *max;
+    hwattr *NewCont, *saveNewCont, *OldCont, *max, h;
     ldat count, common, left;
     ttydata *Data = Window->USE.C.TtyData;
     dat x = Window->XWidth, y = Window->YWidth + Data->ScrollBack;
     
     if (!(Window->Flags & WINDOWFL_BORDERLESS))
 	x -= 2, y -= 2;
-    
+
+    h = HWATTR(Window->ColText, ' ') | extra_POS_INSIDE;
+
     /* safety check: */
     if (x > 0 && y > 0) {
 	if (!(saveNewCont = NewCont = (hwattr *)AllocMem(x*y*sizeof(hwattr))))
@@ -153,14 +169,14 @@ byte ResizeWindowContents(window Window) {
 		    OldCont = Window->USE.C.Contents;
 		NewCont += common;
 		for (left = x - common; left; left--)
-		    *NewCont++ = HWATTR(Window->ColText, ' ');
+		    *NewCont++ = h;
 	    }
 	    FreeMem(Window->USE.C.Contents);
 	}
 	
 	left = (saveNewCont + x*y) - NewCont;
 	while (left--)
-	    *NewCont++ = HWATTR(Window->ColText, ' ');
+	    *NewCont++ = h;
     } else {
 	x = y = 0;
 	if (Window->USE.C.Contents)
@@ -244,7 +260,7 @@ byte EnsureLenRow(row Row, ldat Len, byte DefaultCol) {
     return TRUE;
 }
 
-byte WriteRow(window Window, ldat Len, CONST byte *Text) {
+byte RowWriteAscii(window Window, ldat Len, CONST byte *Text) {
     row CurrRow;
     byte CONST * _Text;
     byte FlagNewRows, ModeInsert;
@@ -254,7 +270,7 @@ byte WriteRow(window Window, ldat Len, CONST byte *Text) {
 #endif
     ldat x, y, max, RowLen;
     
-    if (!W_USE(Window, USEROWS))
+    if (!Window || (Len && !Text) || !W_USE(Window, USEROWS))
 	return FALSE;
     
     x=Window->CurX;
@@ -267,7 +283,7 @@ byte WriteRow(window Window, ldat Len, CONST byte *Text) {
 	ClearHilight(Window);
     
     while (Len) {
-	if (max<=y || (max==y+1 && (*Text==(byte)13 || *Text==(byte)10))) {
+	if (max<=y || (max==y+1 && (*Text=='\n' || *Text=='\r'))) {
 	    if (InsertRowsWindow(Window, Max2(y+1-max,1))) {
 		FlagNewRows=TRUE;
 		max=Window->HLogic;
@@ -281,7 +297,7 @@ byte WriteRow(window Window, ldat Len, CONST byte *Text) {
 	
 	RowLen=(ldat)0;
 	_Text=Text;
-	while (RowLen < Len && *_Text!=(byte)13 && *_Text!=(byte)10)
+	while (RowLen < Len && *_Text!='\n' && *_Text!='\r')
 	    ++RowLen, ++_Text;
 	
 	/*	WINDOWFL_INSERT non implementato */
@@ -330,7 +346,7 @@ byte WriteRow(window Window, ldat Len, CONST byte *Text) {
 	    Len -=RowLen;
 	}
 	
-	if (Len && (*Text=='\r' || *Text=='\n')) {
+	if (Len && (*Text=='\n' || *Text=='\r')) {
 	    Window->CurX=x=(ldat)0;
 	    Window->CurY=++y;
 	    Text++, Len--;
@@ -345,7 +361,103 @@ byte WriteRow(window Window, ldat Len, CONST byte *Text) {
 }
 
 
+#if TW_SIZEOFHWFONT == 1
+byte RowWriteHWFont(window Window, ldat Len, CONST hwfont *Text) {
+    return RowWriteAscii(Window, Len, (CONST byte *)Text);
+}
+#else
+byte RowWriteHWFont(window Window, ldat Len, CONST hwfont *Text) {
+    row CurrRow;
+    CONST hwfont * _Text;
+    byte FlagNewRows, ModeInsert;
+    ldat i;
+    ldat x, y, max, RowLen;
+    
+    if (!Window || !Len || !W_USE(Window, USEROWS))
+	return FALSE;
+    
+    x=Window->CurX;
+    y=Window->CurY;
+    max=Window->HLogic;
+    CurrRow=Window->USE.R.LastRow;
+    ModeInsert=Window->Flags & WINDOWFL_ROWS_INSERT;
+    
+    if (Window->State & WINDOW_ANYSEL)
+	ClearHilight(Window);
+    
+    while (Len) {
+	if (max<=y || (max==y+1 && (*Text=='\n' || *Text=='\r'))) {
+	    if (InsertRowsWindow(Window, Max2(y+1-max,1))) {
+		FlagNewRows=TRUE;
+		max=Window->HLogic;
+		CurrRow=Window->USE.R.LastRow;
+	    } else
+		return FALSE;
+	} else {
+	    FlagNewRows=FALSE;
+	    CurrRow=Act(FindRow,Window)(Window, y);
+	}
+	
+	RowLen=(ldat)0;
+	_Text=Text;
+	while (RowLen < Len && *_Text!='\n' && *_Text!='\r')
+	    ++RowLen, ++_Text;
+	
+	/*	WINDOWFL_INSERT non implementato */
+	/*  Gap non implementato				 */
+	
+	if (RowLen) {
+	    if (ModeInsert || (CurrRow && CurrRow->LenGap))
+		return FALSE;
 
+	    if (!EnsureLenRow(CurrRow, x+RowLen, (Window->Flags & WINDOWFL_ROWS_DEFCOL)))
+		return FALSE;
+	    
+	    if (Window->USE.R.NumRowOne==y)
+		Window->USE.R.RowOne=CurrRow;
+	    if (Window->USE.R.NumRowSplit==y)
+		Window->USE.R.RowSplit=CurrRow;
+	    CurrRow->Flags=ROW_ACTIVE;
+
+	    CopyMem(Text, CurrRow->Text+x, sizeof(hwfont)*RowLen);
+	    if (CurrRow->Len<x)
+		for (i = CurrRow->Len; i < x; i++)
+		    CurrRow->Text[i] = (hwfont)' ';
+
+	    if (!(Window->Flags & WINDOWFL_ROWS_DEFCOL)) {
+		WriteMem(CurrRow->ColText+x, Window->ColText, sizeof(hwcol)*RowLen);
+		if (CurrRow->Len<x)
+		    WriteMem(CurrRow->ColText+CurrRow->Len, Window->ColText,
+			     sizeof(hwcol)*(x-CurrRow->Len));
+	    }
+	    
+	    if (CurrRow->Len<x+RowLen)
+		CurrRow->Len=x+RowLen;
+
+	    DrawLogicWidget((widget)Window, x, y, x+RowLen-(ldat)1, y);
+	
+	    Text+=RowLen;
+	    Len -=RowLen;
+	}
+	
+	if (Len && (*Text=='\n' || *Text=='\r')) {
+	    Window->CurX=x=(ldat)0;
+	    Window->CurY=++y;
+	    Text++, Len--;
+	} else
+	    Window->CurX=x+=RowLen;
+    }
+    
+    if (Window == FindCursorWindow())
+	UpdateCursor();
+    
+    return TRUE;
+}
+#endif /* CONF__UNICODE */
+
+
+
+	    
 
 void ExposeWidget2(widget W, dat XWidth, dat YWidth, dat Left, dat Up, dat Pitch, CONST byte *Text, CONST hwfont *Font, CONST hwattr *Attr) {
     if (w_USE(W, USEEXPOSE)) {
@@ -381,6 +493,120 @@ void ExposeWidget2(widget W, dat XWidth, dat YWidth, dat Left, dat Up, dat Pitch
 }
 
 
+void ExposeWindow2(window W, dat XWidth, dat YWidth, dat Left, dat Up, dat Pitch, CONST byte *Text, CONST hwfont *Font, CONST hwattr *Attr) {
+    ldat CurX, CurY;
+    
+    if (W_USE(W, USEEXPOSE)) {
+	ExposeWidget2((widget)W, XWidth, YWidth, Left, Up, Pitch, Text, Font, Attr);
+	return;
+    }
+
+    if (!W_USE(W, USECONTENTS) && !W_USE(W, USEROWS))
+	return;
+
+    /* handle negative (Left,Up) by clipping */
+    if (Left < 0) {
+	XWidth += Left;
+	if (Text) Text -= Left;
+	if (Font) Font -= Left;
+	if (Attr) Attr -= Left;
+	Left = 0;
+    }
+    if (Up < 0) {
+	YWidth += Up;
+	if (Text) Text -= (ldat)Up * Pitch;
+	if (Font) Font -= (ldat)Up * Pitch;
+	if (Attr) Attr -= (ldat)Up * Pitch;
+	Up = 0;
+    }
+    
+    if (W_USE(W, USECONTENTS)) {
+	/* clip to window size */
+	CurX = W->USE.C.TtyData->SizeX;
+	CurY = W->USE.C.TtyData->SizeY;
+	if (Left >= CurY || Up >= CurY)
+	    return;
+	if ((ldat)XWidth + Left > CurX)
+	    XWidth = CurX - Left;
+	if ((ldat)YWidth + Up > CurY)
+	    YWidth = CurY - Up;
+    }
+    if (XWidth <= 0 || YWidth <= 0)
+	return;
+    
+    if (Text) {
+	void (*WriteString)(window, ldat, CONST byte *);
+	if (W_USE(W, USECONTENTS)) {
+	    WriteString = W->Fn->TtyWriteString;
+	} else
+	    WriteString = (void (*)(window, ldat, CONST byte *))W->Fn->RowWriteString;
+	
+	CurX = W->CurX;
+	CurY = W->CurY;
+	for (; YWidth; YWidth--, Up++, Text += Pitch) {
+	    Act(GotoXY,W)(W, Left, Up);
+	    WriteString(W, XWidth, Text);
+	}
+	Act(GotoXY,W)(W, CurX, CurY);
+	
+    } else if (Font) {
+	void (*WriteHWFont)(window, ldat, CONST hwfont *);
+	if (W_USE(W, USECONTENTS))
+	    WriteHWFont = W->Fn->TtyWriteHWFont;
+	else
+	    WriteHWFont = (void (*)(window, ldat, CONST hwfont *))W->Fn->RowWriteHWFont;
+	
+	CurX = W->CurX;
+	CurY = W->CurY;
+	for (; YWidth; YWidth--, Up++, Font += Pitch) {
+	    Act(GotoXY,W)(W, Left, Up);
+	    WriteHWFont(W, XWidth, Font);
+	}
+	Act(GotoXY,W)(W, CurX, CurY);
+	
+    } else if (Attr) {
+	void (*WriteHWAttr)(window, dat, dat, ldat, CONST hwattr *);
+	if (W_USE(W, USECONTENTS))
+	    WriteHWAttr = W->Fn->TtyWriteHWAttr;
+	else
+	    WriteHWAttr = (void (*)(window, dat, dat, ldat, CONST hwattr *))W->Fn->RowWriteHWAttr;
+			   
+	for (; YWidth; YWidth--, Up++, Attr += Pitch)
+	    WriteHWAttr(W, Left, Up, XWidth, Attr);
+    }
+}
+
+/***************/
+
+
+void ResizeWidget(widget W, dat X, dat Y) {
+    dat mX, mY;
+
+    if (W) {
+	X = Max2(1, X);
+	Y = Max2(1, Y);
+	mX = Max2(X, W->XWidth);
+	mY = Max2(Y, W->YWidth);
+	W->XWidth = X;
+	W->YWidth = Y;
+
+	if (!(W->Flags & WIDGETFL_NOTVISIBLE))
+	    DrawAreaWidget(W);
+    }
+}
+
+void ResizeGadget(gadget G, dat X, dat Y) {
+    if (G) {
+	if (G_USE(G, USETEXT)) {
+	    /* FIXME: finish this */
+	} else {
+	    ResizeWidget((widget)G, X, Y);
+	}
+    }
+}
+
+
+
 /***************/
 
 void DragFirstScreen(ldat DeltaX, ldat DeltaY) {
@@ -898,7 +1124,6 @@ void ResizeRelFirstWindow(dat i, dat j) {
     }
 }
 
-
 void ResizeRelWindow(window Window, dat i, dat j) {
     ldat Left, Up, Rgt, Dwn;
     widget Parent;
@@ -907,12 +1132,14 @@ void ResizeRelWindow(window Window, dat i, dat j) {
     dat YLimit, XWidth, YWidth;
     dat MinXWidth, MinYWidth, MaxXWidth, MaxYWidth;
     dat DWidth, DHeight;
-    byte Shade, DeltaXShade, DeltaYShade;
+    byte Shade, DeltaXShade, DeltaYShade, visible;
     
     if (!Window || (!i && !j)) /* || !(Window->Attrib & WINDOW_RESIZE) */
 	return;
 
-    if ((widget)Window == All->FirstScreen->FirstW) {
+    visible = !(Window->Flags & WIDGETFL_NOTVISIBLE);
+    
+    if (visible && (widget)Window == All->FirstScreen->FirstW) {
 	ResizeRelFirstWindow(i, j);
 	return;
     }
@@ -924,7 +1151,7 @@ void ResizeRelWindow(window Window, dat i, dat j) {
     MaxXWidth=Window->MaxXWidth;
     MaxYWidth=Window->MaxYWidth;
 
-    if ((Parent=Window->Parent) && IS_SCREEN(Parent)) {
+    if (visible && (Parent=Window->Parent) && IS_SCREEN(Parent)) {
 	DWidth=All->DisplayWidth;
 	DHeight=All->DisplayHeight;
 	SetUp=All->SetUp;
@@ -961,7 +1188,7 @@ void ResizeRelWindow(window Window, dat i, dat j) {
 	Dwn+=DeltaY;
     }
     if (DeltaX || DeltaY) {
-	if (Parent && IS_SCREEN(Parent)) {
+	if (visible && Parent && IS_SCREEN(Parent)) {
 	    Up = (dat)Max2(Up, (ldat)YLimit);
 	    DrawArea2((screen)Parent, (widget)0, (widget)0,
 		      (dat)Left, (dat)Up, (dat)Rgt, (dat)Dwn, FALSE);
@@ -969,7 +1196,7 @@ void ResizeRelWindow(window Window, dat i, dat j) {
 		DrawShadeWindow(Window, 0, 0, MAXDAT, MAXDAT, FALSE);
 	}
     
-	if (Window == (window)All->FirstScreen->FocusW)
+	if (visible && Window == (window)All->FirstScreen->FocusW)
 	    UpdateCursor();
 	
 	/* resize contents? for Interactive Resize, let the WM resize it
@@ -986,7 +1213,7 @@ void ResizeRelWindow(window Window, dat i, dat j) {
  * (0,0) is the (XLogic,YLogic) cell of the Window,
  * i.e. the first visible one.
  */
-void ScrollFirstWindowArea(dat X1, dat Y1, dat X2, dat Y2, dat DeltaX, dat DeltaY) {
+void ScrollFirstWindowArea(dat X1, dat Y1, dat X2, dat Y2, ldat DeltaX, ldat DeltaY) {
     screen Screen;
     dat DWidth, DHeight;
     window Window;
@@ -1006,7 +1233,12 @@ void ScrollFirstWindowArea(dat X1, dat Y1, dat X2, dat Y2, dat DeltaX, dat Delta
 
     if (X1 > X2 || X1 > XWidth-3 || X2 < 0 || Y1 > Y2 || Y1 > YWidth-3 || Y2 < 0)
 	return;
-    
+
+    if (DeltaX >= XWidth || -DeltaX >= XWidth || DeltaY >= YWidth || -DeltaY >= YWidth) {
+	DrawWidget((widget)Window, (dat)0, (dat)0, MAXDAT, MAXDAT, FALSE);
+	return;
+    }
+
     DWidth = All->DisplayWidth;
     DHeight = All->DisplayHeight;
     YLimit=Screen->YLimit;
@@ -1067,7 +1299,7 @@ void ScrollFirstWindowArea(dat X1, dat Y1, dat X2, dat Y2, dat DeltaX, dat Delta
     }
 }
 
-void ScrollFirstWindow(dat DeltaX, dat DeltaY, byte byXYLogic) {
+void ScrollFirstWindow(ldat DeltaX, ldat DeltaY, byte byXYLogic) {
     window Window;
     ldat XLogic, YLogic;
     dat XWidth, YWidth;
@@ -1083,24 +1315,24 @@ void ScrollFirstWindow(dat DeltaX, dat DeltaY, byte byXYLogic) {
 	XLogic=Window->XLogic;
 	YLogic=Window->YLogic;
 	
-	if (DeltaX>(dat)0 && XLogic>=MAXLDAT-(ldat)DeltaX)
-	    DeltaX=(dat)(MAXLDAT-XLogic-(ldat)1);
-	else if (DeltaX<(dat)0 && XLogic<(ldat)-DeltaX)
-	    DeltaX=-(dat)XLogic;
+	if (DeltaX>0 && XLogic>=MAXLDAT-DeltaX)
+	    DeltaX=MAXLDAT-XLogic-1;
+	else if (DeltaX<0 && XLogic<-DeltaX)
+	    DeltaX=-XLogic;
 	if (!W_USE(Window, USEROWS)) {
 	    /*
 	     * WARNING: Window->USE.C.Contents and other methods
 	     * may be unable to handle out-of-bound rows
 	     */
-	    if (DeltaY>(dat)0 && (ldat)YLogic+YWidth-2+DeltaY>=Window->HLogic)
-		DeltaY=(dat)(Window->HLogic-YLogic-YWidth+2);
-	    else if (DeltaY<(dat)0 && YLogic<(ldat)-DeltaY)
-		DeltaY=-(dat)YLogic;
+	    if (DeltaY>0 && YLogic+YWidth-2+DeltaY>=Window->HLogic)
+		DeltaY=Window->HLogic-YLogic-YWidth+2;
+	    else if (DeltaY<0 && YLogic<-DeltaY)
+		DeltaY=-YLogic;
 	} else {
-	    if (DeltaY>(dat)0 && YLogic+DeltaY>=Window->HLogic)
-		DeltaY=(dat)(Window->HLogic-YLogic-(ldat)1);
-	    else if (DeltaY<(dat)0 && YLogic<(ldat)-DeltaY)
-		DeltaY=-(dat)YLogic;
+	    if (DeltaY>0 && YLogic+DeltaY>=Window->HLogic)
+		DeltaY=Window->HLogic-YLogic-1;
+	    else if (DeltaY<0 && YLogic<-DeltaY)
+		DeltaY=-YLogic;
 	}
 	if (DeltaX)
 	    Window->XLogic = (XLogic += DeltaX);
@@ -1112,17 +1344,26 @@ void ScrollFirstWindow(dat DeltaX, dat DeltaY, byte byXYLogic) {
 
     ScrollFirstWindowArea(0, 0, XWidth-3, YWidth-3, -DeltaX, -DeltaY);
     
-    if (byXYLogic && Window == (window)All->FirstScreen->FocusW)
+    if (byXYLogic && ContainsCursor((widget)Window))
 	UpdateCursor();
 }
 
-void ScrollWindow(window Window, dat DeltaX, dat DeltaY) {
+void ScrollWindow(window Window, ldat DeltaX, ldat DeltaY) {
     ldat XLogic, YLogic;
     dat YWidth;
     
     if (!Window || !IS_WINDOW(Window) || (!DeltaX && !DeltaY))
 	return;
     
+    if (!(Window->Attrib & WINDOW_X_BAR))
+	DeltaX = 0;
+
+    if (!(Window->Attrib & WINDOW_Y_BAR))
+	DeltaY = 0;
+
+    if (!DeltaX && !DeltaY)
+	return;
+
     if ((widget)Window == All->FirstScreen->FirstW) {
 	ScrollFirstWindow(DeltaX, DeltaY, TRUE);
 	return;
@@ -1132,24 +1373,24 @@ void ScrollWindow(window Window, dat DeltaX, dat DeltaY) {
     XLogic=Window->XLogic;
     YLogic=Window->YLogic;
 	
-    if (DeltaX>(dat)0 && XLogic>=MAXLDAT-(ldat)DeltaX)
-	DeltaX=(dat)(MAXLDAT-XLogic-(ldat)1);
-    else if (DeltaX<(dat)0 && XLogic<(ldat)-DeltaX)
-	DeltaX=-(dat)XLogic;
+    if (DeltaX>0 && XLogic>=MAXLDAT-DeltaX)
+	DeltaX=MAXLDAT-XLogic-1;
+    else if (DeltaX<0 && XLogic<-DeltaX)
+	DeltaX=-XLogic;
     if (!W_USE(Window, USEROWS)) {
 	/*
 	 * WARNING: Window->USE.C.Contents and other methods
 	 * may be unable to handle out-of-bound rows
 	 */
-	if (DeltaY>(dat)0 && (ldat)YLogic+YWidth-2+DeltaY>=Window->HLogic)
-	    DeltaY=(dat)(Window->HLogic-YLogic-YWidth+2);
-	else if (DeltaY<(dat)0 && YLogic<(ldat)-DeltaY)
-	    DeltaY=-(dat)YLogic;
+	if (DeltaY>0 && YLogic+YWidth-2+DeltaY>=Window->HLogic)
+	    DeltaY=Window->HLogic-YLogic-YWidth+2;
+	else if (DeltaY<0 && YLogic<-DeltaY)
+	    DeltaY=-YLogic;
     } else {
-	if (DeltaY>(dat)0 && YLogic+DeltaY>=Window->HLogic)
-	    DeltaY=(dat)(Window->HLogic-YLogic-(ldat)1);
-	else if (DeltaY<(dat)0 && YLogic<(ldat)-DeltaY)
-	    DeltaY=-(dat)YLogic;
+	if (DeltaY>0 && YLogic+DeltaY>=Window->HLogic)
+	    DeltaY=Window->HLogic-YLogic-1;
+	else if (DeltaY<0 && YLogic<-DeltaY)
+	    DeltaY=-YLogic;
     }
     if (DeltaX)
 	Window->XLogic = (XLogic += DeltaX);
@@ -1158,7 +1399,47 @@ void ScrollWindow(window Window, dat DeltaX, dat DeltaY) {
 
     DrawFullWindow2(Window);
 
-    if (Window == (window)All->FirstScreen->FocusW)
+    if (ContainsCursor((widget)Window))
+	UpdateCursor();
+}
+
+void ScrollWidget(widget W, ldat DeltaX, ldat DeltaY) {
+    ldat XLogic, YLogic;
+    dat YWidth;
+
+    if (!W || !(DeltaX || DeltaY))
+	return;
+    
+    if (IS_WINDOW(W)) {
+	ScrollWindow((window)W, DeltaX, DeltaY);
+	return;
+    }
+    
+    YWidth=W->YWidth;
+    XLogic=W->XLogic;
+    YLogic=W->YLogic;
+	
+    if (DeltaX>0 && XLogic>=MAXLDAT-DeltaX)
+	DeltaX=MAXLDAT-XLogic-1;
+    else if (DeltaX<0 && XLogic<-DeltaX)
+	DeltaX=-XLogic;
+    /*
+     * WARNING: content methods may be
+     * unable to handle out-of-bound areas
+     */
+    if (DeltaY>0 && YLogic+YWidth-2>=MAXLDAT-DeltaY)
+	DeltaY=MAXLDAT-YLogic-YWidth+2;
+    else if (DeltaY<0 && YLogic<-DeltaY)
+	DeltaY=-YLogic;
+
+    if (DeltaX)
+	W->XLogic = (XLogic += DeltaX);
+    if (DeltaY)
+	W->YLogic = (YLogic += DeltaY);
+
+    DrawAreaWidget(W);
+
+    if (ContainsCursor(W))
 	UpdateCursor();
 }
 
@@ -1179,7 +1460,7 @@ byte ExecScrollFocusWindow(void) {
     
     Attrib=Window->Attrib;
     State=Window->State;
-    DeltaX=DeltaY=(num)0;
+    DeltaX=DeltaY=(sbyte)0;
     
     if (Attrib & WINDOW_X_BAR && State & X_BAR_SELECT)
 	DeltaX=1;
@@ -1225,6 +1506,9 @@ byte ExecScrollFocusWindow(void) {
     return TRUE;
 }
 
+
+/* menu traversing functions */
+
 void HideMenu(byte on_off) {
     screen Screen=All->FirstScreen;
     
@@ -1251,96 +1535,215 @@ void HideMenu(byte on_off) {
     }
 }
 
-void ChangeMenuFirstScreen(menuitem NewItem, byte ByMouse, byte Flag) {
-    screen Screen;
-    menu Menu, _Menu;
-    menuitem CurrItem;
-    window NewWin;
-    widget CurrW;
+static void OpenSubMenuItem(menu M, menuitem Item, byte ByMouse) {
+    window P = (window)Item->Parent;
+    window W = Item->Window;
+    screen S = All->FirstScreen;
+    ldat y = P->CurY;
     
-    Screen=All->FirstScreen;
-    Menu=Act(FindMenu,Screen)(Screen);
+    P->CurY = Item->WCurY;
+    if (y != MAXLDAT)
+	DrawLogicWidget((widget)P, 0, y, MAXLDAT, y);
+    if ((y = P->CurY) == MAXLDAT)
+	(void)Act(FindRowByCode,P)(P, Item->Code, &P->CurY);
+    if ((y = P->CurY) != MAXLDAT)
+	DrawLogicWidget((widget)P, 0, y, MAXLDAT, y);
     
-    if (Menu) {
-
-	if (All->SetUp->Flags & SETUP_HIDEMENU && Flag==ACTIVATE_MENU_FLAG)
-	    HideMenu(FALSE);
-	
-	CurrW = Screen->FocusW;
-	
-	if (Flag == ACTIVATE_MENU_FLAG)
-	    CurrItem = (menuitem)0;
-	else
-	    CurrItem = Act(GetSelectItem,Menu)(Menu);
-
-	if (CurrItem == NewItem && Flag == KEEP_ACTIVE_MENU_FLAG)
-	    /* this can happen... on an empty menu with common items enabled */
-	    return;
-    
-	/*
-	 * WARNING:
-	 * UnMapWindow() calls us if unmapping a menu window or a menu owner
-	 * while the menu is active (All->State & STATE_ANY) == STATE_MENU
-	 * 
-	 * when shutting down the menu, disable (All->State & STATE_ANY) == STATE_MENU
-	 * _BEFORE_ calling UnMap() or we will enter an infinite loop
-	 */
-	All->State =
-	    (ByMouse && Flag != DISABLE_MENU_FLAG ? STATE_FL_BYMOUSE : (byte)0) |
-	    (Flag != DISABLE_MENU_FLAG ? STATE_MENU : STATE_DEFAULT);
-	
-	if (Flag!=DISABLE_MENU_FLAG) {
-	    
-	    if (Flag==ACTIVATE_MENU_FLAG) {
-		Screen->MenuWindow = (window)CurrW; /* so that it keeps `active' borders */
-		Screen->FocusW = (widget)0;
-	    }
-	    
-	    if (NewItem) {
-		_Menu = (menu)NewItem->Parent; /* may either be Menu or All->CommonMenu */
-		NewWin = NewItem->Window;
-		
-		NewWin->Up = 1;
-		NewWin->Left = NewItem->Left;
-		
-		if (Menu != _Menu && Menu->LastI)
-		    /* adjust common menu NewWin->Left to the Item position in this menu */
-		    NewWin->Left += Menu->LastI->Left + Menu->LastI->Len;
-	    
-		Act(SetSelectItem,Menu)(Menu, NewItem);
-		
-		if (ByMouse)
-		    NewWin->CurY = MAXLDAT;
-		else if (NewWin->CurY == MAXLDAT)
-		    NewWin->CurY = (ldat)0;
-	    
-		if (NewItem->Flags & ROW_ACTIVE)
-		    NewWin->Flags &= ~WINDOWFL_DISABLED;
-		else
-		    NewWin->Flags |= WINDOWFL_DISABLED;
-		
-		Act(MapTopReal,NewWin)(NewWin, Screen);
-	    } else
-		Act(SetSelectItem,Menu)(Menu, (menuitem)0);
+    if (W) {
+	if (!W->Parent) {
+	    W->Left = P->Left + P->XWidth - S->XLogic;
+	    W->Up = P->Up + P->CurY - P->YLogic - S->YLogic + 1;
 	}
-	if (Flag!=ACTIVATE_MENU_FLAG) {
-	    if (Flag==DISABLE_MENU_FLAG) {
-		if ((NewWin = Screen->MenuWindow)) {
-		    Act(KbdFocus,NewWin)(NewWin);
-		    Screen->MenuWindow = (window)0;
-		} else
-		    Do(KbdFocus,Window)((window)0);
-	    }
-	    if (CurrItem && CurrW && IS_WINDOW(CurrW) && (CurrW->Flags & WINDOWFL_MENU))
-		Act(UnMap,CurrW)(CurrW);
-	    UpdateCursor();
-	}
-	if (All->SetUp->Flags & SETUP_HIDEMENU && Flag==DISABLE_MENU_FLAG)
-	    HideMenu(TRUE);
+	if (ByMouse)
+	    W->CurY = MAXLDAT;
+	else if (W->CurY == MAXLDAT)
+	    W->CurY = 0;
+	Act(MapTopReal,W)(W, S);
     }
-    Act(DrawMenu,Screen)(Screen, 0, MAXDAT);
+    if ((widget)P != S->FocusW)
+	Act(Focus,P)(P);
 }
 
+static void OpenTopMenuItem(menu M, menuitem Item, byte ByMouse) {
+    menu _M = (menu)Item->Parent; /* may either be M or All->CommonMenu */
+    window W = Item->Window;
+    
+    if (!W->Parent) {
+	W->Up = 1;
+	W->Left = Item->Left;
+    
+	if (M != _M && M->LastI)
+	    /* adjust common menu W->Left to the Item position in this menu */
+	    W->Left += M->LastI->Left + M->LastI->Len;
+    }
+    
+    Act(SetSelectedItem,M)(M, Item);
+    
+    if (ByMouse)
+	W->CurY = MAXLDAT;
+    else if (W->CurY == MAXLDAT)
+	W->CurY = (ldat)0;
+    
+    if (Item->Flags & ROW_ACTIVE)
+	W->Flags &= ~WINDOWFL_DISABLED;
+    else
+	W->Flags |= WINDOWFL_DISABLED;
+    
+    Act(MapTopReal,W)(W, All->FirstScreen);
+    
+    if (W->CurY != MAXLDAT && (Item = (menuitem)Act(FindRow,W)(W, W->CurY)))
+	OpenSubMenuItem(M, Item, ByMouse);
+}
+
+static void OpenMenuItem(menu M, menuitem Item, byte ByMouse) {
+    if (Item) {
+	obj O = Item->Parent;
+	if (O && IS_WINDOW(O))
+	    OpenSubMenuItem(M, Item, ByMouse);
+	else
+	    OpenTopMenuItem(M, Item, ByMouse);
+    } else
+	Act(SetSelectedItem,M)(M, (menuitem)0);
+}
+
+/* this activates the menu bar */
+static void OpenMenu(menuitem Item, byte ByMouse) {
+    screen S = All->FirstScreen;
+    widget W = S->FocusW;
+    menu M = Act(FindMenu,S)(S);
+    
+    if ((All->State & STATE_ANY) == STATE_DEFAULT) {
+	
+	All->State = STATE_MENU | (ByMouse ? STATE_FL_BYMOUSE : 0);
+
+	if (All->SetUp->Flags & SETUP_MENU_HIDE)
+	    HideMenu(FALSE);
+
+	if (!S->MenuWindow && W) {
+	    S->MenuWindow = (window)W; /* so that it keeps `active' borders */
+	    S->FocusW = (widget)0;
+	}
+    }
+    OpenMenuItem(M, Item, ByMouse);
+}
+
+
+/*
+ * up one level; return new selected item;
+ * do NOT use to close the menu, CloseMenu() does that
+ */
+static menuitem CloseMenuItem(menu M, menuitem Item, byte ByMouse) {
+    window P = (window)Item->Parent, W = Item->Window;
+    
+    if (W)
+	Act(UnMap,W)(W);
+
+    if (P && IS_WINDOW(P)) {
+	if (ByMouse) {
+	    ldat y = P->CurY;
+	    P->CurY = MAXLDAT;
+	
+	    if (y != MAXLDAT)
+		DrawLogicWidget((widget)P, 0, y, MAXLDAT, y);
+	}
+	Item = P->MenuItem;
+	if (Item) {
+	    W = (window)Item->Parent;
+	    if (W && IS_WINDOW(W))
+		Act(Focus,W)(W);
+	    else
+		Act(Focus,P)(P);
+	}
+	return Item;
+    } else {
+	Item = (menuitem)0;
+	Act(SetSelectedItem,M)(M, Item);
+	return Item;
+    }
+}
+
+static dat DepthOfMenuItem(menuitem I) {
+    window W;
+    dat d = 0;
+    
+    while (I && (W = (window)I->Parent) && IS_WINDOW(W)) {
+	I = W->MenuItem;
+	d++;
+    }
+    return d;
+}
+
+/* this traverses the menu bar as needed */
+static void TraverseMenu(menu M, menuitem OldItem, dat Odepth, menuitem NewItem, dat Ndepth, byte ByMouse) {
+    while (Odepth > Ndepth && OldItem) {
+	Odepth--;
+	OldItem = CloseMenuItem(M, OldItem, ByMouse);
+    }
+    /* paranoia */
+    Odepth = DepthOfMenuItem(OldItem);
+    
+    if (Ndepth == Odepth) {
+	if (OldItem != NewItem)
+	    CloseMenuItem(M, OldItem, ByMouse);
+	OpenMenuItem(M, NewItem, ByMouse);
+    } else if (Ndepth == Odepth + 1) {
+	OpenMenuItem(M, NewItem, ByMouse);
+    } else
+	printk("twin: internal error: unsupported menu traversing.\n");
+}
+
+/* close the menu bar */
+void CloseMenu(void) {
+    screen S = All->FirstScreen;
+    menu M = Act(FindMenu,S)(S);
+    menuitem Item;
+    window W;
+    
+    if (M) {
+	if ((W = S->MenuWindow)) {
+	    Act(KbdFocus,W)(W);
+	    S->MenuWindow = (window)0;
+	} else
+	    Do(KbdFocus,Window)((window)0);
+	
+	/* close whole currently open menu tree */
+	Item = Act(GetSelectedItem,M)(M);
+	while (Item && IS_MENUITEM(Item) && (W = (window)Item->Window) && IS_WINDOW(W)) {
+	    Item = (menuitem)Act(FindRow,W)(W,W->CurY);
+	    Act(UnMap,W)(W);
+	}
+    }
+    All->State = STATE_DEFAULT;
+    if (All->SetUp->Flags & SETUP_MENU_HIDE)
+	HideMenu(TRUE);
+    else
+	Act(DrawMenu, S)(S, 0, MAXDAT);
+}
+
+/*
+ * exported interface to open and interact with the menu.
+ * do NOT use to close the menu, CloseMenu() does that
+ */
+void SetMenuState(menuitem Item, byte ByMouse) {
+    screen S = All->FirstScreen;
+    menu M = Act(FindMenu,S)(S);
+    menuitem OldItem = (menuitem)0;
+    dat Odepth = 0;
+    
+    if (M && (Item || ByMouse)) {
+	if ((All->State & STATE_ANY) != STATE_DEFAULT)
+	    OldItem = Act(RecursiveGetSelectedItem,M)(M, &Odepth);
+	if (!OldItem)
+	    OpenMenu(Item, ByMouse);
+	else if (OldItem != Item)
+	    TraverseMenu(M, OldItem, Odepth, Item, DepthOfMenuItem(Item), ByMouse);
+	UpdateCursor();
+    }
+}
+
+
+/* ---------------- */
+
 void UnFocusWidget(widget W) {
     if (W && W->Parent == (widget)All->FirstScreen && W == All->FirstScreen->FocusW) {
 	if (IS_WINDOW(W)) {
@@ -1372,7 +1775,26 @@ void RollUpWindow(window W, byte on_off) {
     }
 }
 
-void MakeFirstWidget(widget W, byte alsoFocus) {
+/* ---------------- */
+
+
+void SetVisibleWidget(widget W, byte on_off) {
+    byte visible;
+    if (W) {
+	on_off = !!on_off;
+	visible = !(W->Flags & WIDGETFL_NOTVISIBLE);
+	
+	if (on_off != visible) {
+	    W->Flags ^= WIDGETFL_NOTVISIBLE;
+	    if (IS_WINDOW(W))
+		DrawAreaWindow2((window)W);
+	    else
+		DrawAreaWidget(W);
+	}
+    }
+}
+
+void RaiseWidget(widget W, byte alsoFocus) {
     screen Screen;
     
     if (W && (Screen = (screen)W->Parent) && IS_SCREEN(Screen)) {
@@ -1395,7 +1817,7 @@ void MakeFirstWidget(widget W, byte alsoFocus) {
     }
 }
 
-void MakeLastWidget(widget W, byte alsoUnFocus) {
+void LowerWidget(widget W, byte alsoUnFocus) {
     screen Screen;
     widget _W;
     
@@ -1425,14 +1847,76 @@ void MakeLastWidget(widget W, byte alsoUnFocus) {
 }
 
 
+void RestackWidgets(widget W, uldat N, CONST widget *arrayW) {
+    widget FW, CW;
+    byte need_redraw = FALSE;
+    
+    if (W && N && arrayW) {
+	for (FW = (widget)0; N; N--, arrayW++) {
+	    /*
+	     * Allow only children that really have the given parent.
+	     * Also deny WINDOFL_MENU windows
+	     */
+	    if ((CW = *arrayW) && CW->Parent == W && !(CW->Flags & WINDOWFL_MENU)) {
+		if (FW && CW != FW->Next) {
+		    /* restack after arrayW[0] */
+		    Remove(CW);
+		    Act(Insert,CW)(CW, W, FW, FW->Next);
+		    need_redraw = TRUE;
+		}
+		FW = CW;
+	    }
+	}
+	/* FIXME: this is gross */
+	if (need_redraw)
+	    DrawAreaWidget(W);
+    }
+}
 
 
 
+/* ---------------- */
+
+
+void RestackRows(obj O, uldat N, CONST row *arrayR) {
+    row FR, CR;
+    byte need_redraw = FALSE;
+    
+    if (O && (IS_MENU(O) || IS_WINDOW(O)) && N && arrayR) {
+	for (FR = (row)0; N; N--, arrayR++) {
+	    /*
+	     * Allow only children that really have the given parent.
+	     */
+	    if ((CR = *arrayR) && (obj)CR->Window == O) {
+		if (FR && CR != FR->Next) {
+		    /* restack after arrayR[0] */
+		    Remove(CR);
+		    Act(Insert,CR)(CR, (window)O, FR, FR->Next);
+		    need_redraw = TRUE;
+		}
+		FR = CR;
+	    }
+	}
+	/* FIXME: this is gross */
+	if (need_redraw) {
+	    if (IS_MENU(O))
+		SyncMenu((menu)O);
+	    else
+		DrawAreaWidget((widget)O);
+	}
+    }
+}
+
+
+
+/* ---------------- */
+
+
 void SendMsgGadget(gadget G) {
     msg Msg;
     event_gadget *Event;
     if (G->Code && !(G->Flags & GADGETFL_DISABLED)) {
-	if ((Msg=Do(Create,Msg)(FnMsg, MSG_WIDGET_GADGET, sizeof(event_gadget)))) {
+	if ((Msg=Do(Create,Msg)(FnMsg, MSG_WIDGET_GADGET, 0))) {
 	    Event = &Msg->Event.EventGadget;
 	    Event->W = G->Parent;
 	    Event->Code = G->Code;
@@ -1539,7 +2023,7 @@ void WriteTextsGadget(gadget G, byte bitmap, dat TW, dat TH, CONST byte *Text, d
 			_W = W;
 			while (_W-- > 0) {
 # ifdef CONF__UNICODE
-			    *GT++ = Tutf_IBM437_to_UTF_16[*TT++];
+			    *GT++ = Tutf_CP437_to_UTF_16[*TT++];
 # else
 			    *GT++ = *TT++;
 # endif
@@ -1621,7 +2105,7 @@ void WriteHWFontsGadget(gadget G, byte bitmap, dat TW, dat TH, CONST hwfont *HWF
 			_W = W;
 			while (_W-- > 0) {
 # ifdef CONF__UNICODE   
-			    *GT++ = Tutf_IBM437_to_UTF_16[*TT++];
+			    *GT++ = Tutf_CP437_to_UTF_16[*TT++];
 # else
 			    *GT++ = *TT++;
 # endif
@@ -1645,6 +2129,24 @@ void WriteHWFontsGadget(gadget G, byte bitmap, dat TW, dat TH, CONST hwfont *HWF
 #endif
 }
 
+void SyncMenu(menu Menu) {
+    menuitem I, PrevI = (menuitem)0;
+    screen Screen;
+    
+    if (Menu) {
+	for (I = Menu->FirstI; I; I = I->Next) {
+	    if (PrevI)
+		I->Left = PrevI->Left + PrevI->Len;
+	    else
+		I->Left = 1;
+	    PrevI = I;
+	}
+	for (Screen = All->FirstScreen; Screen; Screen = Screen->Next) {
+	    if (Act(FindMenu,Screen)(Screen) == Menu)
+		Act(DrawMenu,Screen)(Screen, 0, MAXDAT);
+	}
+    }
+}
 
 #if 0
 void SetNewFont(void) {
