@@ -1,16 +1,37 @@
 
+
+
+#define _obj .TWS_field_obj
+#define _any .TWS_field_scalar
+#define _vec .TWS_field_vecV
+#define _len .TWS_field_vecL
+#define _type .type
+
+#define void_ TWS_void
+#define obj_  TWS_obj
+#define vec_  TWS_vec
+#define vecW_ TWS_vecW
+
+
+
+
+/* defines and wrappers for commonly used stuff */
 #define SIZEOF(type)			AlienSizeof(type,Slot)
 #define REPLY(code, type, lval)		alienReply(code, SIZEOF(type), sizeof(type), lval)
 
 #define POP(s,type,lval)		alienPOP(s,type,lval)
+/* wrap alienPUSH() to work with non-lvalue `val' */
 #define PUSH(s,type,val)		do {type tmp = (val); alienPUSH(s,type,tmp); } while(0)
 #define POPADDR(s,type,len,ptr)		PopAddr(s,type,len,ptr)
 
 
 
-#if TW_BYTE_ORDER == TW_LITTLE_ENDIAN && defined(__i386__) /* due to lack of alignment */
+#if TW_BYTE_ORDER == TW_LITTLE_ENDIAN
 /* we can use hton?() functions to speed up translation */
-#include <netinet/in.h>
+# include <netinet/in.h>
+#endif
+
+#if TW_BYTE_ORDER == TW_LITTLE_ENDIAN && TW_CAN_UNALIGNED != 0 /* due to lack of alignment */
 INLINE void FlipCopyMem(CONST byte *src, byte *dst, uldat len) {
     switch (len) {
       case 2:
@@ -40,6 +61,7 @@ INLINE void FlipCopyMem(CONST byte *src, byte *dst, uldat len) {
 #endif /* TW_BYTE_ORDER == TW_LITTLE_ENDIAN */
 
 
+
 /*translate from alien data, copying srclen bytes to dstlen bytes, optionally flipping byte order*/
 static void alienRead(CONST byte *src, uldat srclen, byte *dst, uldat dstlen, byte flip) {
     
@@ -115,17 +137,12 @@ static void alienPush(CONST byte *src, uldat len, byte **dst, uldat alien_len) {
 
 /*
  * translate from alien data, copying len bytes from srcsize chunks to dstsize chunks, optionally flipping byte order.
- * allocate a new buffer to hold data.
+ * assume dst is large enough to hold translated data.
  */
-static CONST byte *alienReadVec(CONST byte *src, uldat len, uldat srcsize, uldat dstsize, byte flag) {
-    byte *dstbuf, *dst;
-    
+INLINE void alienReadVec(CONST byte *src, byte *dst, uldat len, uldat srcsize, uldat dstsize, byte flag) {
     /* round to srcsize multiple */
     len = (len / srcsize) * srcsize;
     
-    if (!(dstbuf = dst = AllocMem(len / srcsize * dstsize)))
-	return dst;
-    
     /* optimize common cases */
     
 #if TW_BYTE_ORDER == TW_LITTLE_ENDIAN
@@ -171,13 +188,15 @@ static CONST byte *alienReadVec(CONST byte *src, uldat len, uldat srcsize, uldat
 #endif /* TW_BYTE_ORDER == TW_LITTLE_ENDIAN */
 	
     else if (srcsize == dstsize) {
-	/* (flag) must be TRUE, */
-	while (len) {
-	    FlipCopyMem(src, dst, srcsize);
-	    src += srcsize;
-	    dst += dstsize;
-	    len -= srcsize;
-	}
+	if (flag) {
+	    while (len) {
+		FlipCopyMem(src, dst, srcsize);
+		src += srcsize;
+		dst += dstsize;
+		len -= srcsize;
+	    }
+	} else
+	    CopyMem(src, dst, srcsize);
     } else {
 	/* fallback. all other cases */
 	while (len) {
@@ -187,12 +206,57 @@ static CONST byte *alienReadVec(CONST byte *src, uldat len, uldat srcsize, uldat
 	    len -= srcsize;
 	}
     }
-    return dstbuf;
 }
 
 
+/*
+ * translate from alien data, copying len bytes from srcsize chunks to dstsize chunks, optionally flipping byte order.
+ * allocate a new buffer to hold data.
+ */
+static CONST byte *alienAllocReadVec(CONST byte *src, uldat len, uldat srcsize, uldat dstsize, byte flag) {
+    byte *dst;
+    
+    if ((dst = AllocMem(len / srcsize * dstsize)))
+	alienReadVec(src, dst, len, srcsize, dstsize, flag);
+    
+    return dst;
+}
+
+
+/*
+ * translate to alien data, copying len bytes from srcsize chunks to dstsize chunks, optionally flipping byte order.
+ * assume dst is large enough to hold translated data.
+ */
+#ifdef CONF__UNICODE
+static void alienWriteVec(CONST byte *src, byte *dst, uldat len, uldat srcsize, uldat dstsize, byte flag) {
+    /* round to srcsize multiple */
+    len = (len / srcsize) * srcsize;
+
+    /* no optimization for srcsize == 1 or dstsize == 1 here */
+    if (srcsize == dstsize) {
+	if (flag) {
+	    while (len) {
+		FlipCopyMem(src, dst, srcsize);
+		src += srcsize;
+		dst += dstsize;
+		len -= srcsize;
+	    }
+	} else
+	    CopyMem(src, dst, srcsize);
+    } else {
+	/* fallback. all other cases */
+	while (len) {
+	    alienWrite(src, srcsize, dst, dstsize, flag);
+	    src += srcsize;
+	    dst += dstsize;
+	    len -= srcsize;
+	}
+    }
+}
+#endif
+
 static void alienReply(uldat code, uldat alien_len, uldat len, CONST void *data) {
-    byte AlienSizeofUldat = AlienSizeof(uldat,Slot);
+    byte AlienSizeofUldat = SIZEOF(uldat);
     
     if (RemoteWriteQueue(Slot, 3*AlienSizeofUldat + alien_len, NULL)
 	== 3*AlienSizeofUldat + alien_len) {
@@ -216,197 +280,337 @@ static void alienReply(uldat code, uldat alien_len, uldat len, CONST void *data)
 }
 
 #ifdef CONF__UNICODE
-static void alienTranslateHWAttrV_IBM437_to_UTF_16(hwattr *H, uldat Len) {
+static void alienTranslateHWAttrV_CP437_to_UTF_16(hwattr *H, uldat Len) {
     hwfont f;
     while (Len--) {
-	f = Tutf_IBM437_to_UTF_16[HWFONT(*H) & 0xFF];
+	f = Tutf_CP437_to_UTF_16[HWFONT(*H) & 0xFF];
 	*H = HWATTR_COLMASK(*H) | HWATTR(0, f);
 	H++;
     }
 }
 #endif
+
+TW_INLINE ldat alienDecodeArg(uldat id, CONST byte * Format, uldat n, tsfield a, uldat mask[1], byte flag[1], ldat fail) {
+    uldat a0, nlen;
+    byte c;
+    void *av;
+    CONST void *A;
     
-static void alienDecode(uldat id) {
-    static any a[20];
-    uldat mask = 0; /* at least 32 bits. we need 20... */
-    uldat nlen, n = 1;
-    ldat fail = 1;
-    byte *Format = sockF[id].Format;
-    uldat a0;
-    byte called, c, self, tmp, retF, retT;
-    
-    self = *Format++;
-    retF = *Format++;
-    retT = *Format++;
-    
-    while (fail > 0 && (c = *Format++)) {
-	switch (c) {
-	  case '_':
-	    switch ((c = *Format)) {
+    switch ((c = *Format++)) {
+      case '_':
+	switch ((c = *Format)) {
 #define CASE_(type) \
-	      case CAT(TWS_,type) + TWS_base_CHR: \
-    		/* ensure type size WAS negotiated */ \
-		if ((CAT(TWS_,type) <= TWS_hwcol || AlienSizeof(type, Slot)) && Left(SIZEOF(type))) { \
-		    type an; \
-		    POP(s,type,an); \
-		    a[n]._ = (uldat)an; \
-		} else \
-		    fail = -fail; \
-		break
+	  case CAT(TWS_,type): \
+	    /* ensure type size WAS negotiated */ \
+	    if (SIZEOF(type) && Left(SIZEOF(type))) { \
+		type an; \
+		POP(s,type,an); \
+		a[n]_any = (tany)an; \
+		a[n]_type = c; \
+		break; \
+	    } \
+	    fail = -fail; \
+	    break
 		
-		case TWS_hwcol + TWS_base_CHR:
-		/*FALLTHROUGH*/
-		CASE_(byte);
-		CASE_(dat);
-		CASE_(ldat);
-		CASE_(time_t);
-		CASE_(frac_t);
-		CASE_(hwfont);
+	  case TWS_hwcol:
+	    /*FALLTHROUGH*/
+	    CASE_(byte);
+	    CASE_(dat);
+	    CASE_(ldat);
+	    CASE_(topaque);
+	    CASE_(tany);
 #ifndef CONF__UNICODE
-		CASE_(hwattr);
+	    CASE_(hwfont);
 #else
-		case TWS_hwattr + TWS_base_CHR:
-		/* ensure hwattr size WAS negotiated */
-		if ((TWS_hwattr <= TWS_hwcol || AlienSizeof(hwattr, Slot)) && Left(SIZEOF(hwattr))) {
-		    hwattr an;
-		    POP(s,hwattr,an);
-		    a[n]._ = (uldat)an;
-		    if (AlienSizeof(hwattr, Slot) == 1)
-			a[n]._ = Tutf_IBM437_to_UTF_16[a[n]._ & 0xFF];
-		} else
-		    fail = -fail;
-		break;
-#endif
-#undef CASE_
-	      default:
+	  case TWS_hwfont:
+	    /* ensure hwattr size WAS negotiated */
+	    if (SIZEOF(hwfont) && Left(SIZEOF(hwfont))) {
+		hwfont an;
+		POP(s,hwfont,an);
+		a[n]_any = (tany)an & 0xFF;
+		if (SIZEOF(hwfont) == 1)
+		    a[n]_any = Tutf_CP437_to_UTF_16[a[n]_any];
+		a[n]_type = c;
 		break;
 	    }
+	    fail = -fail;
 	    break;
-	  case 'x':
-	    /* all kind of pointers */
-	    if (Left(SIZEOF(uldat))) {
-		POP(s,uldat,a[n]._);
-		c = *Format - base_magic_CHR;
-		a[n].x = Id2Obj(c, a[n]._);
-	    } else
-		fail = -fail;
+#endif
+#ifndef CONF__UNICODE
+	    CASE_(hwattr);
+#else
+	  case TWS_hwattr:
+	    /* ensure hwattr size WAS negotiated */
+	    if (SIZEOF(hwattr) && Left(SIZEOF(hwattr))) {
+		hwattr an;
+		POP(s,hwattr,an);
+		a[n]_any = HWATTR_COLMASK(an);
+		an = HWFONT(an) & 0xFF;
+		if (SIZEOF(hwattr) == 2)
+		    an = Tutf_CP437_to_UTF_16[an];
+		a[n]_any |= HWATTR(0, an);
+		a[n]_type = c;
+		break;
+	    }
+	    fail = -fail;
 	    break;
-	  case 'V':
-	    nlen = sockLengths(id, n, a);
-	    c = *Format - TWS_base_CHR;
+#endif
+#undef CASE_
+	  default:
+	    break;
+	}
+	break;
+      case 'x':
+	/* all kind of pointers */
+	if (Left(SIZEOF(uldat))) {
+	    POP(s,uldat,a0);
+	    c = *Format - base_magic_CHR;
+	    a[n]_obj = Id2Obj(c, a0);
+	    a[n]_type = obj_;
+	    break;
+	}
+	fail = -fail;
+	break;
+      case 'V':
+	nlen = sockLengths(id, n, a);
+	c = *Format;
+	/* ensure type size WAS negotiated */
+	if (AlienMagic(Slot)[c]) {
+	    nlen *= AlienMagic(Slot)[c];
+	    if (Left(nlen)) {
+		PopAddr(s,byte,nlen,av);
+		a[n]_vec = av;
+		a[n]_len = nlen / AlienMagic(Slot)[c] * TwinMagicData[c];
+		a[n]_type = vec_|c;
+		
+		if (AlienMagic(Slot)[c] != TwinMagicData[c] ||
+		    (TwinMagicData[c] != 1 && AlienXendian(Slot) == MagicAlienXendian)) {
+			
+		    a[n]_vec = alienAllocReadVec
+			(av, nlen, AlienMagic(Slot)[c], TwinMagicData[c],
+			 AlienXendian(Slot) == MagicAlienXendian);
+		    if (a[n]_vec) {
+#ifdef CONF__UNICODE
+			if (c == TWS_hwattr && SIZEOF(hwattr) == 2)
+			    alienTranslateHWAttrV_CP437_to_UTF_16((hwattr *)a[n]_vec, nlen);
+#endif
+			*mask |= 1 << n;
+		    } else
+			fail = -fail;
+		}
+		break;
+	    }
+	}
+	fail = -fail;
+	break;
+      case 'W':
+	if (Left(SIZEOF(uldat))) {
+	    POP(s,uldat,nlen);
+		
+	    c = *Format;
 	    /* ensure type size WAS negotiated */
-	    if ((c <= TWS_hwcol || AlienMagic(Slot)[c])) {
-		nlen *= AlienMagic(Slot)[c];
-		if (Left(nlen)) {
-		    PopAddr(s,byte,nlen,a[n].V);
+	    if (AlienMagic(Slot)[c]) {
+		if (!nlen || (Left(nlen) && nlen == sockLengths(id, n, a) * AlienMagic(Slot)[c])) {
+		    PopAddr(s,byte,nlen,av);
+		    a[n]_vec = av;
+		    a[n]_len = nlen / AlienMagic(Slot)[c] * TwinMagicData[c];
+		    a[n]_type = vec_|vecW_|c;
+		    
 		    if (AlienMagic(Slot)[c] != TwinMagicData[c] ||
 			(TwinMagicData[c] != 1 && AlienXendian(Slot) == MagicAlienXendian)) {
 			
-			a[n].V = alienReadVec(a[n].V, nlen, AlienMagic(Slot)[c],
-					      TwinMagicData[c], AlienXendian(Slot) == MagicAlienXendian);
-			if (a[n].V) {
+			a[n]_vec = alienAllocReadVec
+			    (av, nlen, AlienMagic(Slot)[c], TwinMagicData[c],
+			     AlienXendian(Slot) == MagicAlienXendian);
+			    
+			if (a[n]_vec) {
 #ifdef CONF__UNICODE
-			    if (c == TWS_hwattr && AlienMagic(Slot)[c] == 1)
-				alienTranslateHWAttrV_IBM437_to_UTF_16((hwattr *)a[n].V, nlen);
+			    if (c == TWS_hwattr && SIZEOF(hwattr) == 2)
+				alienTranslateHWAttrV_CP437_to_UTF_16((hwattr *)a[n]_vec, nlen);
 #endif
-			    mask |= 1 << n;
+			    *mask |= 1 << n;
 			} else
 			    fail = -fail;
 		    }
 		    break;
 		}
 	    }
-	    fail = -fail;
-	    break;
-	  case 'W':
-	    if (Left(SIZEOF(uldat))) {
-		POP(s,uldat,nlen);
+	}
+	fail = -fail;
+	break;
+    case 'X':
+	nlen = sockLengths(id, n, a) * SIZEOF(uldat);
+	c = *Format - base_magic_CHR;
+	if (Left(nlen)) {
+	    PopAddr(s,byte,nlen,av);
+	    a[n]_vec = av;
+	    a[n]_len = nlen / SIZEOF(uldat) * sizeof(uldat);
+	    a[n]_type = vec_|obj_;
+	    
+	    if (SIZEOF(uldat) != sizeof(uldat) ||
+		AlienXendian(Slot) == MagicAlienXendian) {
+
+		A = a[n]_vec = alienAllocReadVec
+		    (av, nlen, SIZEOF(uldat), sizeof(uldat),
+		     AlienXendian(Slot) == MagicAlienXendian);
+	    } else
+		A = a[n]_vec;
 		
-		c = *Format - TWS_base_CHR;
-		/* ensure type size WAS negotiated */
-		if ((c <= TWS_hwcol || AlienMagic(Slot)[c])) {
-		    if (!nlen || (Left(nlen) && nlen == sockLengths(id, n, a)) * AlienMagic(Slot)[c]) {
-			PopAddr(s,byte,nlen,a[n].V);
+	    if (A) {
+		a[n]_vec = AllocId2ObjVec(flag, c, nlen/sizeof(uldat), (byte *)A);
+		FreeMem((void *)A);
+		if (a[n]_vec) {
+		    *mask |= *flag << n;
+		    break;
+		}
+	    }
+	}
+	fail = -fail;
+	break;
+      case 'Y':
+	c = *Format - base_magic_CHR;
+	if (Left(sizeof(uldat))) {
+	    Pop(s,uldat,nlen);
+	    nlen *= sizeof(uldat);
+	    if (Left(nlen)) {
+		PopAddr(s,byte,nlen,av);
+		a[n]_vec = av;
+		a[n]_len = nlen / SIZEOF(uldat) * sizeof(uldat);
+		a[n]_type = vec_|obj_;
+
+		if (SIZEOF(uldat) != sizeof(uldat) ||
+		    AlienXendian(Slot) == MagicAlienXendian) {
 			
-			if (AlienMagic(Slot)[c] != TwinMagicData[c] ||
-			    (TwinMagicData[c] != 1 && AlienXendian(Slot) == MagicAlienXendian)) {
-			
-			    a[n].V = alienReadVec(a[n].V, nlen, AlienMagic(Slot)[c],
-						  TwinMagicData[c], AlienXendian(Slot) == MagicAlienXendian);
-			    
-			    if (a[n].V) {
-#ifdef CONF__UNICODE
-				if (c == TWS_hwattr && AlienMagic(Slot)[c] == 1)
-				    alienTranslateHWAttrV_IBM437_to_UTF_16((hwattr *)a[n].V, nlen);
-#endif
-				mask |= 1 << n;
-			    } else
-				fail = -fail;
-			}
+		    A = a[n]_vec = alienAllocReadVec
+			(av, nlen, SIZEOF(uldat), sizeof(uldat),
+			 AlienXendian(Slot) == MagicAlienXendian);
+		} else
+		    A = a[n]_vec;
+		    
+		if (A) {
+		    a[n]_vec = AllocId2ObjVec(flag, c, nlen/sizeof(uldat), (byte *)A);
+		    FreeMem((void *)A);
+		    if (a[n]_vec) {
+			*mask |= *flag << n;
 			break;
 		    }
 		}
 	    }
-	    fail = -fail;
-	    break;
-	  default:
-	    fail = -fail;
-	    break;
 	}
+	fail = -fail;
+	break;
+      default:
+	fail = -fail;
+	break;
+    }
+    return fail;
+}
+
+static void alienMultiplexB(uldat id) {
+    static struct s_tsfield a[TW_MAX_ARGS_N];
+    static byte warned = FALSE;
+    uldat mask = 0; /* at least 32 bits. we need 20... */
+    uldat nlen, n = 1;
+    ldat fail = 1;
+    CONST byte *Format = sockF[id].Format;
+    uldat a0;
+    byte c, self, flag, tmp, retT[2];
+    
+    self = *Format++;
+    retT[0] = *Format++;
+    retT[1] = *Format++;
+    
+    while (fail > 0 && *Format) {
+	if (n < TW_MAX_ARGS_N) {
+	    fail = alienDecodeArg(id, Format, n, a, &mask, &flag, fail);
+	
+	} else /* (n >= TW_MAX_ARGS_N) */ {
+	    if (!warned) {
+		warned = TRUE;
+		printk("twin: alienMultiplexB(): got a call with %d args, only %d supported!\n",
+		       n, TW_MAX_ARGS_N);
+	    }
+	    fail = -fail;
+	}
+	
 	if (fail > 0) {
-	    Format++;
+	    Format += 2;
 	    fail++;
 	    n++;
 	} else
 	    break;
     }
 
-    if ((called = (fail > 0 && s == end && !c && (self != '2' || a[1].x))))
-	sockMultiplex(id, a);
-
-    while (mask && n) {
-	if (mask & (1 << n)) {
-	    mask &= ~(1 << n);
-	    /*avoid warnings about discarding CONST on a[n].V*/
-	    FreeMem(a[n].VV);
+    if ((flag = (fail > 0 && s == end && !*Format && (self != '2' || a[1]_obj)))) {
+	
+	if (retT[0] == 'O' && a[n-1]_type == (TWS_vec|TWS_byte) && a[n-1]_len == 2*sizeof(byte)) {
+	    /*
+	     * variable return type. store it in last arg,
+	     * and let function implementation overwrite it
+	     * 
+	     * evil trick: only a[n-1]_vec will be passed to the function,
+	     * but it points to a[n-1] itself!
+	     */
+	    a[n-1]_type = proto_2_TWS(a[n-1]_vec);
+	    if (mask & 1<<(n-1))
+		FreeMem(a[n-1].TWS_field_vecVV);
+	    
+	    a[n-1]_vec = &a[n-1];
+	    a[n-1]_len = 0;
 	}
-	n--;
+	
+	fullMultiplexS(id, n, a);
+    }
+    
+    for (nlen = 0; mask; mask >>= 1, nlen++) {
+	if (mask & 1)
+	    FreeMem(a[nlen].TWS_field_vecVV);
     }
 
-    if (called) {
+    if (flag) {
+        if (retT[0] == 'O') {
+	    /* variable return type. get it from last arg */
+	    /* FIXME: currently, only '_' (scalar) and 'v' (void) return types are supported */
+		
+	    TWS_2_proto(a[n-1]_type, retT);
+	}
 	
-	switch (retF) {
+	switch (retT[0]) {
 	  case '_':
-	    switch (retT) {
+	    switch (retT[1]) {
 #define CASE_(type) \
-	      case CAT(TWS_,type) + TWS_base_CHR: \
+	      case CAT(TWS_,type): \
     		/* ensure type size WAS negotiated */ \
-		if (CAT(TWS_,type) <= TWS_hwcol || AlienSizeof(type, Slot)) { \
-		    *(type *)&a[0] = (type)a[0]._; \
+		if (CAT(TWS_,type) <= TWS_hwcol || SIZEOF(type)) { \
+		    /* move to first bytes on MSB machines */ \
+		    *(type *)&a[0]_any = (type)a[0]_any; \
 		    c = SIZEOF(type); \
 		    tmp = sizeof(type); \
+		    break; \
 		} \
+		fail = 0; \
 		break
 		
-		case TWS_hwcol + TWS_base_CHR:
+		case TWS_hwcol:
 		/*FALLTHROUGH*/
 		CASE_(byte);
 		CASE_(dat);
 		CASE_(ldat);
-		CASE_(time_t);
-		CASE_(frac_t);
+		CASE_(topaque);
+		CASE_(tany);
 		CASE_(hwfont);
 #ifndef CONF__UNICODE
 		CASE_(hwattr);
 #else
-		case TWS_hwattr + TWS_base_CHR:
+	      case TWS_hwattr:
 		/* ensure hwattr size WAS negotiated */
-		if (TWS_hwattr <= TWS_hwcol || AlienSizeof(hwattr, Slot)) {
-		    if (AlienSizeof(hwattr, Slot) == 1) {
-			hwfont f = Tutf_UTF_16_to_IBM437(HWFONT(a[0]._));
-			a[0]._ = HWATTR_COLMASK(a[0]._) | HWATTR(0, f);
+		if (TWS_hwattr <= TWS_hwcol || SIZEOF(hwattr)) {
+		    if (SIZEOF(hwattr) == 1) {
+			hwfont f = Tutf_UTF_16_to_CP437(HWFONT(a[0]_any));
+			a[0]_any = HWATTR_COLMASK(a[0]_any) | HWATTR(0, f);
 		    }
-		    *(hwattr *)&a[0] = (hwattr)a[0]._;
+		    /* move to first bytes on MSB machines */
+		    *(hwattr *)&a[0]_any = (hwattr)a[0]_any;
 		    c = SIZEOF(hwattr);
 		    tmp = sizeof(hwattr);
 		} else
@@ -419,13 +623,13 @@ static void alienDecode(uldat id) {
 		break;
 	    }
 	    if (c && fail > 0) {
-		alienReply(OK_MAGIC, tmp, c, &a[0]);
+		alienReply(OK_MAGIC, tmp, c, &a[0]_any);
 		return;
 	    }
 	    break;
 	    
 	  case 'x':
-	    a0 = a[0].x ? a[0].x->Id : NOID;
+	    a0 = a[0]_obj ? a[0]_obj->Id : NOID;
 	    REPLY(OK_MAGIC, uldat, &a0);
 	    return;
 	    
@@ -437,9 +641,9 @@ static void alienDecode(uldat id) {
 	    break;
 	}
     }
-    if (retF != 'v') {
+    if (retT[0] != 'v') {
 	if (fail > 0) {
-	    if (self != '2' || a[1].x)
+	    if (self != '2' || a[1]_obj)
 		fail = FAIL_MAGIC;
 	    else
 		fail = 1;
@@ -449,10 +653,23 @@ static void alienDecode(uldat id) {
 }
 
 
+#undef _obj
+#undef _any
+#undef _vec
+#undef _len
+#undef _type
+
+#undef void_
+#undef obj_
+#undef vec_
+#undef vecW_
+
+
+
 static void AlienIO(int fd, uldat slot) {
     uldat len, Funct;
     byte *t, *tend;
-    size_t tot;
+    int tot = 0;
 #ifdef CONF_SOCKET_GZ
     uldat gzSlot;
 #endif
@@ -490,7 +707,7 @@ static void AlienIO(int fd, uldat slot) {
 	s = t = RemoteReadGetQueue(Slot, &len);
 	tend = s + len;
 
-	AlienSizeofUldat = AlienSizeof(uldat,Slot);
+	AlienSizeofUldat = SIZEOF(uldat);
 	
 	while (s + 3*AlienSizeofUldat <= tend) {
 	    POP(s,uldat,len);
@@ -504,7 +721,7 @@ static void AlienIO(int fd, uldat slot) {
 		POP(s, uldat, Funct);
 		if (Funct < MaxFunct) {
 		    slot = Slot;
-		    alienDecode(Funct); /* Slot is the uncompressed socket here ! */
+		    alienMultiplexB(Funct); /* Slot is the uncompressed socket here ! */
 		    Slot = slot;	/*
 					 * restore, in case alienF[Funct].F() changed it;
 					 * without this, tw* clients can freeze
@@ -512,7 +729,7 @@ static void AlienIO(int fd, uldat slot) {
 					 */
 		}
 		else if (Funct == FIND_MAGIC)
-		    alienDecode(0);
+		    alienMultiplexB(0);
 		s = end;
 	    } else if (s + len < s) {
 		s = tend;
@@ -538,6 +755,55 @@ static void AlienIO(int fd, uldat slot) {
 }
 
 
+/*move chunk bytes to chunk bytes at time, flipping byte order*/
+static void FlipMoveMem(byte *mem, uldat len, uldat chunk) {
+    uldat i, j;
+    byte c;
+#if TW_BYTE_ORDER == TW_LITTLE_ENDIAN
+    byte32 t;
+    
+    switch (chunk) {
+      case 1:
+	return;
+      case 2:
+	while (len >= 2) {
+	    *(byte16 *)mem = htons(*(CONST byte16 *)mem);
+	    mem += 2;
+	    len -= 2;
+	}
+	return;
+      case 4:
+	while (len >= 4) {
+	    *(byte32 *)mem = htonl(*(CONST byte32 *)mem);
+	    mem += 4;
+	    len -= 4;
+	}
+	return;
+      case 8:
+	while (len >= 8) {
+	    t = htonl(((CONST byte32 *)mem)[0]);
+	    ((byte32 *)mem)[0] = htonl(((CONST byte32 *)mem)[1]);
+	    ((byte32 *)mem)[1] = t;
+	    mem += 8;
+	    len -= 8;
+	}
+	return;
+      default:
+	break;
+    }
+#endif
+    while (len >= chunk) {
+	for (i = 0; i < chunk / 2; i++) {
+	    j = chunk - i - 1;
+	    c = mem[i];
+	    mem[i] = mem[j];
+	    mem[j] = c;
+	}
+	mem += chunk;
+	len -= chunk;
+    }
+}
+
 /*
  * Turn the (msg) into a (tmsg) and write it on the MsgPort file descriptor.
  * Used when MsgPort slot is using non-native data sizes and endianity.
@@ -546,16 +812,21 @@ static void AlienIO(int fd, uldat slot) {
  * when an exclusive one is started. Must preserve Slot, Fd and other globals!
  */
 static void alienSendMsg(msgport MsgPort, msg Msg) {
-    uldat Len = 0, Tot;
-    byte *t;
-    
-#if defined(CONF__MODULES) || defined (CONF_HW_DISPLAY)
-    uldat N;
-    byte *Src, Type;
-#endif
-    
     uldat save_Slot = Slot;
     int save_Fd = Fd;
+    uldat Len = 0, Tot;
+    byte *t;
+    uldat N;
+#if defined(CONF__MODULES) || defined (CONF_HW_DISPLAY) || defined(CONF__UNICODE)
+    byte *Src;
+#endif
+#if defined(CONF__MODULES) || defined (CONF_HW_DISPLAY)
+    byte Type;
+#endif
+#ifdef CONF__UNICODE
+    byte16 h;
+    hwattr H;
+#endif
     
     RequestN = MSG_MAGIC;
     Fd = MsgPort->RemoteData.Fd;
@@ -564,7 +835,6 @@ static void alienSendMsg(msgport MsgPort, msg Msg) {
     switch (Msg->Type) {
 #if defined(CONF__MODULES) || defined (CONF_HW_DISPLAY)
       case MSG_DISPLAY:
-
 	Src = Msg->Event.EventDisplay.Data;
 	N = 1;
 	
@@ -605,14 +875,11 @@ static void alienSendMsg(msgport MsgPort, msg Msg) {
 		PushV(t, N, Src);
 # ifdef CONF__UNICODE
 	    } else if (Type == TWS_hwattr && AlienMagic(Slot)[Type] == 2) {
-		/* on the fly conversion from Unicode to IBM437 */
-		byte16 h;
-		hwattr H;
-		
+		/* on the fly conversion from Unicode to CP437 */
 		while (N--) {
 		    Pop(Src,hwattr,H);
 		    
-		    h = HWATTR16(HWCOL(H),Tutf_UTF_16_to_IBM437(HWFONT(H)));
+		    h = HWATTR16(HWCOL(H),Tutf_UTF_16_to_CP437(HWFONT(H)));
 		    alienPush((CONST byte *)&h, sizeof(hwattr), &t, 2);
 		}
 # endif
@@ -628,34 +895,35 @@ static void alienSendMsg(msgport MsgPort, msg Msg) {
 	
 	break;
 #endif /* defined(CONF__MODULES) || defined (CONF_HW_DISPLAY) */
-	case MSG_WIDGET_KEY:
+      case MSG_WIDGET_KEY:
 	alienReply(Msg->Type, Len = SIZEOF(uldat) + 3*SIZEOF(udat) + 2*SIZEOF(byte) + Msg->Event.EventKeyboard.SeqLen, 0, NULL);
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventKeyboard.W));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventKeyboard.W));
 	    alienPUSH(t, udat,  Msg->Event.EventKeyboard.Code);
 	    alienPUSH(t, udat,  Msg->Event.EventKeyboard.ShiftFlags);
 	    alienPUSH(t, udat,  Msg->Event.EventKeyboard.SeqLen);
 	    alienPUSH(t, byte,  Msg->Event.EventKeyboard.pad);
-	    PushV(t, Msg->Event.EventKeyboard.SeqLen+1, Msg->Event.EventKeyboard.AsciiSeq);
+	    PushV(t, Msg->Event.EventKeyboard.SeqLen + 1, /* the final '\0' */
+		  Msg->Event.EventKeyboard.AsciiSeq);
 	}
 	break;
-	case MSG_WIDGET_MOUSE:
+      case MSG_WIDGET_MOUSE:
 	alienReply(Msg->Type, Len = SIZEOF(uldat) + 4*SIZEOF(udat), 0, NULL);
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventMouse.W));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventMouse.W));
 	    alienPUSH(t, udat,  Msg->Event.EventMouse.Code);
 	    alienPUSH(t, udat,  Msg->Event.EventMouse.ShiftFlags);
 	    alienPUSH(t, dat,   Msg->Event.EventMouse.X);
 	    alienPUSH(t, dat,   Msg->Event.EventMouse.Y);
 	}
 	break;
-	case MSG_WIDGET_CHANGE:
+      case MSG_WIDGET_CHANGE:
 	alienReply(Msg->Type, Len = SIZEOF(uldat) + 4*SIZEOF(dat), 0, NULL);
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventWidget.W));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventWidget.W));
 	    alienPUSH(t, udat,  Msg->Event.EventWidget.Code);
 	    alienPUSH(t, udat,  Msg->Event.EventWidget.Flags);
 	    alienPUSH(t, udat,  Msg->Event.EventWidget.XWidth);
@@ -677,17 +945,17 @@ static void alienSendMsg(msgport MsgPort, msg Msg) {
 	alienReply(Msg->Type, Len = SIZEOF(uldat) + 2*SIZEOF(dat) + SIZEOF(uldat), 0, NULL);
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventMenu.W));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventMenu.W));
 	    alienPUSH(t, udat,  Msg->Event.EventMenu.Code);
 	    alienPUSH(t, udat,  Msg->Event.EventMenu.pad);
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventMenu.Menu));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventMenu.Menu));
 	}
 	break;
       case MSG_SELECTION:
 	alienReply(Msg->Type, Len = SIZEOF(uldat) + 4*SIZEOF(dat), 0, NULL);
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventSelection.W));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventSelection.W));
 	    alienPUSH(t, udat,  Msg->Event.EventSelection.Code);
 	    alienPUSH(t, udat,  Msg->Event.EventSelection.pad);
 	    alienPUSH(t, dat,   Msg->Event.EventSelection.X);
@@ -695,28 +963,65 @@ static void alienSendMsg(msgport MsgPort, msg Msg) {
 	}
 	break;
       case MSG_SELECTIONNOTIFY:
-	alienReply(Msg->Type, Len = 4*SIZEOF(uldat) + 2*SIZEOF(dat) + MAX_MIMELEN
-		   + Msg->Event.EventKeyboard.SeqLen, 0, NULL);
+	N = Msg->Event.EventSelectionNotify.Len;
+#ifdef CONF__UNICODE
+	if (Msg->Event.EventSelectionNotify.Magic == SEL_HWFONTMAGIC)
+	    N = (N / sizeof(hwfont)) * SIZEOF(hwfont);
+#endif
+	alienReply(Msg->Type, Len = 4*SIZEOF(uldat) + 2*SIZEOF(dat) + MAX_MIMELEN + N, 0, NULL);
+
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventSelectionNotify.W));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventSelectionNotify.W));
 	    alienPUSH(t, udat,  Msg->Event.EventSelectionNotify.Code);
 	    alienPUSH(t, udat,  Msg->Event.EventSelectionNotify.pad);
 	    alienPUSH(t, uldat, Msg->Event.EventSelectionNotify.ReqPrivate);
 	    alienPUSH(t, uldat, Msg->Event.EventSelectionNotify.Magic);
 	    PushV(t, MAX_MIMELEN, Msg->Event.EventSelectionNotify.MIME);
-	    alienPUSH(t, uldat, Msg->Event.EventSelectionNotify.Len);
-	    PushV(t, Msg->Event.EventSelectionNotify.Len, Msg->Event.EventSelectionNotify.Data);
+	    
+#ifdef CONF__UNICODE
+	    /* client may be not unicode aware while we are */
+	    if (Msg->Event.EventSelectionNotify.Magic == SEL_HWFONTMAGIC) {
+		N = Msg->Event.EventSelectionNotify.Len;
+		
+		if (SIZEOF(hwfont) == 1) {
+		    /* on the fly conversion from Unicode to CP437 */
+		    Src = Msg->Event.EventSelectionNotify.Data;
+		
+		    N /= sizeof(hwfont);
+		    alienPUSH(t, uldat, N);
+		
+		    while (N--) {
+			Pop(Src,hwattr,H);
+			
+			h = HWATTR16(HWCOL(H),Tutf_UTF_16_to_CP437(HWFONT(H)));
+			alienPush((CONST byte *)&h, sizeof(hwattr), &t, 2);
+		    }
+		} else {
+		    N = (N / sizeof(hwfont)) * SIZEOF(hwfont);
+		    alienPUSH(t, uldat, N);
+		    N = Msg->Event.EventSelectionNotify.Len;
+		    alienWriteVec(Msg->Event.EventSelectionNotify.Data, t, N,
+				  sizeof(hwfont), SIZEOF(hwfont),
+				  AlienXendian(Slot) == MagicAlienXendian);
+		    t += N;
+		}
+	    } else
+#endif
+	    {
+		alienPUSH(t, uldat, Msg->Event.EventSelectionNotify.Len);
+		PushV(t, Msg->Event.EventSelectionNotify.Len, Msg->Event.EventSelectionNotify.Data);
+	    }
 	}
 	break;
       case MSG_SELECTIONREQUEST:
 	alienReply(Msg->Type, Len = SIZEOF(uldat) + 2*SIZEOF(dat) + 2*SIZEOF(ldat), 0, NULL);
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventSelectionRequest.W));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventSelectionRequest.W));
 	    alienPUSH(t, udat,  Msg->Event.EventSelectionRequest.Code);
 	    alienPUSH(t, udat,  Msg->Event.EventSelectionRequest.pad);
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventSelectionRequest.Requestor));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventSelectionRequest.Requestor));
 	    alienPUSH(t, uldat, Msg->Event.EventSelectionRequest.ReqPrivate);
 	}
 	break;
@@ -726,23 +1031,28 @@ static void alienSendMsg(msgport MsgPort, msg Msg) {
 	      + Msg->Event.EventControl.Len, 0, NULL);
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventControl.W));
+	         PUSH(t, uldat, Obj2Id(Msg->Event.EventControl.W));
 	    alienPUSH(t, udat,  Msg->Event.EventControl.Code);
 	    alienPUSH(t, udat,  Msg->Event.EventControl.Len);
 	    alienPUSH(t, dat,   Msg->Event.EventControl.X);
 	    alienPUSH(t, dat,   Msg->Event.EventControl.Y);
-	    PushV(t, Msg->Event.EventControl.Len + 1 /*the '\0'*/, Msg->Event.EventControl.Data);
+	    PushV(t, Msg->Event.EventControl.Len, Msg->Event.EventControl.Data);
 	}
 	break;
 	
       case MSG_USER_CLIENTMSG:
-	alienReply(Msg->Type, Len = SIZEOF(uldat) + 2*SIZEOF(dat) + Msg->Event.EventClientMsg.Len, 0, NULL);
+	alienReply(Msg->Type, Len = 2*SIZEOF(uldat) + 2*SIZEOF(dat) + Msg->Event.EventClientMsg.Len, 0, NULL);
 	if ((t = RemoteWriteGetQueue(Slot, &Tot)) && Tot >= Len) {
+	    event_clientmsg *E = &Msg->Event.EventClientMsg;
+	    
 	    t += Tot - Len;
-	    PUSH(t, uldat, Obj2Id(Msg->Event.EventClientMsg.W));
-	    alienPUSH(t, udat,  Msg->Event.EventClientMsg.Code);
-	    alienPUSH(t, udat,  Msg->Event.EventClientMsg.Len);
-	    PushV(t, Msg->Event.EventClientMsg.Len, Msg->Event.EventClientMsg.Data);
+	         PUSH(t, uldat, Obj2Id(E->W));
+	    alienPUSH(t, udat,  E->Code);
+	    alienPUSH(t, udat,  E->Format);
+	    alienPUSH(t, uldat, E->Len);
+	    if (AlienXendian(Slot) == MagicAlienXendian && E->Format > 1)
+		FlipMoveMem(E->Data.b, E->Len, E->Format);
+	    PushV(t, E->Len, E->Data.b);
 	}
 	break;
 	
@@ -753,6 +1063,103 @@ static void alienSendMsg(msgport MsgPort, msg Msg) {
     Fd = save_Fd;
 }
 
+#ifdef CONF_EXT
+
+#define _obj .TWS_field_obj
+#define _any .TWS_field_scalar
+#define _vec .TWS_field_vecV
+#define _len .TWS_field_vecL
+#define _type .type
+
+static byte alienDecodeExtension(tany *Len, CONST byte **Data, tany *Args_n, tsfield a) {
+    static byte type_warned = 0;
+    topaque n = 0;
+    ldat fail = 1;
+    tany len, left = *Len;
+    CONST byte *data = *Data;
+    tany args_n = *Args_n;
+    udat t;
+
+    while (fail > 0 && n < args_n) {
+	switch ((t = a[n]_type)) {
+
+# define CASE_(type) \
+case CAT(TWS_,type): \
+    /* ensure type size WAS negotiated */ \
+    if ((len = AlienSizeof(type, Slot)) && left >= len) { \
+	type an; \
+	\
+	left -= len; \
+	POP(data,type,an); \
+	a[n]_any = (tany)an; \
+    } else \
+	fail = -fail; \
+    break
+
+	  case TWS_hwcol:
+	    /*FALLTHROUGH*/
+	    CASE_(byte);
+	    CASE_(dat);
+	    CASE_(ldat);
+	    CASE_(topaque);
+	    CASE_(tany);
+	    CASE_(hwfont);
+	    CASE_(hwattr);
+#undef CASE_
+
+	  case TWS_vec|TWS_vecW|TWS_byte:
+	    /* ensure (topaque) size WAS negotiated */
+	    if ((len = AlienSizeof(topaque, Slot)) && left >= len) {
+		topaque nlen;
+		
+		left -= len;
+		POP(data,topaque,nlen);
+		a[n]_len = nlen;
+		
+		if (!nlen || Left(nlen)) {
+		    void *addr;
+		    left -= nlen;
+		    POPADDR(data,byte,nlen,addr);
+		    a[n]_vec = addr;
+		    break;
+		}
+	    }
+	    fail = -fail;
+	    break;
+	  default:
+	    if (type_warned < 5) {
+		type_warned = 5;
+		printk("twin: sockDecodeExtension(): got a call with unknown type 0x%02X' !\n",
+		       (int)t);
+	    }
+	    fail = -fail;
+	    break;
+	}
+	
+	if (fail <= 0)
+	    break;
+	
+	fail++;
+	n++;
+    }
+    
+    if (fail > 0) {
+	*Len -= data - *Data;
+	*Data = data;
+	*Args_n = n;
+    }
+    
+    return fail > 0;
+}
+
+#undef _obj
+#undef _any
+#undef _vec
+#undef _len
+#undef _type
+
+#endif /* CONF_EXT */
+
 #undef SIZEOF
 #undef REPLY
 #undef POP
