@@ -26,13 +26,21 @@
 #include "common.h"
 #include "tty.h"
 
-#include "Tw/Tw.h"
-#include "Tw/Twstat.h"
+#include <Tw/Tw.h>
+#include <Tw/Twstat.h>
+#include <Tw/Twstat_defs.h>
 
 #ifdef CONF__UNICODE
-# include "Tutf/Tutf.h"
+# include <Tutf/Tutf.h>
 #endif
 
+#ifdef CONF__UNICODE
+extern hwattr extra_POS_INSIDE;
+#else
+# define extra_POS_INSIDE 0
+#endif
+
+
 /*
  * VT102 emulator
  */
@@ -75,6 +83,7 @@ static byte dirtyN;
 #define nPar		Data->nPar
 #define currG		Data->currG
 #define Charset		Win->Charset
+#define InvCharset	Data->InvCharset
 #define G		Data->G
 #define G0		Data->G0
 #define G1		Data->G1
@@ -86,7 +95,7 @@ static byte dirtyN;
 #define utf_char	Data->utf_char
 #define newLen		Data->newLen
 #define newMax		Data->newMax
-#define newName	Data->newName
+#define newName		Data->newName
 
 
 /* A bitmap for codes <32. A bit of 1 indicates that the code
@@ -195,7 +204,7 @@ static void invert_screen(void) {
     
     while (count--) {
 	a = *p;
-	*p++ = HWATTR( COL( COLBG(HWCOL(a)), COLFG(HWCOL(a))), HWFONT(a));
+	*p++ = (a & ~HWATTR(MAXHWCOL, 0)) | HWATTR(COL(COLBG(HWCOL(a)), COLFG(HWCOL(a))), 0);
 	if (p == Split) p = Base;
     }
 }
@@ -211,7 +220,7 @@ static void insert_char(ldat nr) {
 	p[nr] = *p;
     
     while (nr--)
-	*q++ = HWATTR(ColText, ' ');
+	*q++ = HWATTR(ColText, ' ') | extra_POS_INSIDE;
 
     *Flags &= ~TTY_NEEDWRAP;
 }
@@ -229,7 +238,7 @@ INLINE void delete_char(ldat nr) {
 	p++;
     }
     while (nr--)
-	*p++ = HWATTR(ColText, ' ');
+	*p++ = HWATTR(ColText, ' ') | extra_POS_INSIDE;
     
     *Flags &= ~TTY_NEEDWRAP;
 }
@@ -364,7 +373,7 @@ static void scrollup(dat t, dat b, dat nr) {
     }
     
     /* clear the last nr lines */
-    fill(d + (b-t-nr) * SizeX, HWATTR(ColText, ' '), nr * SizeX);
+    fill(d + (b-t-nr) * SizeX, HWATTR(ColText, ' ') | extra_POS_INSIDE, nr * SizeX);
     
     if (accel)
 	ScrollFirstWindowArea(0, t, SizeX-1, b-1, 0, -nr);
@@ -391,7 +400,7 @@ static void scrolldown(dat t, dat b, dat nr) {
     step = SizeX * nr;
 
     rev_copy(s, s + step, (b-t-nr)*SizeX);
-    fill(s, HWATTR(ColText, ' '), step);
+    fill(s, HWATTR(ColText, ' ') | extra_POS_INSIDE, step);
 
     if (accel)
 	ScrollFirstWindowArea(0, t, SizeX-1, b-1, 0, nr);
@@ -468,7 +477,7 @@ static void csi_J(int vpar) {
       default:
 	return;
     }
-    fill(start, HWATTR(ColText, ' '), count);
+    fill(start, HWATTR(ColText, ' ') | extra_POS_INSIDE, count);
     
     *Flags &= ~TTY_NEEDWRAP;
 }
@@ -497,7 +506,7 @@ static void csi_K(int vpar) {
 	return;
     }
     while (count--)
-	*start++ = HWATTR(ColText, ' ');
+	*start++ = HWATTR(ColText, ' ') | extra_POS_INSIDE;
 
     *Flags &= ~TTY_NEEDWRAP;
 }
@@ -513,7 +522,7 @@ static void csi_X(int vpar) /* erase the following vpar positions */
     dirty_tty(X, Y, X+vpar-1, Y);
     
     while (vpar--)
-	*start++ = HWATTR(ColText, ' ');
+	*start++ = HWATTR(ColText, ' ') | extra_POS_INSIDE;
     
     *Flags &= ~TTY_NEEDWRAP;
 }
@@ -541,9 +550,18 @@ static void update_eff(void) {
 #ifdef CONF__UNICODE
 
 # define setCharset(g) do switch ((currG = (g))) { \
-    case LAT1_MAP: Charset = Tutf_ISO_8859_1_to_UTF_16; break; \
-    case IBMPC_MAP: Charset = Tutf_IBM437_to_UTF_16; break; \
-    case USER_MAP: Charset = All->Gtranslations[USER_MAP]; break; \
+  case LAT1_MAP: \
+    Charset = Tutf_ISO_8859_1_to_UTF_16; \
+    InvCharset = Tutf_UTF_16_to_ISO_8859_1; \
+    break; \
+  case IBMPC_MAP: \
+    Charset = Tutf_CP437_to_UTF_16; \
+    InvCharset = Tutf_UTF_16_to_CP437; \
+    break; \
+  case USER_MAP: \
+    Charset = All->Gtranslations[USER_MAP]; \
+    InvCharset = Tutf_UTF_16_to_ISO_8859_1; /* very rough :( */ \
+    break; \
 } while (0)
 
 INLINE hwfont applyG(hwfont c) {
@@ -672,7 +690,7 @@ static void respond_string(byte *p) {
 	/* or we may need to send a Msg to Win->Owner */
 	msg Msg;
 	event_keyboard *Event;
-	if ((Msg = Do(Create,Msg)(FnMsg, MSG_WIDGET_KEY, Len + sizeof(event_keyboard)))) {
+	if ((Msg = Do(Create,Msg)(FnMsg, MSG_WIDGET_KEY, Len))) {
 	    /* this is the same code as in KeyboardEvent() in hw.c */
 	    Event = &Msg->Event.EventKeyboard;
 	    Event->W = (widget)Win;
@@ -906,11 +924,13 @@ static void reset_tty(byte do_clear) {
     
     G = saveG = 0;
     /*
-     * this probably violates some standard, 
-     * but starting with the identity mapping
-     * seems the only reasonable choice to me
+     * default to latin1 charset if CONF__UNICODE is enabled
      */
+#ifdef CONF__UNICODE
+    setCharset(G0 = saveG0 = LAT1_MAP);
+#else
     setCharset(G0 = saveG0 = IBMPC_MAP);
+#endif
     G1 = saveG1 = GRAF_MAP;
 
     utf = utf_count = utf_char = 0;
@@ -954,35 +974,18 @@ static byte insert_newtitle(byte c) {
 }
 
 static void set_newtitle(void) {
-    widget P;
+    dat _Len;
+    byte *_Name;
     
-    if (Win->Name)
-	FreeMem(Win->Name);
 
     /* try to shrink... */
-    if (!(Win->Name = ReAllocMem(newName, newLen)))
-	Win->Name = newName;
-    Win->NameLen = newLen;
+    if (!(_Name = ReAllocMem(newName, _Len = newLen)))
+	_Name = newName;
+
     newLen = newMax = 0;
     newName = (byte *)0;
 
-#if 1
-    /*
-     * do not allow changing window borders just because
-     * some untrusted application set a new title
-     */
-    DrawBorderWindow(Win, BORDER_UP);
-#else
-    /* user may have title-dependent borders in ~/.twinrc, honour them: */
-    Win->BorderPattern[0] = Win->BorderPattern[1] = NULL;
-    DrawBorderWindow(Win, BORDER_ANY);
-#endif
-    
-    if ((P = Win->Parent) && IS_SCREEN(P)) {
-	/* need to update window list with new name ? */
-	if (((screen)P)->FnHookW)
-	    ((screen)P)->FnHookW(((screen)P)->HookW);
-    }
+    Act(SetTitle,Win)(Win, _Len, _Name);
 }
 
 static void clear_newtitle(void) {
@@ -1324,7 +1327,7 @@ INLINE void write_ctrl(byte c) {
 	if (c == '8') {
 	    /* DEC screen alignment test */
 	    dirty_tty(0, 0, SizeX-1, SizeY-1);
-	    fill(Start, HWATTR(ColText, 'E'), (ldat)SizeX * SizeY);
+	    fill(Start, HWATTR(ColText, 'E') | extra_POS_INSIDE, (ldat)SizeX * SizeY);
 	}
 	break;
 	
@@ -1529,7 +1532,7 @@ void TtyWriteAscii(window Window, ldat Len, CONST byte *AsciiSeq) {
 		insert_char(1);
 	    
 	    dirty_tty(X, Y, X, Y);
-	    *Pos = HWATTR(Color, c);
+	    *Pos = HWATTR(Color, c) | extra_POS_INSIDE;
 	    
 	    if (X == SizeX - 1) {
 		if (*Flags & TTY_AUTOWRAP) 
@@ -1549,7 +1552,7 @@ void TtyWriteAscii(window Window, ldat Len, CONST byte *AsciiSeq) {
 
 #if TW_SIZEOFHWFONT == 1
 void TtyWriteHWFont(window Window, ldat Len, CONST hwfont *HWFont) {
-    TtyWriteAscii(Window, Len, HWFont);
+    TtyWriteAscii(Window, Len, (CONST byte *)HWFont);
 }
 #else
 /* same as TtyWriteAscii(), but writes hwfont (unicode). Useful only if CONF__UNICODE is enabled. */
@@ -1587,7 +1590,7 @@ void TtyWriteHWFont(window Window, ldat Len, CONST hwfont *HWFont) {
 		insert_char(1);
 	    
 	    dirty_tty(X, Y, X, Y);
-	    *Pos = HWATTR(Color, c);
+	    *Pos = HWATTR(Color, c) | extra_POS_INSIDE;
 	    
 	    if (X == SizeX - 1) {
 		if (*Flags & TTY_AUTOWRAP) 
@@ -1627,7 +1630,7 @@ void TtyWriteString(window Window, ldat Len, CONST byte *String) {
 	}
 	    
 	dirty_tty(X, Y, X, Y);
-	*Pos = HWATTR(Color, c);
+	*Pos = HWATTR(Color, c) | extra_POS_INSIDE;
 	    
 	if (X == SizeX - 1) {
 	    if (*Flags & TTY_AUTOWRAP) 
@@ -1647,6 +1650,10 @@ void TtyWriteString(window Window, ldat Len, CONST byte *String) {
  */
 void TtyWriteHWAttr(window Window, dat x, dat y, ldat len, CONST hwattr *text) {
     ldat left, max, chunk;
+#ifdef CONF__UNICODE
+    ldat i;
+    hwattr extra;
+#endif
     hwattr *dst;
     
     if (!Window || !len || !text || !W_USE(Window, USECONTENTS) || !Window->USE.C.TtyData)
@@ -1682,11 +1689,20 @@ void TtyWriteHWAttr(window Window, dat x, dat y, ldat len, CONST hwattr *text) {
 	    dst -= Split - Base;
 	max = Split - dst;
 	chunk = Min2(left, max);
+#ifdef CONF__UNICODE
+	for (i = chunk; i; i--, text++, dst++) {
+	    if ((extra = HWEXTRA32(*text)))
+		*dst = *text;
+	    else
+		*dst = *text | extra_POS_INSIDE;
+	}
+#else
 	CopyMem(text, dst, chunk * sizeof(hwattr));
 	text += chunk;
 	dst += chunk;
+#endif
     } while ((left -= chunk) > 0);
-    
+
     if (len > SizeX - x)
 	dirty_tty(0, y, SizeX - 1, y + (x + len - 1) / SizeX);
     else
